/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.adapter;

import java.util.Iterator;

import dk.rode.thesis.iterator.IterableSequence;
import dk.rode.thesis.iterator.SequenceIterator;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.model.Sequence;

/**
  * An <i>object adapter</i> that adapts a {@link Sequence} type 
  * to conform to the {@code Iterator} type: when the iterator is 
  * advanced, so is the sequence - hence the adaptation. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    IterableSequence
  * @see    SequenceIterator
  * 
  * @param  <E> The type of values delivered by this iterator.
  */
@ParticipantUsage(value = "Target", type = Iterator.class)
@Participant("Adapter")
public class IteratorSequence<E> implements Iterator<E> { // does NOT implement sequence!
    
    /**
      * The adapted sequence in form of an {@link SequenceIterator}
      * instance: the responsibility is designated to the sequence
      * iterator because it already implements the desired
      * functionality. This makes the adaption easy, as we merely
      * forward the method invocations. <p>
      * 
      * Never null.
      */
    private final SequenceIterator<E> iterator;
    
    /**
      * Constructor, which creates this iterator sequence to iterate
      * through the sequence value from {@code sequence}, starting
      * from its current value. <p>
      * 
      * {@code sequence} must be {@link Sequence#bounded() bounded}.
      * 
      * @param  sequence The sequence to adapt as an iterator; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      * @throws IllegalArgumentException If {@code sequence} is not bounded.
      */
    public IteratorSequence(@Participant("Adaptee") final Sequence<E> sequence) {
        this.iterator = new SequenceIterator<E>(sequence);
    }

    /**
      * 
      */
    public boolean hasNext() {
        return this.iterator.hasNext();
    }

    /**
      * 
      */
    public E next() {
        return this.iterator.next();
    }

    /**
      *  
      */
    public void remove() {
        this.iterator.remove();
    }

    
}
