/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */

/**
  * Implementations and examples of the <b>Adapter</b> 
  * design pattern [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.139]. <p>
  *  
  * <b>Intent</b>:<ul><p><i>
  * 
  *    Convert the interface of a class into another interface 
  *    clients expect. Adapter lets classes work together that 
  *    could not otherwise because of incompatible interfaces.
  *     
  * </i></ul>
  *
  * Here, the <b>Adaptee</b> participant is represented by any
  * {@link dk.rode.thesis.meta.model.Sequence} implementation,
  * for example {@link dk.rode.thesis.meta.model.HexDigitSequence}. <p>
  * 
  * The {@link dk.rode.thesis.adapter.IteratorSequence} and 
  * {@link dk.rode.thesis.adapter.SequenceAdapter} classes
  * represent the <b>Adapter</b> participant. <p>
  * 
  * An {@code IteratorSequence} has <i>object scope</i>. The <b>Target</b> 
  * of an {@code IteratorSequence} is the {@link java.util.Iterator 
  * java.util.Iterator} class. An {@code IteratorSequence} thus allows 
  * iteration over sequence values using standard Java iterators as oppose 
  * to using the specific {@code Sequence} interface. The 
  * {@code IteratorSequence} class does <b>not</b> implement the 
  * {@code Sequence} interface. <p>
  *  
  * A {@code SequenceAdapter} essentially has <i>class scope</i>, but is an
  * <i>object adapter</i>, and its <b>Target</b> participant is represented 
  * by another {@code Sequence} declared with an incompatible type parameter. 
  * It utilises {@link dk.rode.thesis.adapter.AdapterDelegate} instances to 
  * perform the actual representation shift between types {@code E} and 
  * {@code T}, for example from an {@code Integer} value to a {@code String} 
  * representation of that value. <p>
  *   
  * The <b>Client</b> participant is the test application, i.e. the 
  * {@link dk.rode.thesis.adapter.Main} class. <p>
  * 
  * <a name="uml"></a><b>UML Class Diagram</b>:<p>
  * <img src="doc-files/Adapter.png" border="0"><p>
  * 
  * <b>Implementation notes</b>:<br>
  * The {@code SequenceAdapter} class is an <i>object</i> adapter in that 
  * it forwards all sequence requests to the adapted sequence, but a bit 
  * unusual in that it does not adapt the <i>raw type</i>, but the 
  * <i>instantiated generic type</i> of its own raw type. In that respect it 
  * is more related to <i>class</i> adaptation than object adaptation. <p>
  * 
  * Java does not support multiple functional inheritance, so 
  * <i>class adapters</i> as described by Gamma et al. cannot be 
  * constructed directly. However, using {@link java.lang.reflect.Proxy 
  * java.lang.reflect.Proxy} objects, an object can be adapted to support 
  * interfaces dynamically using reflection. As the relationship is dynamic, 
  * however, it is related to specific instances, not to a given class directly. <p>
  *  
  * This technique is extensively used in this thesis in the 
  * {@link dk.rode.thesis.meta.reflect.proxy} package. For example, the 
  * {@link dk.rode.thesis.meta.reflect.proxy.ProxyFactory} class uses a 
  * private internal {@code Proxy} target interface that all created 
  * proxies will be adapted to, regardless of their type. This enables 
  * the proxy factory to access all proxies internally using this common 
  * interface, i.e. the factory acts as the client. The adapted proxies and 
  * their context are unaware of the adaptation, and all requests to the 
  * {@code Proxy} interface is never seen by the adaptee: all functionality 
  * is "inherited" similar to multiple functional inheritance. The 
  * {@link dk.rode.thesis.meta.reflect.proxy.Reference} interface is also 
  * a target interface adapted by <i>shared objects</i> obtained via a 
  * {@code ProxyFactory} as well. <p>
  * 
  * A final example is the {@link dk.rode.thesis.prototype.PrototypeFactory} 
  * from the {@linkplain dk.rode.thesis.prototype Prototype} pattern 
  * implementation, which adapts objects of any class declaring a copy 
  * constructor to the {@link dk.rode.thesis.prototype.Copyable} target 
  * interface. The adaptation is public and the context can thus cast the 
  * (proxy) object to and from the {@code Copyable} type and directly 
  * invoke its {@code copy()} method. <p>
  * 
  * It is standard in Java to use anonymous inner classes to adapt to a
  * required target interface/class, for example to implement a handler of
  * some sort. The adapter strategies declared in the 
  * {@link dk.rode.thesis.adapter.AdapterStrategy} class are all
  * anonymous inner classes that adapts the {@link dk.rode.thesis.adapter.AdapterDelegate}
  * interface. Hence, adaptation is used to construct the required
  * strategies to adapt a given sequence adapter to a target type!
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a> 
  */
@Pattern(name         = "Adapter", 
         scope        = {Scope.Class, Scope.Object},
         purpose      = Purpose.Structural,
         participants = {"Target", "Adapter", "Adaptee", "Client"}) 
package dk.rode.thesis.adapter;

import dk.rode.thesis.meta.Pattern;
import dk.rode.thesis.meta.Purpose;
import dk.rode.thesis.meta.Scope;






