/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.bridge;

import static dk.rode.thesis.meta.util.Const.INFINITE;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.AbstractSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.reflect.Caller;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.SequencePolicy;
import dk.rode.thesis.strategy.StringablePolicy;
import dk.rode.thesis.strategy.StringablePolicy.Type;

/**
  * A <i>sequence abstraction</i> defers the creation of
  * {@link Sequence sequence} values to a value 
  * {@link SequenceValueGenerator generator}. <p>
  * 
  * The generator to use can either be supplied at
  * {@link #SequenceAbstraction(SequenceValueGenerator) construction}
  * time, or any time before sequence use via the 
  * {@link #setGenerator(SequenceValueGenerator)} method. <b>Any</b> 
  * sequence method expect {@link #copy() copy} and {@code setGenerator}
  * will throw an {@code IllegalStateException} if accessed 
  * before a generator has been set! The {@link #toString(StringablePolicy)}
  * method will only fail in case the methods described above is used by 
  * the specified {@link SequencePolicy}. <p>
  * 
  * The {@link SequenceValueGenerator#size() size}, 
  * {@link SequenceValueGenerator#ordered() ordered}, and 
  * {@link SequenceValueGenerator#duplicates() duplicate} properties of a 
  * {@code SequenceValueGenerator} makes it <b>very</b> easy to alter the
  * behaviour of a given sequence abstraction! <p> 
  *
  * <b>Implementation notes</b>:<br>
  * The separation between the abstraction and implementation allows the
  * abstraction to use any kind of implementation that delivers 
  * a sub-type of the value delivered by the sequence. Here, the implementation
  * is only allowed to be set once, because changing a sequence while
  * traversing it seems illogical, but other Bridge implementations may benefit
  * from changing the implementation even after the abstraction is in use. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by this sequence. Note that the
  *             values used by this sequence may generate any type {@code S} 
  *             that satisfies {@code <S extends E>}.
  */
@Participant("Abstraction")
public class SequenceAbstraction<E> extends AbstractSequence<E> implements Sequence<E> {
    
    /**
      * The {@link SequenceValueGenerator generator} used to
      * generate the next sequence value on an invocation of 
      * {@link #next()}. <p> 
      * 
      * Supplied either at construction time, or set explicitly
      * using the {@link #setGenerator(SequenceValueGenerator)}
      * method. Can be set at most once, i.e. the generator
      * cannot change once set, but must be set before this
      * sequence is valid and can be used. <p>
      * 
      * Never empty, and never null for all practical purposes. 
      */
    @Participant("Implementation")
    protected SequenceValueGenerator<? extends E> generator;
    /**
      * The current sequence value. <p>
      * 
      * Never null if this sequence has had a generator set.
      */
    protected E current;

    /**
      * Constructor. <p>
      * 
      * A value {@link SequenceValueGenerator generator} <b>must</b> explicitly 
      * be set using the {@link #setGenerator(SequenceValueGenerator)} 
      * method before this sequence is valid!  
      */
    public SequenceAbstraction() {
        super(false); // no reset!
        this.generator = null;
        this.current = null; // invalid!
        this.state = null;   // invalid!
    }
    
    /**
      * Constructor, which creates this sequence abstraction to 
      * use the value {@link SequenceValueGenerator generator}
      * supplied as {@code generator}. <p>
      *
      * The {@link #setGenerator(SequenceValueGenerator)} must <b>not</b>
      * be invoked for this sequence.
      * 
      * @param  generator The generator to use; cannot be null.
      * @throws NullPointerException If {@code generator} is null.
      */
    public SequenceAbstraction(@Participant("Implementation") SequenceValueGenerator<? extends E> generator) {
        super(false);
        this.setGenerator(generator);
    }

    /**
      * Copy constructor. <p>
      * 
      * If {@code sequence} does not have a value {@link SequenceValueGenerator 
      * generator} set, a generator <b>must</b> explicitly 
      * be set for this sequence using the 
      * {@link #setGenerator(SequenceValueGenerator)} method before 
      * this sequence is valid!  
      * 
      * @param  sequence The sequence abstraction to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public SequenceAbstraction(SequenceAbstraction<E> sequence) {
        super(sequence);
        // A state means a generator has been set...
        if (sequence.state != null) {
            this.generator = sequence.generator.copy();
            int number = sequence.generator.number();
            this.current = (number < 1) 
                         ? (E)this.generator.first() 
                         : this.generator.get(number);
        // Not set...
        } else {
            this.generator = null;
            this.current = null;
            this.state = null;
        }
    }
    
    /**
      * 
      */
    public SequenceAbstraction<E> copy() {
        return new SequenceAbstraction<E>(this);
    }

    /**
      * Sets the {@link SequenceValueGenerator generator} to use
      * for this sequence to {@code generator}. <p>
      * 
      * A generator can only be set once, and cannot be changed 
      * once set. This sequence is not valid until a generator
      * has been set.
      * 
      * @param  generator The generator to use; cannot be null.
      * @return This sequence abstraction; never null.
      * @throws NullPointerException If {@code generator} is null.
      * @throws IllegalStateException If a generator has already been set.
      * @see    #SequenceAbstraction(SequenceValueGenerator)
      */
    public SequenceAbstraction<E> setGenerator(@Participant("Implementation") SequenceValueGenerator<? extends E> generator) 
        throws IllegalStateException {
        if (generator == null) {
            throw new NullPointerException("Generator cannot be null");
        } else if (this.generator != null) {
            throw new IllegalStateException("Generator already set");
        }
        this.generator = generator;
        // Must have an initial value to comply with the Sequence contract...
        this.current = this.generator.first(); 
        this.state = State.START; // valid state!
        return this;
    }
    
    /**
      * Returns true if a generator has been set for this sequence
      * abstraction, false if not. <p>
      * 
      * If this sequence is not initialised, all sequence methods
      * will throw an exception on use!
      *  
      * @return True if initialised, false if not.
      */
    public boolean isInitialised() {
        return (this.generator != null);
    }
    
    /**
      * @throws IllegalStateException If a generator has not been set.
      */
    public State state() throws IllegalStateException {
        if (this.state == null) {
            // Copying is allowed by the super-class (using copy constructors),
            // or by any specialisation of this class. This class never calls
            // this method explicitly...
            Caller caller = new Caller();
            if (!((caller.isSuperClass()) || (caller.isSubClass()))) {
                throw new IllegalStateException("Sequence not initialised");
            }
        }
        return this.state;
    }
    
    /**
      * @throws IllegalStateException If a generator has not been set.
      */
    public final boolean bounded() throws IllegalStateException {
        if (this.state == null) {
            throw new IllegalStateException("Sequence not initialised");
        }
        return (this.generator.size() != INFINITE);
    }

    /**
      * @throws IllegalStateException If a generator has not been set.
      */
    public final boolean consistent() throws IllegalStateException {
        if (this.state == null) {
            throw new IllegalStateException("Sequence not initialised");
        }
        return this.generator.ordered();
    }

    /**
      * @throws IllegalStateException If a generator has not been set.
      */
    public boolean unique() throws IllegalStateException {
        if (this.state == null) {
            throw new IllegalStateException("Sequence not initialised");
        }
        return !this.generator.duplicates();
    }

    /**
      * @throws IllegalStateException If a generator has not been set.
      */
    public E current() throws IllegalStateException {
        if (this.state == null) {
            throw new IllegalStateException("Sequence not initialised");
        }
        return this.current;
    }
    
    /**
      * @throws IllegalStateException If a generator has not been set.
      */
    public E next() throws IllegalStateException {
        if (this.state == null) {
            throw new IllegalStateException("Sequence not initialised");
            
        } else if ((this.current = this.generator.get()) == null) {
            LogFactory.getLog(this).error("Implementation delivered null value: ", this, " -> ", this.generator);
            throw new InternalError("Implementation delivered null value");
        }
        this.state = this.generator.number() == 1 ? State.RESTART : State.NORMAL;
        return this.current;
    }

    /**
      * @throws IllegalStateException If a generator has not been set.
      */
    public void reset() throws IllegalStateException {
        if (this.state == null) {
            throw new IllegalStateException("Sequence not initialised");
        }
        super.reset();
        this.current = this.generator.first();
    }

    /**
      * 
      */
    public CharSequence toString(StringablePolicy<? super Sequence<E>> policy) {
        // Initialised...
        if (this.state != null) {
            policy = this.getStringablePolicy(policy);
            CharSequence cs = policy.toString(this);
            if (policy.type() == Type.CLOSED) {
                return cs;
            }
            StringBuilder sb = new StringBuilder().append(cs);
            if (policy.type() == Type.OPEN) {
                sb.append(']');
            }
            // Try to find a policy with a similar name to use for the implementation,
            // or use the IDENTITY policy as default...
            SequenceValuePolicy svp = 
                StringablePolicy.NameIdiom.getSimilarByName(SequenceValuePolicy.class, policy, SequenceValuePolicy.IDENTITY);
            // ISSUE: the line below does not compile in NetBeans 5.5.1 with Java 1.6.0_01
            sb.append(".[").append(svp.toString(this.generator)); 
            if (svp.type() == Type.OPEN) {
                sb.append(']');
            }
            return sb.append(']');
            
        // Not initialised...
        } else {
            StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
            return sb.append("; initialised: false").append(']').toString();
        }
    }
    
}
