/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.bridge;

import static dk.rode.thesis.meta.util.Const.EOF;
import static dk.rode.thesis.meta.util.Util.serializableCopy;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;

import dk.rode.thesis.meta.Participant;

/**
  * A <i>sequence value array list</i> stores sequence values 
  * in a {@link java.util.ArrayList java.util.ArrayList} instance. <p>
  * 
  * A sequence value array list is {@link #ordered() ordered},
  * can be {@link #sorted() sorted}, and may or may not contain 
  * {@link #duplicates() duplicates}. <p>
  *
  * <b>Implementation notes</b>:<br>
  * The {@link #get(int)} method is fast as the {@code ArrayList} class 
  * offers random access via indexes. The {@link #numberOf(Serializable, int)} 
  * method uses a binary search on the sub-list of length <i>m</i> identified 
  * by the supplied index if the list is sorted. Hence, the running time 
  * is <i>O(log(m))</i>, where <i>m</i> {@literal <=} <i>n</i> and <i>n</i> is the
  * number of values in the entire value list. If all the elements are 
  * identical or not sorted, the worst-case running time is proportional 
  * to the number of elements in the value list, i.e. <i>O(n)</i>. 
  *   
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of serializable values stored in this value list.
  */
@Participant("ConcreteImplementor")
public class SequenceValueArrayList<E extends Serializable> 
    extends SequenceValueCollection<E,ArrayList<E>> {

    /**
      * True if this list is sorted, false if not. 
      */
    private final boolean sorted;
    /**
      * The comparator used to sort this value list, if any. <p>
      * 
      * If null and {@link #sorted} is true, the comparable natural 
      * ordering is assumed. <p>
      * 
      * Can be null.
      */
    private final Comparator<? super E> comparator;
    /**
      * True if this list contains duplicates, false if it does not. <p>
      * 
      * This can either be <i>specified</i> or <i>calculated</i> 
      * at construction time.
      */
    private final boolean duplicates;

    /**
      * Constructor, which creates this <i>sequence value array list</i> 
      * to use the {@code java.util.ArrayList} instance supplied as 
      * {@code values}. <p>
      * 
      * This value list is assumed to be {@link #sorted() unsorted}, but
      * whether or not {@code values} contains {@link #duplicates() duplicates}
      * <b>is</b> calculated. The worst case calculation running time
      * is <i>O(values.size)</i>. <p>
      * 
      * Modifying {@code values} externally <b>will</b> affect this
      * value set and should <b>not</b> be performed. The result
      * of doing so is undefined - but never good. <p>
      * 
      * {@code values} must not contain null entries. If it does, 
      * the sequence {@link SequenceAbstraction abstraction} using this 
      * value list will generate an internal error when that entry is 
      * returned by this value list!   
      * 
      * @param  type       The type of values stored in this list (and 
      *                    {@code values}); cannot be null.
      * @param  values     The actual {@code ArrayList} instance storing the sequence 
      *                    values; cannot be null, empty, or contain null values.
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code values} is empty, or contain a 
      *         null value discovered during duplicate calculation.
      */
    public SequenceValueArrayList(Class<E> type, ArrayList<E> values) {
        this(type, values, false/*unsorted*/, null/*calculate duplicates*/);
    }

    /**
      * Constructor, which creates this <i>sequence value array list</i> 
      * to use the {@code java.util.ArrayList} instance supplied as 
      * {@code values}. <p>
      * 
      * Modifying {@code values} externally <b>will</b> affect this
      * value set and should <b>not</b> be performed. The result
      * of doing so is undefined - but never good. <p>
      * 
      * {@code values} must not contain null entries. If it does, 
      * the sequence {@link SequenceAbstraction abstraction} using this 
      * value list will generate an internal error when that entry is 
      * returned by this value list!   
      * 
      * @param  type       The type of values stored in this list (and 
      *                    {@code values}); cannot be null.
      * @param  values     The actual {@code ArrayList} instance storing the sequence 
      *                    values; cannot be null, empty, or contain null values.
      * @param  sorted     True if {@code values} is sorted, false if not. If
      *                    true, comparable natural ordering is assumed.
      * @param  duplicates If true, {@code values} is known to contain duplicates;
      *                    if false, {@code values} is known not to contain 
      *                    duplicates; or if null, whether or not {@code values}
      *                    contains duplicates is calculated. Potentially expensive,
      *                    worst running time <i>O(values.size)</i>.                 
      * @throws NullPointerException If either {@code type} or {@code values}
      *         are null.
      * @throws IllegalArgumentException If {@code values} is empty, or contain a 
      *         null value discovered during duplicate calculation.
      */
    public SequenceValueArrayList(Class<E> type, ArrayList<E> values, boolean sorted, Boolean duplicates) {
        this(type, values, 0, sorted, null, duplicates);
    }

    /**
      * Constructor, which creates this <i>sequence value array list</i> 
      * to use the {@code java.util.ArrayList} instance supplied as 
      * {@code values}. <p>
      * 
      * Modifying {@code values} externally <b>will</b> affect this
      * value set and should <b>not</b> be performed. The result
      * of doing so is undefined - but never good. <p>
      * 
      * {@code values} is assumed already sorted by the {@code Comparator}
      * supplied as {@code comparator}; if null, the comparable 
      * natural ordering is assumed. <p>
      * 
      * {@code values} must not contain null entries. If it does, 
      * the sequence {@link SequenceAbstraction abstraction} using this 
      * value list will generate an internal error when that entry is 
      * returned by this value list!   
      * 
      * @param  type       The type of values stored in this list (and 
      *                    {@code values}); cannot be null.
      * @param  values     The actual {@code ArrayList} instance storing the sequence 
      *                    values; cannot be null, empty, or contain null values.
      * @param  comparator The comparator used, if any; can be null, in which case
      *                    comparator natural ordering is assumed.
      * @param  duplicates If true, {@code values} is known to contain duplicates;
      *                    if false, {@code values} is known not to contain 
      *                    duplicates; or if null, whether or not {@code values}
      *                    contains duplicates is calculated. Potentially expensive,
      *                    worst running time <i>O(values.size)</i>.                 
      * @throws NullPointerException If either {@code type} or {@code values}
      *         are null.
      * @throws IllegalArgumentException If {@code values} is empty, or contain a 
      *         null value discovered during duplicate calculation.
      */
    public SequenceValueArrayList(Class<E> type, ArrayList<E> values, Comparator<? super E> comparator, Boolean duplicates) {
        this(type, values, 0, true, comparator, duplicates);
    }
    
    /**
      * Constructor, which creates this <i>sequence value array list</i> 
      * to use the {@code java.util.ArrayList} instance supplied as 
      * {@code values}. <p>
      * 
      * The next value to return by {@link #get()} is the value that 
      * is the {@code number}'th value in this value list, and  
      * this constructor ensures that this value list is initialised 
      * accordingly! <p>
      * 
      * Modifying {@code values} externally <b>will</b> affect this
      * value set and should <b>not</b> be performed. The result
      * of doing so is undefined - but never good. <p>
      * 
      * {@code values} must not contain null entries. If it does, 
      * the sequence {@link SequenceAbstraction abstraction} using this 
      * value list will generate an internal error when that entry is 
      * returned by this value list!   
      * 
      * @param  type       The type of values stored in this list (and 
      *                    {@code values}); cannot be null.
      * @param  values     The actual {@code ArrayList} instance storing the sequence 
      *                    values; cannot be null, empty, or contain null values.
      * @param  number     The number pointing the the {@code number}'th value
      *                    to be returned by the next call to {@link #get()}.                
      * @param  sorted     True if {@code values} is sorted, false if not.
      * @param  comparator The comparator used, if any; can be null, in which case
      *                    comparator natural ordering is assumed.
      * @param  duplicates If true, {@code values} is known to contain duplicates;
      *                    if false, {@code values} is known not to contain 
      *                    duplicates; or if null, whether or not {@code values}
      *                    contains duplicates is calculated. Potentially expensive,
      *                    worst running time <i>O(values.size)</i>.                 
      * @throws NullPointerException If either {@code type} or {@code values}
      *         are null.
      * @throws IllegalArgumentException If {@code values} is empty, or contain a 
      *         null value discovered during duplicate calculation, or if
      *         {@code number} is illegal.
      */
    private SequenceValueArrayList(Class<E> type, ArrayList<E> values, int number, boolean sorted, Comparator<? super E> comparator, Boolean duplicates) {
        super(type, values, number);
        this.sorted = sorted;
        this.comparator = comparator;
        // Calculate duplicates? Expensive, but we can also test for null, but no
        // guarantee that we will traverse the entire list of values. Hence,
        // an error may still occur at runtime during "next()" in the sequence
        // abstraction...
        if (duplicates == null) { 
            duplicates = false;
            HashSet<E> set = new HashSet<E>();
            for (int i = 0, n = this.values.size(); i < n; i++) {
                E value = values.get(i);
                if (value == null) {
                    throw new IllegalArgumentException("Values cannot contain null");
                } else if (set.contains(value)) {
                    duplicates = true;
                    break;
                } else {
                    set.add(value);
                }
            }
        }
        this.duplicates = duplicates;
    }
    
    /**
      * 
      */
    public E first() {
        this.number = 1;
        return this.values.get(0);
    }
    
    /**
      * 
      */
    public E get() {
        if (this.number >= this.values.size()) {
            this.number = 0;
        }
        return this.values.get(this.number++);
    }

    /**
      * 
      */
    public E get(int number) {
        this.checkNumber(number, true);
        E value = this.values.get(number - 1);
        this.number = number;
        return value;
    }

    /**
      * 
      */
    public int numberOf(E value, int fromNumber) {
        if (this.checkNumber(fromNumber, false)) {
            // Sorted...
            if (this.sorted) {
                // Generate a view with the interesting values... This will
                // not cause copy of list values since the sub list is backed
                // its parent, i.e. "values"...
                List<E> view = this.values.subList(fromNumber - 1, this.values.size());
                // "index" is zero-based...
                int index = Collections.binarySearch(view, value, this.comparator);
                if (index == EOF) {
                    return EOF;
                    
                } 
                // Make sure we find the first one always...
                if (this.duplicates) {
                    for (int i = index - 1; i >= 0; i--) {
                        if (view.get(i).equals(value)) {
                            index--;
                        } else {
                            break;
                        }
                    }
                }
                return fromNumber + index;
                
            // Unsorted...
            } else {
                // From "fromNumber" (one-based) onwards...
                for (int i = fromNumber - 1; i < this.values.size(); i++) {
                    if (value.equals(this.values.get(i))) {
                        return i + 1;
                    }
                }
            }
        }
        return EOF;
    }
    
    /**
      * 
      */
    public boolean ordered() {
        return true;
    }
    
    /**
      *
      */
    public boolean sorted() {
        return this.sorted;
    }
    
    /**
      * 
      */
    public boolean duplicates() {
        return this.duplicates;
    }
    
    /**
      * 
      */
    public SequenceValueArrayList<E> copy() {
        return new SequenceValueArrayList<E>(this.type, 
                                             serializableCopy(this.values), 
                                             this.number == 0 ? 0 : this.number - 1,
                                             this.sorted, 
                                             this.comparator,
                                             this.duplicates);
    }
    
}
