/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.bridge;

import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.StringablePolicy;

/**
  * Standard policies for formatting {@link SequenceValueGenerator} 
  * objects into char sequences (<b>not</b> part of the core Bridge implementation). <p>
  *
  * <b>Implementation notes:</b><br>
  * <b>This is NOT part of the core Bridge functionality, merely 
  * to provide better log descriptions!</b><p>
  * 
  * See implementation notes for the {@link ObjectPolicy} enumeration.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    ObjectPolicy
  */
public enum SequenceValuePolicy implements StringablePolicy<SequenceValueGenerator<?>> {
    
    /**
      * Generates a char sequence representation of a value generator 
      * having the following format:
      * <pre>
      *   SequenceValueGenerator
      * </pre>
      * 
      * Where {@code SequenceValueGenerator} is the simple class name of 
      * the class of the values.
      * 
      * @see ObjectPolicy#NAME
      */
    NAME {
        public CharSequence toString(SequenceValueGenerator<?> values) {
            return ObjectPolicy.NAME.toString(values);
        }
    },

    /**
      * Generates a char sequence representation of a value generator 
      * having the following format:
      * <pre>
      *   SequenceValueGenerator[..]
      * </pre>
      * 
      * Where {@code SequenceValueGenerator} is the simple class name 
      * of the class of the values, and {@code ..} is the <b>identity</b> 
      * hash-code of the values.
      * 
      * @see ObjectPolicy#ID
      */
    ID {
        public CharSequence toString(SequenceValueGenerator<?> values) {
            return ObjectPolicy.ID.toString(values);
        }
    },
    
    /**
      * Generates a char sequence representation of a value generator
      * having the following format:
      * <pre>
      *   SequenceValueGenerator[hash-code: ..[; identity-hash-code: ..]; size: ..]
      * </pre>
      * 
      * Where {@code SequenceValueGenerator} is the simple class name 
      * of the class of the values, and {@code ..} are the hash-code(s) 
      * and size of the values, respectively.
      * 
      * @see ObjectPolicy#IDENTITY_OPEN
      */
    IDENTITY {
        public CharSequence toString(SequenceValueGenerator<?> values) {
            StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(values);
            return sb.append("; size: ").append(values.size())
                     .append("; number: ").append(values.number())
                     .append(']');
        }
    },
    
    /**
      * Generates a char sequence representation of a value generator 
      * having the following format:
      * <pre>
      *   SequenceValueGenerator[hash-code: ..[; identity-hash-code: ..]; size: ..; number: ..; [values: ..]] 
      * </pre>
      * 
      * Where {@code SequenceValueGenerator} is the simple class name of
      * the class of the values, and {@code values} list the
      * actual values for value {@link SequenceValueCollection collections}, 
      * ordered and sorted, if possible. <p>
      *
      * @see ObjectPolicy#IDENTITY_OPEN
      */
    VALUES {
        public CharSequence toString(SequenceValueGenerator<?> values) {
            StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(values);
            sb.append("; size: ").append(values.size());
            sb.append("; number: ").append(values.number());
            if (values instanceof SequenceValueCollection) {
                sb.append("; values: ").append(((SequenceValueCollection<?,?>)values).values);
            }
            return sb.append(']');
        }
    },
    
    /**
      * Generates a char sequence representation of a value generator 
      * having the following format:
      * <pre>
      *   SequenceValueGenerator[hash-code: ..[; identity-hash-code: ..]; type: ..; size: ..; number: ..; ordered: ..; sorted: ..; duplicates: ..] 
      * </pre>
      * 
      * Where {@code SequenceValueGenerator} is the simple class name of
      * the class of the values. <p>
      *
      * @see ObjectPolicy#IDENTITY_OPEN
      */
    VERBOSE {
        public CharSequence toString(SequenceValueGenerator<?> values) {
            StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(values);
            return sb.append("; type: ").append(values.getValueType().getSimpleName())
                     .append("; size: ").append(values.size())
                     .append("; number: ").append(values.number())
                     .append("; ordered: ").append(values.ordered())
                     .append("; sorted: ").append(values.sorted())
                     .append("; duplicates: ").append(values.duplicates())
                     .append(']');
        }
    };
    
    /**
      * @return {@link dk.rode.thesis.strategy.StringablePolicy.Type#APPENDABLE}. 
      */
    public Type type() {
        return Type.APPENDABLE;
    }
        
}
