/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.bridge;

import static dk.rode.thesis.meta.util.Util.serializableCopy;

import java.io.Serializable;
import java.util.TreeSet;

import dk.rode.thesis.meta.Participant;

/**
  * A <i>sequence value tree set</i> stores sequence values 
  * in a {@link java.util.TreeSet java.util.TreeSet} instance. <p>
  * 
  * A sequence value tree set is always {@link #ordered() ordered},
  * {@link #sorted() sorted}, and contains no {@link #duplicates() duplicates}.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of serializable values stored in this value set.
  */
@Participant("ConcreteImplementor")
public class SequenceValueTreeSet<E extends Serializable> 
     extends SequenceValueSet<E,TreeSet<E>> {
    
    /**
      * Constructor, which creates this <i>sequence value tree set</i> to
      * use the {@code java.util.TreeSet} instance supplied as 
      * {@code values}. <p>
      * 
      * Modifying {@code values} externally <b>will</b> affect this
      * value set and should <b>not</b> be performed. The result
      * of doing so is undefined - but never good. <p>
      * 
      * @param  type   The type of values stored in this set (and 
      *                {@code values}); cannot be null.
      * @param  values The actual {@code TreeSet} instance storing the sequence 
      *                values; cannot be null or empty.
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code values} is empty. 
      */
    public SequenceValueTreeSet(Class<E> type, TreeSet<E> values) {
        this(type, values, 0);
    }
    
    /**
      * Constructor, which creates this <i>sequence value tree set</i> to
      * use the {@code java.util.TreeSet} instance supplied as 
      * {@code values}. <p>
      * 
      * Modifying {@code values} externally <b>will</b> affect this
      * value set and should <b>not</b> be performed. The result
      * of doing so is undefined - but never good. <p>
      * 
      * @param  type   The type of values stored in this set (and 
      *                {@code values}); cannot be null.
      * @param  values The actual {@code TreeSet} instance storing the sequence 
      *                values; cannot be null or empty.
      * @param  number The number pointing the the {@code number}'th value
      *                to be returned by the next call to {@link #get()}.                
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code values} is empty, 
      *         contain a null value, or if {@code number} is illegal.
      */
    private SequenceValueTreeSet(Class<E> type, TreeSet<E> values, int number) {
        super(type, values, number);
    }
    
    /**
      * Returns true.
      * 
      * @return True.
      */
    public boolean ordered() {
        return true; 
    }
    
    /**
      * Returns true.
      * 
      * @return True.
      */
    public boolean sorted() {
        return true;
    }
    
    /**
      * 
      */
    public SequenceValueTreeSet<E> copy() {
        return new SequenceValueTreeSet<E>(this.type, 
                                           serializableCopy(this.values), 
                                           this.number);
    }
    
}
