/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.builder;

import dk.rode.thesis.interpreter.Context;
import dk.rode.thesis.interpreter.Expression;
import dk.rode.thesis.interpreter.TypedExpression;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>typed comparable expression builder</i> construct various 
  * {@link TypedExpression} types having a {@code Comparable} bound on 
  * the value produced by the constructed expressions and on the 
  * expression operands used. <p>
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The {@code Comparable} type of values the constructed expressions 
  *             operate on and/or produces.
  */
@Participant("ConcreteBuilder")
public class TypedComparableExpressionBuilder<E extends Comparable<? super E>> 
     extends TypedExpressionBuilder<E> 
  implements ComparableExpressionBuilder<E> {
    
    /**
      * Constructor. <p>
      * 
      * The same {@link Sequence sequence} as used by {@code builder} will be 
      * used by this builder, but a new local {@link Context context} will be
      * created and used.
      *  
      * @param  type    The class literal representing the type of value the 
      *                 evaluation of constructed expressions produces; cannot
      *                 be null.
      * @param  builder The actual builder to use; cannot be null.
      * @throws NullPointerException If either argument is null.
      */
    public TypedComparableExpressionBuilder(Class<E> type, ComparableExpressionBuilder<E> builder) {
        super(type, builder);
    }
   
    /**
      * Copy constructor. <p>
      * 
      * The same {@link Sequence sequence} as used by {@code builder} will be 
      * used by this builder, but a new local {@link Context context} will be
      * created and used. 
      * 
      * The root expression from {@code builder} is <b>not</b> copied.
      * Hence, this builder is ready to construct new expressions, and
      * return a unique root expression from {@link #getRootExpression()}. 
      * 
      * @param  builder The builder to copy; cannot be null.
      * @throws NullPointerException If {@code builder} is null.
      */
    public TypedComparableExpressionBuilder(TypedExpressionBuilder<E> builder) {
        super(builder);
    }

    /**
      * Copy constructor. <p>
      * 
      * The same {@link Context context} and {@link Sequence sequence}
      * as used by {@code builder} will be used by this builder. <p>
      * 
      * The root expression from {@code builder} is <b>not</b> copied.
      * Hence, this builder is ready to construct new expressions, and
      * return a unique root expression from {@link #getRootExpression()}. 
      * 
      * @param  builder The builder to copy; cannot be null.
      * @throws NullPointerException If {@code builder} is null.
      */
    public TypedComparableExpressionBuilder(TypedComparableExpressionBuilder<E> builder) {
        super(builder);
    }
    
    /**
      * 
      */
    public TypedComparableExpressionBuilder<E> copy() {
        return new TypedComparableExpressionBuilder<E>(this);   
    }

    /**
      * 
      */
    public TypedExpression<Boolean> buildSmallerThanExpression(Expression<E> first, Expression<E> second) {
        return getTypedExpression(((ComparableExpressionBuilder<E>)this.builder).buildSmallerThanExpression(first, second), Boolean.class);
    }
    
    /**
      * 
      */
    public TypedExpression<Boolean> buildGreaterThanExpression(Expression<E> first, Expression<E> second) {
        return getTypedExpression(((ComparableExpressionBuilder<E>)this.builder).buildGreaterThanExpression(first, second), Boolean.class);
    }

    /**
      * 
      */
    public TypedExpression<E> buildSetExpression(Expression<E> value) {
        return getTypedExpression(((ComparableExpressionBuilder<E>)this.builder).buildSetExpression(value), this.type);
    }
           
}
