/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.command;

import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.prototype.StrictCopyable;

/**
  * A <i>command</i> represents a possibly {@link #isUndoable() undoable}
  * {@link #execute() operation} on a given <i>receiver</i>. <p>
  * 
  * The result of the execution is called the <i>command value</i> and 
  * can be acquired using the {@link #getResult()} method <b>after</b>
  * execution. Commands should <b>always</b> be executed by a 
  * {@link CommandProcessor} to ensure proper undo functionality. <p>
  * 
  * When executed, commands may <i>spawn</i> new commands to be executed
  * immediately, i.e. <b>before</b> other commands in the current
  * list of commands being {@link CommandProcessor#execute(List) executed}
  * by the command {@link CommandProcessor processor} used. The execution
  * of spawned commands is thus depth-first. <p>   
  * 
  * Commands are <b>not</b> thread-safe, and generally <b>cannot</b> be 
  * shared. Instead, they can be {@link #copy() copied} for "reuse".
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    CommandProcessor
  * @see    CommandProcessingResult
  * 
  * @param  <E> The type of result obtained by executing this command. 
  */
@Participant("Command")
public interface Command<E> extends StrictCopyable<Command<E>> {

    /**
      * Performs the execution of this command. <p>
      * 
      * After successful execution, the result of the
      * execution can be obtained via the {@link #getResult()}
      * method. <p>
      * 
      * @return A list of spawned commands, if any, to be executed by the
      *         command {@link CommandProcessor processor} currently
      *         executing this command. Can be null or empty.
      * @throws CommandException If the execution fails.
      * @see    #undo() 
      */
    public List<Command<E>> execute() throws CommandException;
    
    /**
      * Perform undo of this command, if possible. <p>
      * 
      * If this method is invoked and this command is not
      * {@link #isUndoable() undoable}, an exception will 
      * be thrown. <p>
      * 
      * The result of invoking undo on a command that has
      * not been {@link #execute() executed} is undefined.
      *  
      * @return The result of the undo operation; may be null.
      * @throws CommandException If the undo fails.
      */
    public E undo() throws CommandException;

    /**
      * Returns true if this command has been executed and is 
      * undoable, false if not. <p>
      * 
      * If undoable, the effect of the {@link #execute() execution}
      * can be undone by invoking {@link #undo()}.
      *   
      * @return True if undoable, false if not.
      */
    public boolean isUndoable();

    /**
      * Returns the result of the {@link #execute() execution} of this
      * command. <p>
      * 
      * Invoking this method before execution will cause an illegal
      * state exception to be thrown. If {@link #undo() undone}, 
      * this command must be re-executed before the result can be
      * fetched again.
      *  
      * @return The result of the execution; can be null.
      */
    public E getResult();
    
    /**
      * Creates a copy of this command, excluding internal state
      * information required for execution and undo.
      *  
      * @return A new instance of this command; never null.
      */
    public Command<E> copy();
    
}
