/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.command;

import java.util.List;

import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>command processing result</i> represents the result of
  * {@link CommandProcessor#execute(List) executing}
  * a list commands by a given command {@link CommandProcessor 
  * processor}. <p>  
  * 
  * The result can represent two scenarios: <ol><p>
  * 
  * <li> The execution succeeded: {@link #executed} is true, and
  *      {@link #result} stores the value, which may be null. The
  *      result is the command {@link Command#getResult() value}
  *      fetched from the <b>last</b> - possibly spawned - command 
  *      executed. <p> 
  * 
  * <li> The execution failed, but was able to be completely undone:
  *      {@link #executed} is false, and {@link #result} is always
  *      null.
  * 
  * </ol>
  *
  * In case the execution and undo fails completely, the processor
  * will throw an exception, and not return a result.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    CommandProcessor#execute(List)
  * @see    Command#getResult()
  * 
  * @param  <E> The type of result obtained from the executed commands. 
  */
public class CommandProcessingResult<E> {

    /**
      * True if the commands that generated this result were successfully
      * executed, and hence the value of {@link #result} is known to be valid. <p>
      * 
      * If false, the execution failed, but all commands were successfully
      * undone. The value of {@link #result} will always be null in such
      * cases. 
      */
    public final boolean executed; 
    /**
      * The result of the command execution. <p>
      * 
      * The value is the result {@link Command#getResult() value} from
      * the <b>last</b> - possibly spawned - command executed. <p>
      * 
      * Can be null, even if {@link #executed} is true.
      */
    public final E result;
    
    /**
      * Constructor.
      * 
      * @param executed The executed flag.
      * @param result   The result of the execution; can be null.
      */
    CommandProcessingResult(boolean executed, E result) {
        this.executed = executed;
        this.result = result;
    }
    
    /**
      * Returns the string representation of this result.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.NAME_OPEN.toString(this);
        return sb.append("executed: ").append(this.executed)
                 .append("result: ").append(this.result).append(']').toString();
    }
}
