/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.command;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>composite command</i> is a command that represents
  * an number of other contained commands, possibly composite
  * as well, thus forming a tree-like structure. <p>
  * 
  * Composite commands are simple containers that do not maintain
  * parent references. Contained commands do not know they are
  * contained unless this information is stored locally. Hence,
  * when a composite command is {@link #execute() executed}, 
  * graph traversal is top-down only, in a depth-first manner; the
  * result of the last executed command is returned. <p>
  * 
  * A composite command is {@link #isUndoable() undoable} if all
  * contained commands are undoable. If the execution fails
  * for the <i>n</i>'th command, the commands from <i>n</i> to 
  * zero will be undone when {@link #undo()} is invoked, i.e. in 
  * reverse order. The returned results is the result from the
  * last undone command, i.e. the command at index zero (which
  * may be composite as well). <p> 
  * 
  * The value returned from {@code execute()} and {@code undo()} 
  * is the result from the <i>last</i> command in this composite 
  * sequence! For {@code execute}, this is the last command executed, 
  * while for {@code undo} is it the value of the first undone command. <p> 
  * 
  * The result of executing a failed composite command again without 
  * having undo the previous error is undefined. <p>
  * 
  * Composite commands are compared based on identity (==). <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  *  
  * @param  <E> The type of result obtained by executing all contained
  *             commands in this composite command. 
  */
@Participant("ConcreteCommand")
public class CompositeCommand<E> implements Command<E>, Iterable<Command<E>> {

    /**
      * The commands associated with this command, in order. <p>
      * 
      * Never null, but can be null.
      */
    private final List<Command<E>> commands;
    /**
      * The index identifying the current command being 
      * executed or undone. <p>
      * 
      * Range: 0 <= index <= {@link #commands}.size()
      */
    private int index;
    
    /**
      * No-arg constructor.
      */
    public CompositeCommand() {
        this.commands = new ArrayList<Command<E>>();
        this.index = -1;
    }

    /**
      * Copy constructor, which {@link #copy() copies} each 
      * associated command in {@code command} and adds it
      * to this composite command.
      * 
      * @param  command The composite command to copy; cannot be null.
      * @throws NullPointerException If {@code command} is null.
      */
    public CompositeCommand(CompositeCommand<E> command) {
        this();
        for (Command<E> c : command) {
            this.addCommand(c.copy());
        }
    }
    
    /**
      * Constructor, which adds each {@link #copy() copies} each command
      * in {@code commands} and adds it to this composite command. <p>
      * 
      * Duplicate commands are ignored.
      * 
      * @param  commands The (copied) commands to add; cannot contain null 
      *                  entries.
      * @throws NullPointerException If {@code commands} contain a null
      *         entry.
      */
    public CompositeCommand(Command<E>... commands) {
        this();
        for (Command<E> c : commands) {
            this.addCommand(c.copy());
        }
    }
    
    /**
      * 
      */
    public CompositeCommand<E> copy() {
        return new CompositeCommand<E>(this);
    }
    
    /**
      * Associates the {@link Command command} supplied
      * as {@code command} to this composite command, if
      * not already. <p>
      * 
      * If already added, this method does nothing.  
      * 
      * @param  command The command to add; cannot be null.
      * @return True if {@code command} was added, false if not, 
      *         i.e. if already added.
      * @throws NullPointerException If {@code command} is null.
      * @throws IllegalArgumentException If this composite command is
      *         tried added to it self.  
      * @see    #removeCommand(Command)         
      */
    public boolean addCommand(Command<E> command) {
        if (command == null) {
            throw new NullPointerException("Command cannot be null");
        } else if (this.commands.contains(command)) {
            return false;
        } else if (command == this) {
            throw new IllegalArgumentException("Cannot add self");
        }
        // Returns true...
        return this.commands.add(command);
    }
    
    /**
      * Removes the {@link Command command} supplied
      * as {@code command} from this composite command, if
      * associated. <p>
      * 
      * If not associated, this method does nothing.  
      * 
      * @param  command The command to remove; cannot be null.
      * @return True if {@code command} was removed, false if not, 
      *         i.e. if not previously added.
      * @throws NullPointerException If {@code command} is null.
      * @see    #addCommand(Command)         
      */
    public boolean removeCommand(Command<?> command) {
        if (command == null) {
            throw new NullPointerException("Command cannot be null");
        }
        return this.commands.remove(command);
    }

    /**
      * Returns an iterator to iterate through this composite
      * command, which is backed by this composite commands, i.e.
      * changes to the iterator changes this command.
      * 
      * @return An iterator backed by this composite command;
      *         never null.
      */
    public Iterator<Command<E>> iterator() {
        // Backed...
        return this.commands.iterator();
    }
    
    /**
      * Returns the {@link Command commands} currently
      * associated with this composite command, in order. <p>
      * 
      * Modifying the returned collection will not affect 
      * this command.
      *  
      * @return The collection of associated commands; never 
      *         null, but can be empty.
      * @see    #size()         
      */
    public List<Command<E>> getCommands() {
        return new ArrayList<Command<E>>(this.commands);
    }
    
    /**
      * Returns the {@link Command commands} currently
      * associated with this composite command, in order,
      * assignable to the type supplied as {@code type}. <p>
      * 
      * Invoking this method with the {@code Command} interface
      * corresponds to invoking {@link #getCommands()}. <p>
      * 
      * Modifying the returned collection will not affect 
      * this command.
      *  
      * @param  <V>  The type of applicable commands.
      * @param  type The (super-)class of commands to include; cannot 
      *              be null.
      * @return The collection of associated commands of the type
      *         supplied as {@code type}; never null, but can be 
      *         empty.
      * @throws NullPointerException If {@code type} is null.
      */
    public <V extends Command<E>> List<V> getCommands(Class<V> type) {
        List<V> commands = new ArrayList<V>(this.commands.size());
        for (Command<E> s : this) {
            if (type.isInstance(s)) {
                commands.add(type.cast(s));
            }
        }
        return commands;
    }

    /**
      * Returns the number of associated commands to this
      * composite command.
      *  
      * @return The number of commands.
      */
    public int size() {
        return this.commands.size();
    }

    //
    // Command interface
    //
    
    /**
      * @return A list containing all spawned commands from contained
      *         commands; never null, but can be empty.
      */
    public List<Command<E>> execute() throws CommandException {
        List<Command<E>> commands = new ArrayList<Command<E>>(this.commands.size());
        for (this.index = 0; this.index < this.commands.size(); this.index++) {
            Command<E> command = this.commands.get(this.index);
            LogFactory.getLog(this).println("Executing: ", command);
            List<Command<E>> spawned = command.execute();
            if (spawned != null) {
                commands.addAll(spawned);
            }
        }
        return commands;
    }

    /**
      * 
      */
    public boolean isUndoable() {
        for (Command<E> c : this) {
            // Will recurse if "c" is composite... 
            if (!c.isUndoable()) {
                return false;
            }
        }
        return true;
    }

    /**
      * 
      */
    public E undo() throws CommandException {
        if (!this.isUndoable()) {
            throw new CommandException("Command is not undoable");
        }
        // We return the *last* value to be compliant with 
        // "execute", which is the first command undone because
        // we traverse the list backwards...
        E value = null;
        for (this.index-- ; this.index >= 0; this.index--) {
            Command<E> command = this.commands.get(this.index);
            LogFactory.getLog(this).println("Undoing: ", command);
            E v = command.undo();
            if (value == null) {
                value = v;
            }
        }
        return value;
    }

    /**
      * Returns the result of the last executed command contained
      * in this composite command. 
      */
    public E getResult() {
        int size = this.commands.size();
        return size == 0 ? null : this.commands.get(size - 1).getResult();
    }
    
    /**
      * Returns the string representation if this composite command.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.ID_OPEN.toString(this);
        return sb.append("; size: ").append(this.commands.size()).append(']').toString();
    }
    
}
