/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.command;

import java.util.List;

import dk.rode.thesis.memento.MemorizableSequence;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.ReversibleSequence;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>next command</i> command will invoke {@link Sequence#next() 
  * next()} on its receiving {@link Sequence sequence} when it is 
  * {@link #execute() executed}. <p>
  * 
  * Next commands are {@link #isUndoable() undoable} if the receiving 
  * sequence is a {@link ReversibleSequence reversible} and/or a 
  * {@link MemorizableSequence memorizable} sequence!
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of result obtained by executing this command.
  *             The type corresponds to the type of values delivered by
  *             the receiver sequence. 
  */
@Participant("ConcreteCommand")
public class NextCommand<E> extends SequenceCommand<E> {
    
    /**
      * Constructor.
      * 
      * @param  sequence The sequence; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public NextCommand(@Participant("Receiver") Sequence<E> sequence) {
        super(sequence);
    }

    /**
      * 
      */
    public NextCommand<E> copy() {
        return new NextCommand<E>(this.sequence);
    }
    
    /**
      * Sets the result of the execution to the {@link Sequence#next() next} 
      * sequence value.
      */
    protected List<Command<E>> doExecute(boolean saved) {
        this.result = this.sequence.next();
        return null;
    }

    /**
      * Returns true if this command has been executed and the receiving 
      * sequence used by this command is either {@link MemorizableSequence 
      * memorizable} and/or {@link ReversibleSequence#reversible() 
      * reversible}.
      * 
      * @return True if undoable, false if not.
      */
    public boolean isUndoable() {
        if (super.isUndoable()) {
            return true;
            
        } else if (this.sequence instanceof ReversibleSequence) {
            ReversibleSequence<E> rs = (ReversibleSequence<E>)this.sequence;
            return rs.reversible();
        }
        return false;
    }

    /**
      *  
      */
    public E undo() throws CommandException {
        // Try undo using super class, i.e. mementos...
        E value = super.undo();
        if (value != null) {
            return value;
        }
        LogFactory.getLog(this).println("Undoing reversible sequence: ", this.sequence);
        // Cast known to succeed since we are undoable, hence 
        // a reversible sequence...
        ReversibleSequence<E> rs = (ReversibleSequence<E>)this.sequence;
        // Reverse sequence...
        rs.reverse();
        // Go back to the previous value...
        rs.next();
        // And reverse so we will advance in the proper direction again...
        return rs.reverse();
    }
    
}
