/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.command;

import static dk.rode.thesis.meta.util.Util.add;

import java.util.ArrayList;
import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.ReversibleSequence;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>reserve command</i> command will {@link ReversibleSequence#reverse()
  * reverse} a given sequence when it is {@link #execute() executed}, if 
  * possible. <p>
  * 
  * Only currently {@link ReversibleSequence#reversible() reversible} 
  * sequences are affected by this command per default. However, a list of 
  * commands can be supplied at construction time to be spawned during
  * execution for non-reversible sequences. <p>
  * 
  * Undo of spawned commands is <b>not</b> handled by this
  * command. <p>
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of result obtained by executing this command.
  *             The type corresponds to the type of values delivered by
  *             the receiver sequence. 
  */
@Participant("ConcreteCommand")
public class ReverseCommand<E> extends SequenceCommand<E> {
    
    /**
      * True if this command successfully reversed a reversible
      * sequence, false if not. <p>
      * 
      * Will be null if this command has not been executed.
      */
    private Boolean reversed;
    /**
      * Command <i>instances</i> to spawn in case the sequence 
      * is not reversible, if any. <p>
      * 
      * Undo of spawned commands is <b>not</b> handled by this
      * command. <p>
      * 
      * Can be null.
      */
    private final List<Command<E>> commands;
    
    /**
      * Constructor.
      * 
      * @param  sequence The sequence; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public ReverseCommand(@Participant("Receiver") Sequence<E> sequence) {
        super(sequence);
        this.commands = new ArrayList<Command<E>>(0);
    }

    /**
      * Constructor. <p>
      * 
      * The commands supplied as {@code commands}, if any, will be 
      * spawned in order during {@link #execute() execution} in case
      * {@code sequence} cannot be reversed. 
      * 
      * @param  sequence The sequence; cannot be null.
      * @param  commands  The commands to possible spawn; can be null or empty.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public ReverseCommand(@Participant("Receiver") Sequence<E> sequence, Command<E>... commands) {
        super(sequence);
        this.commands = new ArrayList<Command<E>>(commands == null ? 0 : commands.length);
        add(this.commands, commands);
    }

    /**
      * Copy constructor. <p>
      * 
      * Commands to spawned, if any, will also be copied. 
      * 
      * @param  command The command to copy; cannot be null.
      * @throws NullPointerException If {@code command} is null.
      */
    public ReverseCommand(ReverseCommand<E> command) {
        super(command.sequence);
        this.commands = new ArrayList<Command<E>>(command.commands.size());
        for (Command<E> c : command.commands) {
            this.commands.add(c.copy());
        }
    }
    
    /**
      * 
      */
    public ReverseCommand<E> copy() {
        return new ReverseCommand<E>(this);
    }
    
    /**
      * Reverses the sequence if {@link ReversibleSequence#reversible() 
      * reversible} and returns the current (same) value. <p>
      * 
      * If the sequence is not reversible, the commands supplied
      * at construction time will be spawned, if any. <p>
      * 
      * @return If the sequence was not reversible, a list containing 
      *         the commands supplied at construction time, if any;
      *         never null, but can be empty.
      */
    protected List<Command<E>> doExecute(boolean saved) {
        // List of spawned commands...
        List<Command<E>> spawned = new ArrayList<Command<E>>(1);
        // Do NOT invoke super execute because we cannot use mementos...
        if (this.sequence instanceof ReversibleSequence) {
            ReversibleSequence<E> rs = (ReversibleSequence<E>)this.sequence;
            if (rs.reversible()) {
                rs.reverse();
                this.reversed = Boolean.TRUE;
                LogFactory.getLog(this).println("Reversed: ", this.sequence);
            } else {
                LogFactory.getLog(this).println("Reversible sequence cannot be reversed: ", this.sequence);
                this.reversed = Boolean.FALSE;
                // Spawn, if any...
                spawned.addAll(this.commands);
            }
        }
        this.result = this.sequence.current();
        return spawned;
    }

    /**
      * Returns true if the receiving sequence is 
      * {@link ReversibleSequence#reversible() reversible},
      * false if not.
      * 
      * @return True if undoable, false if not.
      */
    public boolean isUndoable() {
        // No need to invoke super...
        if (this.sequence instanceof ReversibleSequence) {
            // If null, this command is not executed, hence not undoable...
            if (this.reversed == null) {
                return false;
            }
            ReversibleSequence<E> rs = (ReversibleSequence<E>)this.sequence;
            if (Boolean.TRUE.equals(this.reversed)) {
                return rs.reversible();
            }
            // If we did not reverse during execution this command
            // is undoable because we did not do anything in this
            // commands - though we might have spawned commands, but
            // they will be handled independently. Hence, undo will 
            // nothing as well...
        }
        // Not reversed...
        return true;
    }

    /**
      *  
      */
    public E undo() throws CommandException {
        // If we did not reverse, we simply return the current value...
        if (Boolean.FALSE.equals(this.reversed)) {
            return this.sequence.current();
            
        // Otherwise use standard semantics...
        } else if (!this.isUndoable()) {
            throw new CommandException("Command is not undoable");
        }
        // Reverse (again)...
        this.doExecute(true);
        this.reversed = null;
        return this.result;
    }
    
}
