/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.command;

import java.util.List;

import dk.rode.thesis.memento.MemorizableSequence;
import dk.rode.thesis.memento.SequenceMemento;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>sequence command</i> is a command that manipulates
  * the state of a {@link Sequence sequence} when it
  * is {@link Command#execute() executed}. The sequence in
  * question is the <i>receiver</i>. <p>
  * 
  * Sequence commands are <b>not</b> thread-safe and 
  * <b>cannot</b> be shared, as they carry information required
  * for a specific execution. However, sequence commands supply
  * a {@link #copy() copy} method to return a new command of 
  * the same type using the same sequence as is. Any state
  * related to the execution is not copied. <p>
  * 
  * By default, two sequence commands having the exact same
  * class <b>and</b> using the same sequence are considered 
  * equal. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * A simple version of the {@linkplain dk.rode.thesis.templatemethod 
  * Template Method} pattern is applied here [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.325]: the 
  * (final) {@link #execute()} method is the template method that 
  * utilises the <i>primitive operation</i> {@link #doExecute(boolean)}, 
  * which sub-classes must implement. If the sequence in question is a
  * {@link MemorizableSequence memorizable} sequence, this command is 
  * able to be undone using the default {@link #isUndoable()} and 
  * {@link #undo()} implementations as {@code execute()} saves the 
  * <i>before state</i>. Hence, {@code isUndoable()} and {@code undo()} 
  * are <i>hook operations</i> that can be overridden by sub-commands 
  * if specific control of the undo mechanism is required, as for example 
  * done in the {@link ResetCommand} class. 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values returned by this command, i.e.
  *             the type of values delivered by the receiving sequence. 
  */
@Participant("Command")
public abstract class SequenceCommand<E> implements Command<E> {
    
    /**
      * The sequence receiver used by this command. <p>
      * 
      * Never null.
      */
    @Participant("Receiver")
    protected final Sequence<E> sequence;
    /**
      * The result of the execution. <p>
      * 
      * Can be null.
      */
    protected E result;
    /**
      * A sequence memento to restore the previous state
      * if {@link #sequence} is a {@link MemorizableSequence
      * memorizable} sequence. <p>
      * 
      * Will only be set if {@link #execute()} is invoked by
      * sub-classes <b>before</b> their actual execution. <p>
      * 
      * Can be null.
      */
    private SequenceMemento<E> memento;
    
    /**
      * Constructor, which creates this sequence command
      * to use the {@link Sequence} supplied as 
      * {@code sequence} as the receiver.
      * 
      * @param  sequence The sequence; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    protected SequenceCommand(@Participant("Receiver") Sequence<E> sequence) {
        if ((this.sequence = sequence) == null) {
            throw new NullPointerException("Sequence cannot be null");
        }
    }

    /**
      * Saves the <i>before state</i> of this sequence command if the
      * receiver sequence is a {@link MemorizableSequence memorizable}
      * sequence before {@link #doExecute(boolean) execution}. 
      * 
      * @see #doExecute(boolean)
      */
    public final List<Command<E>> execute() throws CommandException {
        boolean saved = false;
        if (this.sequence instanceof MemorizableSequence) {
            MemorizableSequence<E> ms = (MemorizableSequence<E>)this.sequence;
            this.memento = ms.getMemorizableState();
            saved = true;
        }
        return this.doExecute(saved);
    }

    /**
      * Performs the actual execution of this command. <p>
      * 
      * After successful execution, the result of the
      * execution can be obtained via the {@link #getResult()}
      * method. <p>
      * 
      * @param  saved True if the before state was successfully saved before
      *               this method is invoked, false if not.
      * @return A list of spawned commands, if any, to be executed by the
      *         command {@link CommandProcessor processor} currently
      *         executing this command. Can be null or empty.
      * @throws CommandException If the execution fails.
      * @see    #execute()
      */
    protected abstract List<Command<E>> doExecute(boolean saved) throws CommandException;
    
    /**
      * Performs undo of this sequence command if possible. <p>
      * 
      * If this command is not {@link #isUndoable() undoable} an 
      * exception will be thrown. If this command cannot be undone
      * otherwise, null will be returned. <p>
      * 
      * The command result value is cleared.
      * 
      * @return If this command could be undone, the (new) current 
      *         sequence value is returned; otherwise null.
      * @throws CommandException In case this command is not undoable.         
      */
    public E undo() throws CommandException  {
        if (!this.isUndoable()) {
            throw new CommandException("Command not undoable");
            
        } if ((this.sequence instanceof MemorizableSequence) && (this.memento != null)) {
            LogFactory.getLog(this).println("Undoing memorizable sequence: ", this.sequence);
            try {
                MemorizableSequence<E> ms = (MemorizableSequence<E>)this.sequence;
                ms.setMemorizableState(this.memento);
                // Clear memento...
                this.memento = null;
                // Clear result...
                this.result = null;
                return this.sequence.current();
                
            } catch (Exception e) {
                throw new CommandException(e); 
            }
        }
        return this.result = null;
    }

    /**
      * Returns true if this sequence command has been executed and 
      * can be undone by this sequence command, false if not.
      *  
      * @return Returns true of the receiving sequence is a 
      *         {@link MemorizableSequence memorizable}
      *         sequence that has been executed, false if not.
      */
    public boolean isUndoable() {
        return ((this.sequence instanceof MemorizableSequence) && 
                (this.memento != null));
    }
    
    /**
      * 
      */
    public E getResult() {
        return this.result;
    }
    
    /**
      * Returns the hash code of this sequence command. 
      *
      * @return The hash code.
      */
    public int hashCode() {
        return this.sequence.hashCode();
    }
    
    /**
      * Returns true if {@code object} has the same class
      * as this sequence command <b>and</b> uses the same
      * sequence receiver, false if not. 
      *
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (object == null) {
            return false;
        } else if (object == this) {
            return true;
        } else if (this.getClass().equals(object.getClass())) {
            return this.sequence.equals(((SequenceCommand<?>)object).sequence);
        }
        return false;
    }
    
    /**
      * Returns the string representation of this sequence command.
      *  
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; sequence: ").append(this.sequence.toString(SequencePolicy.IDENTITY))
                 .append(']').toString();
    }
    
}
