/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.decorator;

import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.EnglishAlphabetSequence;
import dk.rode.thesis.meta.model.LongSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.singleton.DanishAlphabetSequence;

/**
  * <b>Decorator</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Main implements Testable {

    /**
      * Executes the <b>Decorator</b> tests.
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }
    
    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) {
        out.heading("Decorator tests...");
       
        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }
        
        // 
        // Create sequences and decorators
        //
        
        Sequence<?>[] sequences = {
            // Decorate sequences...
            new AppenderDecorator(new UppercaseDecorator(new EnglishAlphabetSequence()), "x", "x"),
            new UppercaseDecorator(new AppenderDecorator(new EnglishAlphabetSequence(), "z", "z")), 
            // Cannot explicitly invoke the constructor of the Danish alphabet
            // sequence class because it is a singleton...
            new UppercaseDecorator(DanishAlphabetSequence.Instance),  
            new DuplexDecorator(DanishAlphabetSequence.Instance), 
            // Different type of sequence...
            new LongSequence()
        };
        
        // Print instances...
        out.heading("Sequence instances");
        for (int i = 0; i < sequences.length; i++) {
            out.println("Sequence ", i, ": ", sequences[i]);
        }
        
        //
        // Test equality
        //
        
        // Equality (do not need to test both ways because of the standard
        // Java hashCode/equals contract...
        out.heading("Sequence equality");
        for (int i = 0; i < sequences.length; i++) {
            Sequence<?> sequence = sequences[i];
            out.println(i, ": ", sequence, " equals...");
            for (int j = i; j < sequences.length; j++) {
                out.println("--> ", j, ": ", sequences[j], ": ", sequence.equals(sequences[j]));
            }
        }
        
        //
        // Usage
        //
        
        out.heading("Sequence usage");
        for (int i = 0; i < 30; i++) {
            out.println("Iteration: ", i + 1);
            for (int j = 0; j < sequences.length; j++) {
                Sequence<?> sequence = sequences[j];
                // Advance sequence...
                Object value = sequence.next();
                out.println(j, ": ", sequence, " -> ", value);
            }
            out.println();
        }
        
        out.heading("Sequence equality (copies)");
        for (int i = 0; i < sequences.length; i++) {
            Sequence<?> s = sequences[i].copy();
            out.println(i, ": ", s, " equals...");
            for (int j = i; j < sequences.length; j++) {
                Sequence<?> sequence = sequences[j].copy();
                out.println("--> ", j, ": ", sequence, ": ", s.equals(sequence));
            }
        }
        
        out.heading("Decorator tests succeeded!");
        
        return true;
    }
    
}
