/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.factorymethod;

import static dk.rode.thesis.meta.model.Sequence.State.NORMAL;
import static dk.rode.thesis.meta.model.Sequence.State.RESET;
import static dk.rode.thesis.meta.model.Sequence.State.RESTART;
import static dk.rode.thesis.meta.model.Sequence.State.START;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import dk.rode.thesis.command.Command;
import dk.rode.thesis.command.CommandException;
import dk.rode.thesis.command.CommandProcessingResult;
import dk.rode.thesis.command.CommandProcessor;
import dk.rode.thesis.command.EvilCommand;
import dk.rode.thesis.command.LogCommand;
import dk.rode.thesis.command.NextCommand;
import dk.rode.thesis.command.NullCommand;
import dk.rode.thesis.command.ResetCommand;
import dk.rode.thesis.command.ReverseCommand;
import dk.rode.thesis.facade.RandomSequence;
import dk.rode.thesis.memento.RangeSequence;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.model.Sequence.State;
import dk.rode.thesis.meta.reflect.InstantiableTypeLiteral;
import dk.rode.thesis.meta.reflect.TypeLiteral;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;

/**
  * <b>Factory Method</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Main implements Testable {

    /**
      * Executes the <b>Factory Method</b> tests.
      * 
      * @param  args The arguments to supply to {@link #test(Log, Arguments)},
      *              if any; can be null.
      * @throws Exception If the test failed unexpectedly.             
      */
    public static void main(String[] args) throws Exception {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }

    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(final Log out, Arguments arguments) throws Exception {
        out.heading("Factory Method tests...");
        
        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }

        boolean ok = true;
        
        // True to enable class and/or verbose logging...
        boolean logCommands  = arguments.getArgumentValue("log",         true,  Boolean.class),
                logExecution = arguments.getArgumentValue("log.verbose", false, Boolean.class),
                logCreators  = arguments.getArgumentValue("log",         true,  Boolean.class);
        
        LogFactory.getLog(NextCommand.class).setActive(logCommands);
        LogFactory.getLog(ResetCommand.class).setActive(logCommands);
        LogFactory.getLog(ReverseCommand.class).setActive(logCommands);
        LogFactory.getLog(LogCommand.class).setActive(logCommands);
        
        LogFactory.getLog(CommandProcessor.class).setActive(logExecution);
        
        LogFactory.getLog(CommandCreator.class).setActive(logCreators);
        LogFactory.getLog(SequenceCommandCreator.class).setActive(logCreators);
        LogFactory.getLog(ReversibleSequenceCommandCreator.class).setActive(logCreators);
        LogFactory.getLog(EvilSequenceCommandCreator.class).setActive(logCreators);
        LogFactory.getLog(ReflectiveCommandCreator.class).setActive(logCreators);
        // Generic factories...
        LogFactory.getLog(TypeLiteral.class).setActive(true);
        LogFactory.getLog(Factory.class).setActive(false);
        LogFactory.getLog(TypedFactory.class).setActive(false);
        
        //
        // Create sequences
        //
        
        List<Sequence<Integer>> sequences = 
            new ArrayList<Sequence<Integer>>(4);
        // To test the command creation...
        sequences.add(null);
        // bounded, not unique, not consistent
        sequences.add(new RandomSequence(10));
        // unbounded, consistent, unique, reversible, memorizable
        sequences.add(new RangeSequence());        
        // bounded, consistent, unique, reversible, memorizable             
        sequences.add(new RangeSequence(10, 20));  

        //
        // Creators based on sequence states
        //
        
        List<CommandCreator<Integer,State>> stateCreators = 
            new ArrayList<CommandCreator<Integer,State>>(4); 
        // Standard
        stateCreators.add(new SequenceCommandCreator<Integer>());
        // Handles reversible as well...
        stateCreators.add(new ReversibleSequenceCommandCreator<Integer>());
        // For testing...
        stateCreators.add(new EvilSequenceCommandCreator<Integer>());

        //
        // Creator based on command type literals (reflection)
        //
        
        CommandCreator<Integer,TypeLiteral<? extends Command<Integer>>> reflectionCreator = 
            new ReflectiveCommandCreator<Integer>();
                
        // Print instances...
        out.heading("Created sequences and creators");
        for (int i = 0; i < sequences.size(); i++) {
            out.println("Sequence ", i, ": ", sequences.get(i));
        }
        out.println();
        for (int i = 0; i < stateCreators.size(); i++) {
            out.println("Creator ", i, ": ", stateCreators.get(i));
        }
        out.println("Creator ", stateCreators.size(), ": ", reflectionCreator)
           .println()
           .println();
        
        //
        // Using creators based on sequence state
        //
        
        out.heading("Using creators based on sequence states");

        // States identifying the types of commands to create... 
        List<State> states = Arrays.asList(new State[]{
            NORMAL, RESET, NORMAL, null, START, NORMAL, RESTART, RESET, null, RESTART 
        });
        
        out.println("Using the following sequence states: ");
        for (State state : states) {
            out.print(state, " ");
        }
        out.println();
        
        for (int i = 0; i < stateCreators.size(); i++) {
            CommandCreator<Integer,State> creator = stateCreators.get(i);
            out.println();
            out.println("Creator ", i, ": ", creator);
            for (int j = 0; j < sequences.size(); j++) {
                Sequence<Integer> sequence = sequences.get(j);
                ok = testCreator(out, creator, sequence, j, states) && ok;
            }
            out.println();
        }

        //
        // Using creator based on command classes (reflection)
        //
        
        // Type literals identifying the types of commands to create. 
        // We only need to make a single literal for each command...
        List<TypeLiteral<? extends Command<Integer>>> literals = 
            new ArrayList<TypeLiteral<? extends Command<Integer>>>(10);
        
        // Next command...
        TypeLiteral<NextCommand<Integer>> next 
            = new TypeLiteral<NextCommand<Integer>>(){/*magic*/};
        // Restart command...
        TypeLiteral<ReverseCommand<Integer>> reverse
            = new TypeLiteral<ReverseCommand<Integer>>(){/*magic*/};
        // Reset command, already declared to use the constructor accepting
        // a Sequence argument...
        TypeLiteral<ResetCommand<Integer>> reset 
            = new InstantiableTypeLiteral<ResetCommand<Integer>>(Sequence.class){/*magic*/};
        // Evil command utilises the no-arg constructor, already declared here...
        TypeLiteral<EvilCommand<Integer>> evil 
            = new InstantiableTypeLiteral<EvilCommand<Integer>>(){/*magic*/};
        // Null command utilises the no-arg constructor, already declared here...
        TypeLiteral<NullCommand<Integer>> nil 
            = new InstantiableTypeLiteral<NullCommand<Integer>>(){/*magic*/};

        // Log command: will fail because there is no constructor accepting
        // a Sequence object!
        TypeLiteral<LogCommand<Integer>> logger
            = new TypeLiteral<LogCommand<Integer>>(){/*magic*/};
            
        // Add so they roughly correspond to the states just tested above...
        literals.add(next);
        literals.add(reset);
        literals.add(logger); // will fail and create default command!
        literals.add(next);
        literals.add(null);
        literals.add(evil);
        literals.add(next);
        literals.add(reverse);
        literals.add(reset);
        literals.add(nil); // null command
        literals.add(reverse);
        
        out.heading("Using creator based on command type literals");
        
        out.println("Using the following command type literals: ");
        for (TypeLiteral<?> literal : literals) {
            out.println(literal);
        }
        out.println()        
           .println("Creator 3: ", reflectionCreator);
        for (int j = 0; j < sequences.size(); j++) {
            Sequence<Integer> sequence = sequences.get(j);
            ok = testCreator(out, reflectionCreator, sequence, j, literals) && ok;
        }
        
        //
        // Using generic creator directly
        //
        
        out.heading("Using generic factories directly");
        out.println("(this is essentially what is being done in the ReflectiveCommandCreator class)").println();
        
        // Represents a generic type...
        TypeLiteral<Sequence<Integer>> type = new TypeLiteral<Sequence<Integer>>(){/*magic*/};
        
        // Next command factory...
        TypedFactory<NextCommand<Integer>,Sequence<Integer>> nextFactory 
            = new TypedFactory<NextCommand<Integer>,Sequence<Integer>>(type){/*magic*/};
        // Restart command factory...
        Factory<ReverseCommand<Integer>> reverseFactory
            = new Factory<ReverseCommand<Integer>>(Sequence.class){/*magic*/};
        // Reset command factory...
        Factory<ResetCommand<Integer>> resetFactory 
            = new Factory<ResetCommand<Integer>>(Sequence.class){/*magic*/};

        List<Factory<? extends Command<Integer>>> factories = new ArrayList<Factory<? extends Command<Integer>>>();
        factories.add(nextFactory);
        factories.add(nextFactory);
        factories.add(reverseFactory);
        factories.add(nextFactory);
        factories.add(resetFactory);
        factories.add(nextFactory);
        factories.add(reverseFactory);
        factories.add(nextFactory);
        factories.add(resetFactory);
        factories.add(nextFactory);
        
        for (int j = 0; j < sequences.size(); j++) {
            Sequence<Integer> sequence = sequences.get(j);
            ok = testFactory(out, sequence, j, factories) && ok;
        }
        
        out.println()
           .heading("Factory Method tests ", ok ? "succeeded!" : "failed!");
        
        return ok;
    }

    /**
      * Tests the creator supplied as {@code creator} using {@code sequence}
      * for each token in {@code tokens}
      *  
      * @param  <E>      The type of sequence and command values.
      * @param  <T>      The type of tokens used to identify the type of
      *                  command to create. 
      * @param  out      Log to use; never null.
      * @param  creator  The creator to test; never null.
      * @param  sequence The sequence to use; never null.
      * @param  index    The index of the sequence among all sequences tested.
      * @param  tokens   The tokens to identify the types of commands
      *                  to create.
      * @return True if the test succeeded, false if it did not.                 
      */
    private final static <E,T> boolean testCreator(Log out, CommandCreator<E,T> creator, Sequence<E> sequence, int index, List<T> tokens) {
        out.println()
           .println("Sequence ", index, ": ", sequence)
           .println()
           .println("Creating commands...");
        
        try {
            // Create a new command per token...
            List<Command<E>> commands = new ArrayList<Command<E>>(tokens.size());
            for (T token : tokens) {
                commands.add(creator.getCommand(sequence, token));
            }
            out.println()
               .println("Executing commands...");
            
            // Execution may fail, because we may create evil commands!
            // Not important here, since we test the factory creation
            // process...
            CommandProcessingResult<E> result = null;
            try {
                result = new CommandProcessor().execute(commands);
                if (result.executed) {
                    out.println("Result: ", result.result);
                } else {
                    out.println("Execution failed, but could be undone!");
                }
            } catch (CommandException e) {
                out.warn("Command execution failed").warn(e).println();
            }

            return true;
            
        } catch (Exception e) {
            out.error("Creator test failed").error(e).println();
        }
        return false;
    }

    /**
      * Tests the factories supplied in {@code factories} that creates commands
      * based on {@code sequence} as the single argument.
      *  
      * @param  <E>       The type of sequence and command values.
      * @param  out       Log to use; never null.
      * @param  sequence  The sequence to use; never null.
      * @param  index     The index of the sequence among all sequences tested.
      * @param  factories The factories to create the commands using 
      *                   {@code sequence} as the argument.
      * @return True if the test succeeded, false if it did not.                 
      */
    private final static <E> boolean testFactory(Log out, Sequence<E> sequence, int index, List<Factory<? extends Command<E>>> factories) {
        out.println()
           .println("Sequence ", index, ": ", sequence)
           .println()
           .println("Creating commands...");

        if (sequence == null) {
            out.println("Cannot create based on null value - continuing");
            return true;
        }
        
        try {
            // Create a new command per token...
            List<Command<E>> commands = new ArrayList<Command<E>>(factories.size());
            for (Factory<? extends Command<E>> factory : factories) {
                Command<E> command = factory.newInstance(sequence);
                out.println("Created command: ", factory, " -> ", command);
                commands.add(command);
            }
            out.println()
               .println("Executing commands...");
            
            CommandProcessingResult<E> result = null;
            try {
                result = new CommandProcessor().execute(commands);
                if (result.executed) {
                    out.println("Result: ", result.result);
                } else {
                    out.println("Execution failed, but could be undone!");
                }
            } catch (CommandException e) {
                out.error("Command execution failed").error(e).println();
            }

            return true;
            
        } catch (Exception e) {
            out.error("Factory test failed").error(e).println();
        }
        return false;
    }
    
}