/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 *
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.factorymethod;

import dk.rode.thesis.command.Command;
import dk.rode.thesis.command.LogCommand;
import dk.rode.thesis.command.NextCommand;
import dk.rode.thesis.command.NullCommand;
import dk.rode.thesis.command.ResetCommand;
import dk.rode.thesis.command.ReverseCommand;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.ReversibleSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.model.Sequence.State;

/**
  * A <i>reversible sequence command creator</i> extends
  * the sequence command {@link SequenceCommandCreator creator}
  * to allow for the creation of {@link ReverseCommand reverse} 
  * commands for {@link ReversibleSequence reversible} sequences. <p>
  *
  * This creator can create the following commands based on the
  * supplied internal sequence state: <ul>
  *
  * <li> {@code null:} a {@link NullCommand} is returned. <p>
  *
  * <li> {@link dk.rode.thesis.meta.model.Sequence.State#NORMAL}:
  *      creates a new {@link NextCommand} instance (inherited
  *      functionality). <p>
  *
  * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESET}:
  *      creates a new {@link ResetCommand} instance (inherited
  *      functionality). <p>
  *
  * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESTART}:
  *      if the sequence is reversible, a new {@link ReverseCommand}
  *      instance is created and returned. <p>
  *
  * <li> Otherwise a {@link LogCommand} is returned (default). <p>
  *
  * </ul>
  *
  * A reversible sequence command creator is stateless and thread-safe. <p>
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by created commands (and thus sequences).
  */
@Participant("ConcreteCreator")
public class ReversibleSequenceCommandCreator<E> extends SequenceCommandCreator<E> {

    /**
      * No-arg constructor.
      */
    public ReversibleSequenceCommandCreator() {
        super();
    }

    /**
      *
      */
    @Override
    protected Command<E> create(Sequence<E> sequence, State state) throws Exception {
        if (state == State.RESTART) {
            if (sequence instanceof ReversibleSequence) {
                return new ReverseCommand<E>(sequence);
            }
            // Fall-through...
        }
        // Let super handle it...
        return super.create(sequence, state);
    }

    /**
      * @return A new {@link LogCommand} instance; never null
      */
    @Override
    protected LogCommand<E> createDefault(Sequence<E> sequence) {
        return new LogCommand<E>(sequence);
    }
}
