/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.flyweight;

import java.util.Locale;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.util.Strings;
import dk.rode.thesis.strategy.StringablePolicy;

/**
  * An <i>abstract character</i> implements the basic traits of 
  * the {@link Character} interface. <p>
  * 
  * <b>Implementation notes</b>:<br> 
  * Many of the methods from {@code CharSequence}could be implemented in 
  * this class, but it is expected that individual sub-classes do this for 
  * better performance. This primary concern is that {@code hashCode} and 
  * {@code equals(Object)} are implemented properly to ensure correct 
  * flyweight behaviour (when stored in collections). <p>
  * 
  * The specific {@link Textual} methods are all implemented here as
  * they all implicitly work on {@code CharSequence} instances. As 
  * {@code Textual} implements {@code CharSequence}, we can simply
  * use a reference to {@code this} when actual string functionality
  * is forwarded to the {@link Strings} utility class. 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Flyweight")
public abstract class AbstractCharacter implements Character {

    /**
      * No-arg constructor.
      */
    protected AbstractCharacter() {
        super();
    }

    /**
      * 
      */
    public String lowerCaseFirst(Locale locale) {
        // Can use "this" because we implement CharSequence!
        return Strings.lowerCaseFirst(this, locale).toString();
    }
   
    /**
      * 
      */
    public String upperCaseFirst(Locale locale) {
        // Can use "this" because we implement CharSequence!
        return Strings.upperCaseFirst(this, locale).toString();
    }
    
    /**
      * 
      */
    public int indexOfIgnoreCase(CharSequence sequence, int index, Locale locale) {
        // Can use "this" because we implement CharSequence!
        return Strings.indexOfIgnoreCase(this, sequence, index, locale);
    }
    
    /**
      * Returns the hash code of this character.
      * 
      * @return The hash code.
      */
    public final int hashCode() {
        return this.toString().hashCode();
    }

    /**
      * Returns true if and only if {@code object} is a 
      * character having the same type and value as this
      * character, false otherwise.
      * 
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public final boolean equals(Object object) {
        // Implementation note: compare logically equivalent,
        // because several character factories may be used!
        
        if (object == null) {
            return false;
        } else if (object == this) {
            return true;
            
        // Works because the value of a Character is its string
        // value...
        } else if (this.getClass().equals(object.getClass())) {
            return this.toString().equals(object.toString());
        }
        return false;
    }

    /**
      *
      */
    public StringablePolicy<? super Character> getStringablePolicy(StringablePolicy<? super Character> policy) {
        if (policy == null) {
            return CharacterPolicy.SHORT;
        }
        return policy;
    }

    /**
      * 
      */
    public final CharSequence toString(StringablePolicy<? super Character> policy) {
        return this.getStringablePolicy(policy).toString(this);
    }
    
}
