/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.flyweight;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Locale;
import java.util.Set;

import dk.rode.thesis.iterator.SequenceIterator;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.ArraySequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.util.Strings;

/**
  * A <i>word</i> represents an ordered sequence of {@link Character characters} 
  * making up that word. <p>
  *  
  * A word is {@link Sequence#bounded() bounded},
  * {@link Sequence#consistent() consistent}, but not 
  * {@link #unique() unique} if the same character is 
  * contained more than once in the same word.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    Sentence
  */
@Participant("ConcreteFlyweight")
public class Word 
     extends ArraySequence<Character> 
  implements Sequence<Character>, Iterable<Character>, Textual<Word> {

    /**
      * The hash code for this word. 
      */
    private final int hashCode;
    /**
      * True if this word is {@link #unique() unique}, false
      * if not. <p>
      * 
      * If the same {@link Character character} is contained
      * more than once, this word is not unique. 
      */
    private final boolean unique;
    
    /**
      * Constructor, which creates this word to be comprised
      * of the characters in {@code characters}.
      * 
      * @param  characters The characters; cannot be null or empty.
      * @throws NullPointerException If {@code characters} is null.
      * @throws IllegalArgumentException If {@code characters} is empty.
      */
    Word(Character[] characters) {
        super(characters);
        this.hashCode = this.current().hashCode();
        Set<Character> set = new HashSet<Character>();
        boolean unique = true;
        for (Character c : characters) {
            if (set.contains(c)) {
                unique = false;
            }
            set.add(c);
        }
        this.unique = unique;
    }

    /**
      * Copy constructor. <p>
      * 
      * @param  word The word to copy; cannot be null.
      * @throws NullPointerException If {@code word} is null.
      */
    Word(Word word) {
        super(word);
        this.hashCode = word.hashCode;
        this.unique = word.unique;
    }

    /**
      * 
      */
    public Word copy() {
        // Cannot copy because it acts as a flyweight!
        return this;
    }

    /**
      * Returns true if this word represents a single white space
      * character.
      *  
      * @return True if white space, false if not.
      */
    public boolean isWhiteSpace() {
        return ((this.elements.length == 1) && (this.elements[0] instanceof Whitespace));
    }
    
    /** 
      * Returns true if any given {@link Character character} 
      * in this word is only contained once, false otherwise.
      * 
      * @return True if unique, false if not.
      */
    public boolean unique() {
        return this.unique;
    }
    
    /**
      * 
      */
    public Iterator<Character> iterator() {
        // Hidden copy to avoid sharing of state...
        return new SequenceIterator<Character>(new Word(this));
    }

    /**
      * 
      */
    public String lowerCaseFirst(Locale locale) {
        return this.elements[0].lowerCaseFirst(locale);
    }
    
    /**
      * 
      */
    public String upperCaseFirst(Locale locale) {
        return this.elements[0].upperCaseFirst(locale);
    }
    
    /**
      * 
      */
    public int indexOfIgnoreCase(CharSequence sequence, int index, Locale locale) {
        // Can use "this" because we implement CharSequence!
        return Strings.indexOfIgnoreCase(this, sequence, index, locale);
    }
    
    /**
      * 
      */
    public char charAt(int index) {
        return this.toString().charAt(index);
    }

    /**
      * 
      */
    public int length() {
        return this.toString().length();
    }

    /**
      * 
      */
    public CharSequence subSequence(int start, int end) {
        return this.toString().subSequence(start, end);
    }
    
    /**
      * 
      */
    public final int hashCode() {
        return this.hashCode; // hash code of first character in this word
    }

    /**
      * 
      */
    public final boolean equals(Object object) {
        if (!(object instanceof Word)) {
            return false;
        } else if (this == object) {
            return true;
        }
        return (this.compareTo((Word)object) == 0);
    }
    
    /**
      * 
      */
    public int compareTo(Word word) {
        // Invariant: words are bounded!

        // Simulates complex behaviour! Could just have
        // used this.toString().compareTo(word.toString())
        // or traversed the alphabet directly!
        
        Iterator<Character> i = this.iterator(),
                            j = word.iterator();

        boolean iNext = true, jNext = true;
        
        // Iterate through all characters...
        while ((iNext = i.hasNext()) & (jNext = j.hasNext())) { // No short circuit!
            int n = i.next().compareTo(j.next());
            // Zero = equals...
            if (n != 0) {
                return n;
            }
        }

        // this word contains fewer letters... 
        if (jNext) {
            return 1;
            
        // "word" contains fewer letters...
        } else if (iNext) {
            return -1;
            
        // Same number of letters...
        } else {
            return 0;
        }
    }

    /**
      * Returns this word as a string.
      * 
      * @return This word as a string; never null.
      */
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (Character c : this) {
            sb.append(c);
        }
        return sb.toString();
    }
    
}
