/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import java.util.ArrayList;
import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;

/**
  * A <i>flow expression</i> represents one or more expressions
  * to be evaluated in order, one at a time. The result of the 
  * evaluation is the result of the <i>last</i> expression evaluated.<p>
  *
  * Expressions to be evaluated by a flow expression must explicitly be
  * {@link #add(Expression) added}. Once the last expression has been 
  * added, the flow expression must be {@link #initialise() initialised} 
  * before evaluation, or the evaluation will fail. Once initialised, 
  * additional expressions cannot be added. <p> 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of value the evaluation of this expression 
  *             produces.
  */
@Participant("NonTerminalExpression")
public class FlowExpression<E> 
     extends AbstractExpression<E>
  implements NonTerminalExpression<E>, InitialisableExpression<E> {

    /**
      * The expressions to evaluate, in order. <p>
      * 
      * Never null, and never empty.
      */
    private final List<Expression<? extends E>> expressions;
    /**
      * True if this expression has been initialised, 
      * false if not. 
      */
    private boolean initialised;
    
    /**
      * No-arg constructor. <p>
      *
      * Expressions to be evaluated <b>must</b> be 
      * {@link #add(Expression) added} before evaluation or the
      * evaluation of this expression will fail.
      */
    public FlowExpression() {
        this.expressions = new ArrayList<Expression<? extends E>>();
        this.initialised = false;
    }
    
    /**
      * Copy constructor.
      * 
      * @param  expression The expression to copy; cannot be null.
      * @throws NullPointerException If {@code expression} is null.
      */
    public FlowExpression(FlowExpression<E> expression) {
        this();
        for (Expression<? extends E> e : expression.expressions) {
            this.expressions.add(e.copy());
        }
        this.initialised = expression.initialised;
    }
 
    /**
      * 
      */
    public FlowExpression<E> copy() {
        return new FlowExpression<E>(this);
    }

    /**
      *
      */
    public boolean isInitialised() {
        return this.initialised;
    }

    /**
      * 
      */
    public FlowExpression<E> initialise() throws ExpressionException {
        if (this.initialised) {
            throw new ExpressionException("Flow expression already initialised");
        }
        this.initialised = true;
        return this;
    }

    /**
      * Adds the expression supplied as {@code expression} to
      * this flow expression. <p>
      * 
      * The same expression can be added several times. <p>
      * 
      * After the last expression has been added, this flow 
      * expression must be {@link #initialise() initialised}
      * before it can be evaluated.
      *  
      * @param  expression The expression to add; cannot be null.
      * @return This flow expression; never null.
      * @throws NullPointerException If {@code expression} is null.
      * @throws ExpressionException If this flow expression has already
      *         been initialised.
      */
    public FlowExpression<E> add(Expression<? extends E> expression) throws ExpressionException {
        if (this.initialised) {
            throw new ExpressionException("Flow expression has already been initialised");
        } else if (expression == null) {
            throw new NullPointerException("Expression cannot be null");
        }
        this.expressions.add(expression);
        return this;
    }
    
    /**
      * @return The expressions contained in this expression. 
      */
    public List<Expression<?>> operands() {
        return new ArrayList<Expression<?>>(this.expressions);
    }

    /**
      * @return {@code flow}
      */
    public final String name() {
        return "flow";
    }
    
    /**
      * @throws ExpressionException If this expression has not been 
      *         initialised, or if the evaluation fails. 
      */
    public E evaluate(Context context) throws ExpressionException {
        if (!this.initialised) {
            throw new ExpressionException("Flow expression not initialised");
        }
        LogFactory.getLog(this).println("Evaluating: ", this.asSymbol(context.reset()));
        E value = null;
        for (Expression<? extends E> e : this.expressions) {
            value = e.evaluate(context);
        }
        return value;
    }

    /**
      * Returns {@code x, y, z, ..}, where {@code x}, {@code y}, 
      * {@code z}, .., are the symbol representations of the
      * contained expressions.   
      */
    public String asSymbol(Context context) throws ExpressionException {
        if (!this.initialised) {
            throw new ExpressionException("Flow expression not initialised");
        }
        StringBuilder sb = new StringBuilder();
        for (int i = 0, n = this.expressions.size(); i < n; i++) {
            Expression<? extends E> expression = this.expressions.get(i);
            if (i > 0) {
                sb.append(", ");
            }
            SymbolIdiom.asSymbol(context, this, expression, sb);
        }
        return sb.toString();
    }
    
}
