/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import java.util.List;

import static dk.rode.thesis.meta.util.Util.add;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>next expression</i> will invoke {@link Sequence#next() next()}
  * a number of times on a given {@link Sequence sequence} when 
  * evaluated. <p>
  * 
  * The number of times to invoke {@code next()} can be 
  * {@link #NextExpression(Sequence, int) constant},
  * but it can also be represented by {@link #NextExpression(Sequence, Expression) any}
  * expression of the type {@code Expression<? extends Number>}, 
  * where the actual count is the absolute integer value of the 
  * result of the expression evaluation (if zero, a count of
  * one is used). 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    CurrentExpression
  * @see    SetExpression
  * @see    ResetExpression
  * @see    ReverseExpression
  * 
  * @param  <E> The type of value the evaluation of this expression 
  *             produces. The value corresponds to the (super-) type of 
  *             values delivered by the manipulated {@link Sequence sequence}.
  */
@Participant("TerminalExpression")
public class NextExpression<E> extends SequenceExpression<E,E> {

    /**
      * The number of times {@link Sequence#next()} is invoked
      * for the manipulated {@link SequenceExpression#sequence 
      * sequence}. <p>
      * 
      * The absolute integer value is used. If zero, a count of
      * one is used. 
      */
    private final Expression<? extends Number> count;
    
    /**
      * Constructor, which will call {@link Sequence#next()} a single
      * time on {@code sequence} when this expression is evaluated. <p>
      * 
      * The identity hash code of {@code sequence} is not included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression.
      * 
      * @param  sequence The sequence to manipulate by this expression;
      *                  cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public NextExpression(Sequence<? extends E> sequence) {
        this(sequence, false);
    }
    
    /**
      * Constructor, which will call {@link Sequence#next()} a single
      * time on {@code sequence} when this expression is evaluated. <p>
      * 
      * The identity hash code of {@code sequence} is included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression if {@code includeHashCode} is true.
      * 
      * @param  sequence        The sequence to manipulate by this expression;
      *                         cannot be null.
      * @param  includeHashCode True to include the identity hash code
      *                         of {@code sequence} in the symbolic
      *                         representation, false not to.                 
      * @throws NullPointerException If {@code sequence} is null.
      */
    public NextExpression(Sequence<? extends E> sequence, boolean includeHashCode) {
        this(sequence, 1, includeHashCode);
    }
    
    /**
      * Constructor, which will call {@link Sequence#next()} {@code count}
      * time(s) on {@code sequence} when this expression is evaluated. <p>
      * 
      * The absolute value of count is used. <p>
      * 
      * The identity hash code of {@code sequence} is not included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression.
      * 
      * @param  sequence The sequence to manipulate by this expression;
      *                  cannot be null.
      * @param  count    The number of times to call {@code next()} on
      *                  {@code sequence} when this expression is 
      *                  evaluated.                  
      * @throws NullPointerException If {@code sequence} is null.
      */
    public NextExpression(Sequence<? extends E> sequence, int count) {
        this(sequence, count, false);
    }

    /**
      * Constructor, which will call {@link Sequence#next()} {@code count}
      * time(s) on {@code sequence} when this expression is evaluated. <p>
      * 
      * The absolute value of count is used. <p>
      * 
      * The identity hash code of {@code sequence} is included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression if {@code includeHashCode} is true.
      * 
      * @param  sequence        The sequence to manipulate by this expression;
      *                         cannot be null.
      * @param  count           The number of times to call {@code next()} on
      *                         {@code sequence} when this expression is 
      *                         evaluated.                  
      * @param  includeHashCode True to include the identity hash code
      *                         of {@code sequence} in the symbolic
      *                         representation, false not to.                 
      * @throws NullPointerException If {@code sequence} is null.
      */
    public NextExpression(Sequence<? extends E> sequence, int count, boolean includeHashCode) {
        this(sequence, new ConstantExpression<Integer>(Integer.class, count), includeHashCode);
    }
    
    /**
      * Constructor, which will call {@link Sequence#next()} the 
      * number of times corresponding to the absolute value of 
      * the result of evaluating {@code count} when this
      * expression is evaluated. <p> 
      * 
      * The identity hash code of {@code sequence} is not included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression.
      *
      * @param  sequence The sequence to manipulate by this expression;
      *                  cannot be null.
      * @param  count    The expression that will determine the number of
      *                  times {@code next()} will be invoked on {@code sequence}
      *                  when this expression is evaluated.
      * @throws NullPointerException If either argument is null.
      */
    public NextExpression(Sequence<? extends E> sequence, Expression<? extends Number> count) {
        this(sequence, count, false);
    }
    
    /**
      * Constructor, which will call {@link Sequence#next()} the 
      * number of times corresponding to the absolute value of 
      * the result of evaluating {@code count} when this
      * expression is evaluated. <p> 
      * 
      * The identity hash code of {@code sequence} is included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression if {@code includeHashCode} is true.
      *
      * @param  sequence        The sequence to manipulate by this expression;
      *                         cannot be null.
      * @param  count           The expression that will determine the number of
      *                         times {@code next()} will be invoked on {@code sequence}
      *                         when this expression is evaluated.
      * @param  includeHashCode True to include the identity hash code
      *                         of {@code sequence} in the symbolic
      *                         representation, false not to.                 
      * @throws NullPointerException If {@code sequence} or {@code count} are null.
      */
    public NextExpression(Sequence<? extends E> sequence, Expression<? extends Number> count, boolean includeHashCode) {
        super(sequence, includeHashCode);
        if ((this.count = count) == null) {
            throw new NullPointerException("Count expression cannot be null");
        }
    }
    
    /**
      * Copy constructor. <p>
      * 
      * The actual {@link Sequence sequence} manipulated by
      * {@code expression} is <b>not</b> copied, but used
      * as is.
      * 
      * @param  expression The expression to copy; cannot be null.
      * @throws NullPointerException If {@code expression} is null.
      */
    public NextExpression(NextExpression<E> expression) {
        super(expression);
        this.count = expression.count.copy();
    }
 
    /**
      *
      */
    public NextExpression<E> copy() {
        return new NextExpression<E>(this);
    }
    
    /**
      * 
      */
    @Override
    public List<Expression<?>> operands() {
        return add(super.operands(), this.count);
    }
    
    /**
      * @return {@code next}
      */
    public final String name() {
        return "next";
    }
    
    /**
      * @return The result of the last invocation of 
      *         {@link Sequence#next()}; never null.
      */
    public E evaluate(Context context) throws ExpressionException {
        String s = this.asSymbol(context.reset());
        Log out = LogFactory.getLog(this).println("Evaluating: ", s);
        // Fetch count...
        int count = Math.max(1, Math.abs(this.count.evaluate(context).intValue()));
        E value = null;
        for (int i = 0; i < count; i++) {
            value = this.sequence.next();
        }
        out.println(s, " -> ", value);
        return value;
    }
    
    /**
      * Returns {@code next[x]}, where {@code x} is the symbolic
      * representation of the expression that will determine the 
      * number of times to invoke {@code next()}, or {@code next#123[x]},
      * where {@code 123} is the identity hash code of the manipulated
      * sequence.    
      */
    @Override
    public String asSymbol(Context context) throws ExpressionException {
        StringBuilder sb = new StringBuilder();
        sb.append(super.asSymbol(context)).append('[');
        SymbolIdiom.asSymbol(context, this, this.count, sb);
        return sb.append(']').toString();
    }
    
    /**
      * 
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; sequence: ").append(this.sequence.toString(SequencePolicy.IDENTITY))
                 .append("; count: ").append(this.count).append(']').toString();
    }
    
}
