/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import static dk.rode.thesis.meta.util.Util.add;

import java.util.ArrayList;
import java.util.List;

import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>not expression</i> (!) for any expression evaluating
  * to a {@code Boolean} value. <p>
  *
  * This class does not declare a copy constructor, but can
  * still be {@link #copy() copied}.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    AndExpression
  * @see    OrExpression
  * @see    CompareExpression
  * @see    EqualExpression
  */
public class NotExpression 
     extends AbstractExpression<Boolean>
  implements NonTerminalExpression<Boolean> {

    /**
      * The expression to evaluated and the result negated. <p>
      * 
      * Never null.
      */
    private final Expression<Boolean> expression;

    /**
      * Constructor.
      * 
      * @param  expression The expression to use; cannot be null and
      *                    is not copied.
      * @throws NullPointerException If {@code expression} is null.                   
      */
    public NotExpression(Expression<Boolean> expression) {
        if ((this.expression = expression) == null) {
            throw new NullPointerException("Expression cannot be null");
        }
    }

    /**
      *
      */
    public NotExpression copy() {
        // No copy constructor, so simply copy expression and use
        // as argument for normal constructor...
        return new NotExpression(this.expression.copy());
    }

    /**
      * @return {@code not}
      */
    public final String name() {
        return "not";
    }
    
    /**
      *
      */
    public List<Expression<?>> operands() {
        return add(new ArrayList<Expression<?>>(), this.expression);
    }
    
    /**
      * 
      */
    public Boolean evaluate(Context context) throws ExpressionException {
        LogFactory.getLog(this).println("Evaluating: ", this.asSymbol(context.reset()));
        return !this.expression.evaluate(context);
    }
    
    /**
      * Returns {@code (!x)}, where {@code x} is the symbol
      * representation of the expression negated by this
      * expression.    
      */
    public String asSymbol(Context context) throws ExpressionException {
        StringBuilder sb = new StringBuilder().append("(!");
        SymbolIdiom.asSymbol(context, this, this.expression, sb);
        return sb.append(')').toString();
    }
    
    /**
      * 
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; expression: ").append(this.expression).append(']').toString();
    }
    
}
