/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>reset expression</i> will invoke {@link Sequence#reset() reset()}
  * on a given {@link Sequence sequence} when evaluated. 
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    CurrentExpression
  * @see    NextExpression
  * @see    SetExpression
  * @see    ReverseExpression
  * 
  * @param  <E> The type of value the evaluation of this expression 
  *             produces. The value corresponds to the (super-) type of 
  *             values delivered by the manipulated {@link Sequence sequence}.
  */
@Participant("TerminalExpression")
public class ResetExpression<E> extends SequenceExpression<E,E> {
    
    /**
      * Constructor. <p>
      * 
      * The identity hash code of {@code sequence} is not included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression.
      * 
      * @param  sequence The sequence to manipulate by this expression;
      *                  cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public ResetExpression(Sequence<? extends E> sequence) {
        this(sequence, false);
    }

    /**
      * Constructor. <p>
      * 
      * The identity hash code of {@code sequence} is included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression if {@code includeHashCode} is true.
      * 
      * @param  sequence        The sequence to manipulate by this expression;
      *                         cannot be null.
      * @param  includeHashCode True to include the identity hash code
      *                         of {@code sequence} in the symbolic
      *                         representation, false not to.                 
      * @throws NullPointerException If {@code sequence} is null.
      */
    public ResetExpression(Sequence<? extends E> sequence, boolean includeHashCode) {
        super(sequence, includeHashCode);
    }
    
    /**
      * Copy constructor. <p>
      * 
      * The actual {@link Sequence sequence} manipulated by
      * {@code expression} is <b>not</b> copied, but used
      * as is.
      * 
      * @param  expression The expression to copy; cannot be null.
      * @throws NullPointerException If {@code expression} is null.
      */
    public ResetExpression(ResetExpression<E> expression) {
        super(expression);
    }
 
    /**
      *
      */
    public ResetExpression<E> copy() {
        return new ResetExpression<E>(this);
    }

    /**
      * @return {@code reset}
      */
    public final String name() {
        return "reset";
    }
    
    /**
      * @return The current value of the manipulated sequence after
      *         {@link Sequence#reset()} has been invoked; never null.
      */
    public E evaluate(Context context) throws ExpressionException {
        this.sequence.reset();
        E current = this.sequence.current();
        LogFactory.getLog(this).println("Evaluating: ", this.asSymbol(context.reset()), " -> ", current);
        return current;
    }
    
}
