/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import java.util.ArrayList;
import java.util.List;

import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.ObjectPolicy;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>sequence expression</i> represents a terminal expression
  * used to manipulate a given {@link Sequence sequence}. <p>
  * 
  * When a sequence expression is copied, the manipulated sequence 
  * is <b>not</b> copied, but used as is.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <T> The type of sequence values delivered by the  
  *             manipulated {@link Sequence sequence}. Normally, 
  *             but not required to be, identical with or a 
  *             sub-type of {@code E}.
  * @param  <E> The type of value the evaluation of this expression 
  *             produces.             
  */
public abstract class SequenceExpression<T,E> 
              extends AbstractExpression<E> 
           implements TerminalExpression<E> {

    /**
      * The sequence manipulated by this expression. <p>
      * 
      * Never null.
      */
    protected final Sequence<? extends T> sequence;
    /**
      * True to include the identity hash code of {@link #sequence} 
      * in the symbolic representation, false if not. <p>
      * 
      * Default false.
      */
    protected final boolean includeHashCode;
    
    /**
      * Constructor. <p>
      * 
      * The identity hash code of {@code sequence} is included
      * in the {@link #asSymbol(Context) symbolic} representation of 
      * this expression if {@code includeHashCode} is true.
      * 
      * @param  sequence        The sequence to manipulate by this 
      *                         expression; cannot be null.
      * @param  includeHashCode True to include the identity hash code
      *                         of {@code sequence} in the symbolic
      *                         representation, false not to.                 
      * @throws NullPointerException If {@code sequence} is null.                 
      */
    protected SequenceExpression(Sequence<? extends T> sequence, boolean includeHashCode) {
        if ((this.sequence = sequence) == null) {
            throw new NullPointerException("Sequence cannot be null");
        }
        this.includeHashCode = includeHashCode;
    }
    
    /**
      * Copy constructor. <p>
      * 
      * The actual {@link Sequence sequence} manipulated by
      * {@code expression} is <b>not</b> copied, but used
      * as is.
      * 
      * @param  expression The expression to copy; cannot be null.
      * @throws NullPointerException If {@code expression} is null.                 
      */
    protected SequenceExpression(SequenceExpression<? extends T,E> expression) {
        this.sequence = expression.sequence;
        this.includeHashCode = expression.includeHashCode;
    }
    
    /**
      * 
      */
    public List<Expression<?>> operands() {
        return new ArrayList<Expression<?>>(1);
    }

    /**
      * Returns {@code name}, where {@code name} is the 
      * actual {@link #name() name} of this expression, or
      * {@code name#123}, where {@code 123} is the identity
      * hash code of the manipulated sequence.
      */
    public String asSymbol(Context context) throws ExpressionException { // can be thrown by sub-classes
        String name = this.name();
        if (this.includeHashCode) {
            name += "#" + System.identityHashCode(this.sequence);
        }
        return name; 
    }
    
    /**
      * Returns the hash code of this expression. 
      *
      * @return The hash code.
      */
    public int hashCode() {
        return super.hashCode() + 7 * this.sequence.hashCode();
    }

    /**
      * Returns true if {@code object} is a sequence expression
      * of the same type as this expression manipulating the
      * same sequence using the same operands, false if not. 
      *
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (super.equals(object)) {
            SequenceExpression<?,?> se = (SequenceExpression<?,?>)object;
            return this.sequence.equals(se.sequence);
        }
        return false;
    }
    
    /**
      * 
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; sequence: ").append(this.sequence.toString(SequencePolicy.IDENTITY)).append(']').toString();
    }
    
}
