/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import java.util.List;

/**
  * A <i>type expression decorator</i> allows <b>any</b> expression
  * to be explicitly associated with the type of the values the
  * evaluation of it will produce. <p>
  * 
  * In case the decorated expression is {@link InitialisableExpression 
  * initialisable}, it will be initialised by the {@link #initialise()}
  * method defined in this decorator. If the decorated expression is not
  * initialisable, the {@code initialise()} and 
  * {@link #isInitialised()} methods does nothing. <p> 
  * 
  * <b>Implementation notes</b>:<br>
  * This is a fine example of the 
  * {@linkplain dk.rode.thesis.decorator Decorator} pattern
  * at work.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of value the evaluation of the decorated
  *             expression produces.             
  */
public class TypedExpressionDecorator<E> implements TypedExpression<E>, InitialisableExpression<E> {

    /**
      * The decorated expression. <p>
      * 
      * Never null.
      */
    protected final Expression<E> expression;
    /**
      * The type of values the evaluation of {@link #expression}
      * produces. <p>
      * 
      * Never null.
      */
    protected final Class<E> type;
    /**
      * True if the evaluation type of {@link #expression} will be 
      * included in the symbolic representation of it, false if not. <p>
      * 
      * Default false.
      */
    protected final boolean includeTypeInfo;
    
    /**
      * Constructor, which decorates the typed expression
      * supplied as {@code expression}. <p>
      * 
      * The symbolic representation of {@code expression} will 
      * include type information regarding the type of 
      * {@code expression}.
      * 
      * @param  expression The typed expression to decorate; cannot be null.
      * @throws NullPointerException If {@code expression} is null.                 
      */
    public TypedExpressionDecorator(TypedExpression<E> expression) {
        this(expression, expression.type());
    }

    /**
      * Constructor, which decorates the expression
      * supplied as {@code expression} having the type 
      * supplied as {@code type}. <p>
      * 
      * The symbolic representation of {@code expression} will 
      * include type information regarding {@code type}.
      * 
      * @param  expression The expression to decorate; cannot be null.
      * @param  type       The actual type; cannot be null.
      * @throws NullPointerException If either argument is null.                 
      */
    public TypedExpressionDecorator(Expression<E> expression, Class<E> type) {
        this(expression, type, true);
    }
    
    /**
      * Constructor, which decorates the expression
      * supplied as {@code expression} having the type 
      * supplied as {@code type}. <p>
      * 
      * The symbolic representation of {@code expression} will 
      * include type information regarding {@code type}
      * if {@code includeTypeInfo} is true, not if false.
      * 
      * @param  expression      The expression to decorate; cannot be null.
      * @param  type            The type of values the evaluation of 
      *                         {@code expression} produces. 
      * @param  includeTypeInfo True to include type information in
      *                         the symbolic representation of 
      *                         {@code expression}, false if not.             
      * @throws NullPointerException If {@code expression} or {@code type} is null.                 
      */
    public TypedExpressionDecorator(Expression<E> expression, Class<E> type, boolean includeTypeInfo) {
        if ((this.expression = expression) == null) {
            throw new NullPointerException("Expression cannot be null");
        } else if ((this.type = type) == null) {
            throw new NullPointerException("Type cannot be null");
        }
        this.includeTypeInfo = includeTypeInfo;
    }
    
    /**
      * Copy constructor. <p>
      * 
      * @param  expression The expression to copy; cannot be null.
      * @throws NullPointerException If {@code expression} is null.                 
      */
    public TypedExpressionDecorator(TypedExpressionDecorator<E> expression) {
        this.expression = expression.expression.copy();
        this.type = expression.type; // immutable
        this.includeTypeInfo = expression.includeTypeInfo;
    }
    
    /**
      *
      */
    public TypedExpressionDecorator<E> copy() {
        return new TypedExpressionDecorator<E>(this);
    }
 
    /**
      * 
      */
    public boolean isInitialised() {
        if (this.expression instanceof InitialisableExpression) {
            InitialisableExpression<E> ie = (InitialisableExpression<E>)this.expression;
            return ie.isInitialised();
        }
        return true;
    }
    
    /**
      * 
      */
    public TypedExpressionDecorator<E> initialise() throws ExpressionException {
        if (this.expression instanceof InitialisableExpression) {
            InitialisableExpression<E> ie = (InitialisableExpression<E>)this.expression;
            ie.initialise();
        }
        return this;
    }
    
    /**
      * 
      */
    public String name() {
        return this.expression.name();
    }

    /**
      * 
      */
    public Class<E> type() {
        return this.type;
    }
    
    /**
      * 
      */
    public List<Expression<?>> operands() {
        return this.expression.operands();
    }

    /**
      *
      */
    public E evaluate(Context context) throws ExpressionException {
        return this.expression.evaluate(context);
    }
    
    /**
      * 
      */
    public boolean contains(Expression<?> expression) {
        return (this == expression) || (this.expression.contains(expression));
    }
    
    /**
      * Returns {@code x}, where {@code x} is the symbolic 
      * representation of the decorated expression, or if type information 
      * is included, {@code Type{x}}, where {@code Type} is the simple
      * name of the type.   
      */
    public String asSymbol(Context context) throws ExpressionException {
        StringBuilder sb = new StringBuilder();
        if (this.includeTypeInfo) {
            sb.append(this.type().getSimpleName()).append('{'); 
        }
        sb.append(this.expression.asSymbol(context));
        if (this.includeTypeInfo) {
            sb.append('}');
        }
        return sb.toString();
    }
    
    /**
      * Returns the hash code of this expression. 
      *
      * @return The hash code.
      */
    public int hashCode() {
        return this.expression.hashCode();
    }

    /**
      * Returns true if {@code object} is a typed expression
      * decorator decorating the same expressions as this
      * decorator, tested using {@code equals(Object)}, false
      * if not.
      *
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (!(object instanceof TypedExpressionDecorator)) {
            return false;
        }
        TypedExpressionDecorator<?> ted = (TypedExpressionDecorator<?>)object;
        return this.expression.equals(ted.expression);
    }
    
    /**
      * 
      */
    public String toString() {
        return this.expression.toString();
    }
    
}
