/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.interpreter;

import java.util.ArrayList;
import java.util.List;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>variable expression</i> represents an expression that
  * can be assigned a given value, which will be stored
  * in a given {@link Context context} supplied at evaluation
  * time. <p>
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of value the evaluation of this expression 
  *             produces.
  */
@Participant("NonTerminalExpression")
public class VariableExpression<E> implements NonTerminalExpression<E>, TypedExpression<E> {

    /**
      * The unique name identifying this expression. <p>
      * 
      * Names are case-sensitive. <p>
      * 
      * Never null, nor empty. <p>
      * 
      * This name is <b>not</b> the same as {@link #name()}!
      * 
      * @see Context#lookup(String)
      */
    public final String name;
    /**
      * The type identifying the value returned by this expression. <p>
      * 
      * Never null.
      */
    private final Class<E> type;
    /**
      * True if the type of this variable will be included in the
      * symbolic representation of it, false if not. <p>
      * 
      * Default false.
      */
    protected final boolean includeTypeInfo;
    
    /**
      * Constructor. <p>
      * 
      * The symbolic representation of this variable will not 
      * include type information regarding {@code type}.
      * 
      * @param  type The type of the value represented by this expression;
      *              cannot be null.
      * @param  name The name of this expression; cannot be null or
      *              empty.
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code name} is illegal.
      * @see    Context#lookup(String)
      */
    public VariableExpression(Class<E> type, String name) {
        this(type, name, false);
    }
    
    /**
      * Constructor. <p>
      * 
      * The symbolic representation of this variable will 
      * include type information regarding {@code type}
      * if {@code includeTypeInfo} is true, not if false.
      * 
      * @param  type            The type of the value represented by this expression;
      *                         cannot be null.
      * @param  name            The name of this expression; cannot be null or
      *                         empty.
      * @param  includeTypeInfo True to include type information in
      *                         the symbolic representation, false if not.             
      * @throws NullPointerException If either {@code type} or {@code name} are null.
      * @throws IllegalArgumentException If {@code name} is illegal.
      * @see    Context#lookup(String)
      */
    public VariableExpression(Class<E> type, String name, boolean includeTypeInfo) {
        if ((this.type = type) == null) {
            throw new NullPointerException("Type cannot be null");
        } 
        if ((this.name = name.trim()).length() == 0) {
            throw new IllegalArgumentException("Name cannot be empty");
        }
        this.includeTypeInfo = includeTypeInfo;
    }

    /**
      * Copy constructor.
      * 
      * @param  expression The expression to copy; cannot be null.
      * @throws NullPointerException If {@code expression} is null.
      */
    public VariableExpression(VariableExpression<E> expression) {
        this(expression.type, expression.name);
    }

    /**
      * 
      */
    public VariableExpression<E> copy() {
        return new VariableExpression<E>(this);
    }

    /**
      * @return {@code variable}
      */
    public String name() {
        return "variable";
    }
    
    /**
      * 
      */
    public Class<E> type() {
        return this.type;
    }
    
    /**
      * 
      */
    public List<Expression<?>> operands() {
        return new ArrayList<Expression<?>>(0);
    }

    /**
      * 
      */
    public E evaluate(Context context) throws ExpressionException {
        Log out = LogFactory.getLog(this).print("Evaluating: ", this.asSymbol(context.reset()));
        E value = context.lookup(this);
        out.println(" -> ", value);
        return value;
    }

    /**
      * 
      */
    public boolean contains(Expression<?> expression) {
        return (this == expression);
    }
    
    /**
      * Returns {@code x}, where {@code x} is the name of 
      * this variable, or if type information is included,
      * {@code Type{x}}, where {@code Type} is the simple
      * name of the type.  
      */
    public String asSymbol(Context context) {
        String value = this.name;
        if (this.includeTypeInfo) {
            value = this.type.getSimpleName() + "{" + value + "}"; 
        }
        return value;
    }
    
    /**
      * Returns the hash code of this expression.
      * 
      * @return The hash code; never null.
      */
    public int hashCode() {
        return this.name.hashCode();
    }
    
    /**
      * Returns true if {@code object} is an instance of this
      * expression type and have the same name and type, false if not.
      * 
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (!(object instanceof VariableExpression)) {
            return false;
        }
        VariableExpression<?> ve = (VariableExpression<?>)object;
        return this.name.equals(ve.name) && (this.type.equals(ve.type));
    }
    
    /**
      * 
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.NAME_OPEN.toString(this);
        return sb.append(this.name).append(']').toString();
    }
    
}
