/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */

/**
  * Implementations and examples of the <b>Interpreter</b> 
  * design pattern [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.243]. <p>
  * 
  * <b>Intent</b>:<ul><p><i>
  * 
  *   Given a language, define a representation for its grammar along 
  *   with an interpreter that uses the representation to interpret 
  *   sentences in the language.
  * 
  * </i></ul>
  * 
  * Here, a small language have been constructed that can manipulate 
  * {@link dk.rode.thesis.meta.model.Sequence} objects and
  * perform simple tests using boolean expressions that may or may 
  * not depend on sequence values. <p>
  *
  * The <b>AbstractExpression</b> participant is represented by
  * the {@link dk.rode.thesis.interpreter.Expression} interface. 
  * Any expression can be evaluated by invoking its 
  * {@link dk.rode.thesis.interpreter.Expression#evaluate(Context)}
  * method, but the {@link dk.rode.thesis.interpreter.Interpreter} 
  * class should be used for interpretation. <p>
  * 
  * The <b>TerminalExpression</b> participant is represented by literals 
  * that describe sequence actions, such as {@code 'next'} for
  * {@link dk.rode.thesis.meta.model.Sequence#next()}, 
  * {@code 'reset'} for {@link dk.rode.thesis.meta.model.Sequence#reset()},
  * etc. The {@link dk.rode.thesis.interpreter.TerminalExpression}
  * marker interface denotes terminal symbols, and any terminal
  * expression manipulating a given {@code Sequence} inherits the
  * {@link dk.rode.thesis.interpreter.SequenceExpression} 
  * class. The {@code 'next'} literal, for example, is represented 
  * by the {@link dk.rode.thesis.interpreter.NextExpression}
  * class. The other terminal expressions are 
  * {@link dk.rode.thesis.interpreter.CurrentExpression}, 
  * {@link dk.rode.thesis.interpreter.SetExpression},
  * {@link dk.rode.thesis.interpreter.ResetExpression}, and
  * {@link dk.rode.thesis.interpreter.ReverseExpression}. <p>
  * 
  * The grammar rules are represented by the <b>NonTerminalExpression</b>
  * participant. In this package, each class that implements the marker 
  * interface {@link dk.rode.thesis.interpreter.NonTerminalExpression} 
  * is a non terminal. Examples include {@link dk.rode.thesis.interpreter.FlowExpression},
  * {@link dk.rode.thesis.interpreter.ConditionalExpression},
  * {@link dk.rode.thesis.interpreter.AssignmentExpression}, etc. <p>
  * 
  * The <b>Context</b> participant is represented by the 
  * {@link dk.rode.thesis.interpreter.Context} class, which
  * can store both {@link dk.rode.thesis.interpreter.VariableExpression 
  * variables} and {@link dk.rode.thesis.interpreter.ConstantExpression 
  * constants}. <p>
  * 
  * The <b>Client</b> participant is the test application, i.e. the 
  * {@link dk.rode.thesis.interpreter.Main} class. <p> 
  *  
  * Below, an <b>informal</b> description of the grammar is presented:
  * <pre>
  *    
  *    expression            ::= literal | constant | assignment | condition | argument | flow | '(' expression ')' | break
  *    flow                  ::= expression ',' expression
  *    break                 ::= expression (break and evaluate) | null (terminate)
  *    argument              ::= expression&lt;literal> '[' expression ']'   
  *    expression&lt;literal>   ::= current-expression | next-expression | set-expression | reset-expression | reverse-expression
  *    expression&lt;boolean>   ::= and-expression | or-expression | not-expression | equal-expression | comparison-expression
  *    and-expression        ::= expression1&lt;boolean> '&&' expression2&lt;boolean>  
  *                            | expression1&lt;boolean> '&'  expression2&lt;boolean> 
  *    or-expression         ::= expression1&lt;boolean> '||' expression2&lt;boolean> 
  *                            | expression1&lt;boolean> '|'  expression2&lt;boolean>
  *    not-expression        ::= '!' expression&lt;boolean>
  *    equal-expression      ::= expression1 '==' expression2 
  *    comparison-expression ::= expression1&lt;comparable> '<'  expression2&lt;comparable>
  *                            | expression1&lt;comparable> '==' expression2&lt;comparable>
  *                            | expression1&lt;comparable> '>'  expression2&lt;comparable> 
  *    condition             ::= expression&lt;boolean> '?' expression1 ':' expression2
  *    assignment            ::= variable '=' expression
  *    constant              ::= any 
  *    variable              ::= [a-zA-Z]+ 
  *    literal               ::= 'current' | 'next' | 'set' | 'reset' | 'reverse'
  * </pre>
  * 
  * Precedence is the responsibility of the object that constructs the
  * syntax-tree (here, the {@link dk.rode.thesis.interpreter.Main}
  * class). Cyclic expression references based on identity is allowed
  * at the discretion of the developer at the cost of a potentially
  * very deep call stack. Certain expressions methods will not function
  * properly in case of cyclic references, though! Furthermore, in 
  * general, there is <b>no guarantee that the interpretation of a 
  * given expression and/or syntax tree will terminate</b>! <p>
  * 
  * An example of a syntax-tree for manipulating a single 
  * {@code Sequence}, say <i>foo</i>, could be:
  * <pre> 
  *   1     next[20], 
  *   2     y = current, 
  *   3     reverse 
  *   4       ? next[10], x = current, x < 11 
  *   5                                  ? x 
  *   6                                  : break[y] 
  *   7       : exit
  * </pre>
  * Which reads: <ol><p>
  * 
  * <li> call {@code next()} 20 times on <i>foo</i>; 
  * <li> set the value of <i>y</i> to the current value of <i>foo</i>;
  * <li> try and reverse <i>foo</i>
  * <li> if reversed (line 3), call {@code next()} 10 times on <i>foo</i>,
  *      set <i>x</i> to the current value of <i>foo</i>, and
  *      test if <i>x</i> is smaller than 11; 
  * <li> if <i>x</i> is smaller, return x;
  * <li> if <i>x</i> is equal or larger, break out of the current 
  *      evaluation, and transfer control to <i>y</i>, which will
  *      be evaluated and its result returned;     
  * <li> if not reversed (line 3), exit evaluation with no value (null).
  * </ol><p>     
  * 
  * If <i>foo</i> is a 
  * {@link dk.rode.thesis.meta.model.ReversibleSequence 
  * reversible} integer sequence having the initial value 1, the 
  * evaluation exits by evaluating <i>y</i> in line 6, which yields
  * 21 because of the assignment in line 2; <i>y</i> could be <b>any</b>
  * kind of expression, not just simple values. If the initial value is
  * 0, the evaluation stops normally and returns 10, which is the 
  * value of <i>x</i> in line 5. Assuming <i>foo</i> is not 
  * reversible in line 3, null will be returned from line 7. <p>
  *   
  * <a name="uml"></a><b>UML Class Diagram</b>:<br>
  * <img src="doc-files/Interpreter.png" border="0"><p>
  * 
  * <b>Implementation notes</b>:<br>
  * The {@link dk.rode.thesis.interpreter.Expression} interface
  * is generic, having a single type parameter, {@code E}, that 
  * specifies the type of value returned by the evaluation. An 
  * implication of this is that a given syntax-tree can only evaluate
  * expressions evaluating the same (sub-) type of values. As {@code Sequence}
  * objects are the target of the terminal expressions, the type
  * parameter of the given sequence is often what determines
  * the value of {@code E}. The Gamma et al. examples all operate
  * on a single type only, e.g. boolean or string, and the type
  * is fixed! This is not the case here: any {@code Sequence}
  * can be used, regardless of which values it deliver. Furthermore,
  * some expressions can operate on sub-types of expression values,
  * i.e. {@code <? extends E>}, as is the case with 
  * {@link dk.rode.thesis.interpreter.AssignmentExpression}
  * and {@link dk.rode.thesis.interpreter.FlowExpression}, 
  * for example. When collaborating with {@code Context} and 
  * {@code AssignmentExpression}, 
  * {@link dk.rode.thesis.interpreter.VariableExpression}
  * type allows polymorphic values as well. <p>
  * 
  * More so, the type parameter provide type safety and makes certain things
  * easier. There is, for example, no need for a specific <i>boolean 
  * expression</i> class, since any type matching {@code Expression<Boolean>}
  * will do, and we do not have to cast the result of the evaluation
  * to the desired type. However, certain things become much more 
  * difficult with type parameters, especially in Java where type erasure
  * ensures that the type information is not available at runtime. For
  * example, <i>replacement of expressions</i> as described in the 
  * Gamma et al. C++ example [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.252-254] is very tricky, if 
  * not impossible, when expressions use different type parameters, as 
  * for example the {@link dk.rode.thesis.interpreter.ConditionalExpression}
  * class. In {@code ConditionalExpression}, the condition is expressed 
  * as {@code Expression<Boolean>}, while the two operands are 
  * expressed as {@code Expression<? extends E>}. Clearly, the two type 
  * parameters are incompatible unless {@code E} is a {@code Boolean}  
  * or a sub-type of it; but no such sub-class exist since {@code Boolean}
  * is declared final! Now consider a {@code replace} method declared 
  * in {@code Expression<E>} as: 
  * <pre> 
  *   public Expression&lt;E> replace(String name, Expression&lt;E> exp);
  * </pre>
  * How should {@code ConditionalExpression} implement this? An attempt
  * could be the following:
  * <pre> 
  * 1   public Expression&lt;E> replace(String name, Expression&lt;E> exp) {
  * 2     return new ConditionalExpression&lt;E>(
  * 3       this.condition.replace(name, exp),       // Expression&lt;Boolean>
  * 4       this.first.replace(name, exp),           // Expression&lt;? extends E>
  * 5       this.second.replace(name, exp)           // Expression&lt;? extends E>
  * 6     );
  * 7   }    
  * </pre>
  * But, lines 3-5 will generate compiler errors because of incompatible
  * types. Replacing {@code E} with {@code <? extends E>} in the
  * method signature will not help. Pure will cards cannot be utilised either 
  * as the type information is vital for the condition check to work <b>and</b> 
  * for the creating of a new expression eventually when an actual replacement
  * occurs. Hence, the {@code Expression} type do <b>not</b> have a such 
  * a {@code replace} method. <p>
  * 
  * Of cause, one could simply choose not to use type parameters and
  * store {@code Class} objects in expression with the type information,
  * but then compile-time type safety is gone. <p>
  * 
  * Here, we provide a mixture of the two approaches: the {@code Expression} 
  * interface uses type parameters to enforce compile-time type safety 
  * for expressions as well the sequences they operate on at the expense
  * of replacement of existing expressions, but we also provide the 
  * {@link dk.rode.thesis.interpreter.TypedExpression} interface
  * that can associate the type of the expression values with the expression
  * it self at runtime. <p> 
  * 
  * An example of explicit usage of type parameters is the 
  * {@link dk.rode.thesis.interpreter.NextExpression} class. It 
  * can use any expression of the type {@code Expression<? extends Number>}
  * to deliver the number of times {@code next()} should be invoked on a
  * given sequence. An example where the type of the expression value must
  * be known is the {@link dk.rode.thesis.interpreter.BreakExpression} 
  * class. It aids the {@link dk.rode.thesis.interpreter.Interpreter} 
  * class to transfer control to any expression having a <i>sub-type</i> of 
  * the original expression value type. Furthermore, the 
  * {@link dk.rode.thesis.interpreter.TypedExpressionDecorator}
  * class can be used to associate <i>any</i> expression with type of the 
  * expression value to which it evaluates. <p>  
  * 
  * In line with the Gamma et al. definition of an <i>interpreter</i> (method) 
  * [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.255], {@code Expression} declares a 
  * {@link dk.rode.thesis.interpreter.Expression#copy()},
  * a {@link dk.rode.thesis.interpreter.Expression#asSymbol(Context)},
  * and a {@link dk.rode.thesis.interpreter.Expression#contains(Expression)}  
  * method as well, the latter quite similar in nature to the replacement
  * functionality described by Gamma et al. <p>
  * 
  * The traversal of the expression syntax tree is handled by 
  * the expressions themselves, though with some aid from
  * the {@link dk.rode.thesis.interpreter.Context} class. 
  * External visitors could easily be constructed to traverse
  * the expression graph as well, as expressions expose their
  * composite structure in form of their expression 
  * {@link dk.rode.thesis.interpreter.Expression#operands()
  * operands}. The {@linkplain dk.rode.thesis.builder Builder}
  * pattern illustrates such an external traversal of 
  * expressions. <p>
  *  
  * There are many similarities between terminal expressions and
  * the {@link dk.rode.thesis.command.Command commands} 
  * used in the {@linkplain dk.rode.thesis.command Command} pattern. 
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a> 
  */
@Pattern(name         = "Interpreter", 
         scope        = Scope.Class,
         purpose      = Purpose.Behavioural,
         participants = {"AbstractExpression", 
                         "TerminalExpression", 
                         "NonTerminalExpression", 
                         "Context",
                         "Client"})
package dk.rode.thesis.interpreter;

import dk.rode.thesis.meta.Pattern;
import dk.rode.thesis.meta.Purpose;
import dk.rode.thesis.meta.Scope;
