/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.memento;

import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.reflect.CallerClass;

/**
  * A <i>guarded sequence memento</i> offer the same functionality
  * as {@link SequenceMemento}, but enforce access rules to ensure
  * that only a (memorizable) sequence can set or get the memento
  * state. <p>
  * 
  * Memento state is defined as data that can be set or acquired using
  * the the state methods {@link #setSequence(Sequence)}, {@link #getSequence()}, 
  * and {@link #current()}, respectively. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * All memento functionality has to be declared public if it must
  * be used outside this package. It cannot have "two interfaces" in 
  * Java, a narrow (public) and a wide (public and private, using friends 
  * to access private part), as described by Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.285-287]. 
  * This implementation therefore divides the memento functionality into 
  * two interfaces, namely {@link GuardedSequenceMemento} (private part of 
  * wide) and {@link MemorizableSequence} (narrow, public). However, the 
  * state methods {@link #setSequence(Sequence)}, {@link #getSequence()}, 
  * and {@link #current()} still have to be public. To enforce an access check 
  * at runtime, this class uses a {@link CallerClass} to identify the caller 
  * of the state methods, and only allows calls from valid contexts (memorizable 
  * sequences). The {@code GuardedSequenceMemento} class is thus a  
  * <a href="http://www.rode.dk/thesis/ComparativeEvaluation.htm#section-7.1.2.1" target="thesis"><i>Guarded Type</i></a>. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by applicable memorizable sequences.
  */
public class GuardedSequenceMemento<E> extends SequenceMemento<E> {

    /**
      * No-arg constructor. <p>
      * 
      * The sequence representing the state to memorise must
      * be set explicitly using the {@link #setSequence(Sequence)}
      * method.
      */
    public GuardedSequenceMemento() {
        super();
    }
   
    /**
      * Constructor.
      * 
      * @param  sequence The sequence representing the state to 
      *                  memorise; cannot be null. This memento does
      *                  not keep a reference to {@code sequence},
      *                  but copies it.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public GuardedSequenceMemento(final Sequence<E> sequence) {
        super(sequence);
    }

    /**
      * {@inheritDoc}
      * 
      * @throws IllegalStateException {@inheritDoc}
      * @throws UnsupportedOperationException If the caller is not a sequence
      *         of the type stored internally in this memento.
      */
    @Override 
    public final Sequence<E> getSequence() {
        // This will in any case cause a copy...
        Sequence<E> sequence = super.getSequence();
        // Check that the caller class corresponds to the class of the sequence 
        // stored. This ensures that only a (memorizable) sequence of the same
        // type can invoke this method...
        CallerClass caller = new CallerClass().getNonClassCaller(); // may be called from "current" 
        if (!sequence.getClass().equals(caller.callerClass)) {
            LogFactory.getLog(this).error("Caller does not have access: ", caller);
            throw new UnsupportedOperationException("Method not supported for caller: " + caller);
        }
        return sequence;
    }

    /**
      * {@inheritDoc}
      * 
      * @throws IllegalStateException {@inheritDoc}
      * @throws UnsupportedOperationException If the caller is not a sequence
      *         of the type stored internally in this memento.
      */
    @Override 
    public final E current() {
        return this.getSequence().current();
    }
   
    /**
      * {@inheritDoc}
      * 
      * @throws NullPointerException {@inheritDoc}
      * @throws IllegalStateException {@inheritDoc}
      * @throws UnsupportedOperationException If {@code sequence} does not have the 
      *         same class as the caller of this method.
      */
    @Override 
    public final void setSequence(Sequence<E> sequence) {
        // Check that the caller class corresponds to the sequence class. This 
        // ensures that only a (memorizable) sequence will invoke this method...
        CallerClass caller = new CallerClass();
        // May be called from the constructor...
        if (caller.isSuperClass()) {
            caller = caller.getCaller(); 
        }
        caller = caller.getNonAssignableCaller(); 
        if (!sequence.getClass().equals(caller.callerClass)) {
            LogFactory.getLog(this).error("Caller does not have access: ", caller);
            throw new UnsupportedOperationException("Method not supported for caller: " + caller);
        }
        super.setSequence(sequence);
    }
    
}
