/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.memento;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import dk.rode.thesis.facade.MathFacade;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.EnglishAlphabetSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.test.Testable;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.singleton.DanishAlphabetSequence;

/**
  * <b>Memento</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Caretaker")
public class Main implements Testable {

    /**
      * Executes the <b>Memento</b> tests.
      * 
      * @param args The arguments to supply to {@link #test(Log, Arguments)},
      *             if any; can be null.
      */
    public static void main(String[] args) {
        new Main().test(LogFactory.getLog(), new Arguments(args));
    }
    
    /**
      * No-arg constructor.
      */
    public Main() {
        super();
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) {
        out.heading("Memento tests...");

        if (arguments.size() > 0) {
            out.println("Supplied arguments: ", arguments);
        }
               
        boolean ok = true;

        // True to enable class related log output (verbose)...
        boolean log = arguments.getArgumentValue("log", true, Boolean.class);
        LogFactory.getLog(MemorizableEnglishAlphabetSequence.class).setActive(log);
        LogFactory.getLog(RangeSequence.class).setActive(log);
        LogFactory.getLog(GuardedSequenceMemento.class).setActive(log);
        
        //
        // Create sequences
        //
        
        List<Sequence<String>> sequences = new ArrayList<Sequence<String>>();

        sequences.add(new EnglishAlphabetSequence());
        MemorizableEnglishAlphabetSequence meas = new MemorizableEnglishAlphabetSequence();

        
        sequences.add(meas);
        sequences.add(DanishAlphabetSequence.Instance); 

        RangeSequence rs = new RangeSequence(100, 110);
        
        // Print instances...
        out.heading("Created sequences");
        for (int i = 0; i < sequences.size(); i++) {
            Sequence<String> sequence = sequences.get(i);
            out.println("Sequence ", i, ": ", sequence);
        }
        out.println("Sequence ", sequences.size(), ": ", rs);
        
        //
        // Test mementos
        //
        
        // true = create memento, false = use memento
        boolean create = true, create1 = true;
        // Map storing the mementos
        Map<MemorizableSequence<String>, 
            SequenceMemento<String>> mementos = createMap(); 
        Map<MemorizableSequence<Integer>, 
            SequenceMemento<Integer>> rsMementos = createMap(); 
        
        out.heading("Memento usage");
        for (int i = 0; i < 30; i++) {
            out.println("Iteration: ", (i + 1));
            for (int j = 0; j < sequences.size(); j++) {
                Sequence<String> sequence = sequences.get(j);
                // Advance sequence...
                String value = sequence.next();
                out.println(j, ": ", sequence, " -> ", value);
                if ((i % 10 == 0) && (sequence instanceof MemorizableSequence)) {
                    MemorizableSequence<String> ms = 
                        (MemorizableSequence<String>)sequence;
                    if (create) {
                        ok = create(out, mementos, ms) && ok;
                        create = false;
                    } else {
                        ok = use(out, mementos, ms) && ok;
                    }
                }
            }
            // Integer... verbose, I know...
            Integer value = rs.next();
            out.println(sequences.size(), ": ", rs, " -> ", value);
            if (i % 10 == 0) {
                if (create1) {
                    ok = create(out, rsMementos, rs) && ok;
                    create1 = false;
                } else {
                    ok = use(out, rsMementos, rs) && ok;
                }
            }
            
            out.println();
        }
        for (int i = 0; i < sequences.size(); i++) {
            Sequence<String> sequence = sequences.get(i);
            String value = sequence.next();
            out.println(i, ": ", sequence, " -> ", value);
        }
        out.println(sequences.size(), ": ", rs, " -> ", rs.next());

        //
        // Test illegal access
        //
        
        out.heading("Testing illegal access");
        
        for (int j = 0; j < sequences.size(); j++) {
            Sequence<String> sequence = sequences.get(j);
            ok = checkAccess(out, mementos, sequence) && ok;
        }
        ok = checkAccess(out, rsMementos, rs) && ok;
        out.println();

        //
        // Test serialization
        //
        
        // True to enable logging from range sequence as well...
        LogFactory.getLog(rs).setActive(log);
        
        out.heading("Serialization test of: ", rs);
        
        // May wrap...
        int n = MathFacade.getInstance().getRandomNumber(12);
        out.println("Advancing sequence ", n, " times (based on random sequence)");
        for (int i = 0; i < n; i++) {
            rs.next();
        }
        
        // Serialization...
        try {
            ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream("memento.tmp"));
            oos.writeObject(rs);
            oos.close();
            out.println();
            out.println("Serialized:   ", rs);
            
        } catch (Exception e) {
            out.error("Serialization error").error(e);
            ok = false;
        }

        SequenceMemento<Integer> memento = rs.getMemorizableState();
        out.println("Serialized:   ", memento);
        
        // Deserialization...
        RangeSequence deserialized = null;
        try {
            ObjectInputStream ios = new ObjectInputStream(new FileInputStream("memento.tmp"));
            deserialized = (RangeSequence)ios.readObject();
            ios.close();
            out.println("Deserialized: ", deserialized);
            
        } catch (Exception e) {
            out.error("Deserialization error").error(e);
            ok = false;
        }
        
        //
        // Test possibly illegal state
        //
        
        try {
            deserialized.setMemorizableState(memento);
            out.println("Deserialized: ", memento, " -> ", deserialized);
            
        } catch (Exception e) {
            out.error("Deserialization error").error(e);
            ok = false;
        }
        
        // Test (possibly) illegal state...
        RangeSequence rrs = new RangeSequence(104, 115);
        out.println();
        out.println("Testing memento on different sequence: ", memento, " -> ", rrs);
        try {
            rrs.setMemorizableState(memento);
            out.println("Ok!");
        } catch (MemorizableException me) {
            out.println("Fail (as expected): ", me);
        }
        
        out.heading("Memento tests ", ok ? "succeeded!" : "failed!");
        
        return ok;
    }

    /**
      * Creates a new map to store (memorizable sequence, memento) pairs.
      * 
      * @param  <E> The type of values delivered by applicable sequences.
      * @return A new map to use; never null.
      */
    private final static <E> Map<MemorizableSequence<E>,SequenceMemento<E>> createMap() {
        return new HashMap<MemorizableSequence<E>,SequenceMemento<E>>();
    }
    
    /**
      * Creates a new memento acquired from {@code sequence}.
      * 
      * @param  <E>      The type of values delivered by {@code sequence}.
      * @param  out      The log to use; never null.
      * @param  mementos The map to store the created memento; never null.
      * @param  sequence The sequence to create the memento; never null.
      * @return True upon success, false if not.
      */
    private final static <E> boolean create(Log out, Map<MemorizableSequence<E>,SequenceMemento<E>> mementos, MemorizableSequence<E> sequence) {
        SequenceMemento<E> sm = sequence.getMemorizableState();
        out.println("Created memento: ", sequence, " -> ", sm);
        mementos.put(sequence, sm);
        if ((!(sm instanceof GuardedSequenceMemento)) && (!sequence.current().equals(sm.getSequence().current()))) {
            out.error("Current value different for between sequence and created memento: ", sequence.current(), " != ", sm.getSequence().current());
            return false;
        }
        return true;
    }

    /**
      * Updates {@code sequence} using the relevant memento stored in
      * {@code mementos}.
      * 
      * @param  <E>      The type of values delivered by {@code sequence}.
      * @param  out      The log to use; never null.
      * @param  mementos The map to storing the created memento; never null.
      * @param  sequence The sequence to be updated; never null.
      * @return True upon success, false if not.
      */
    private final static <E> boolean use(Log out, Map<MemorizableSequence<E>,SequenceMemento<E>> mementos, MemorizableSequence<E> sequence) {
        SequenceMemento<E> sm = mementos.get(sequence);
        try {
            sequence.setMemorizableState(sm);
            out.println("Used memento: ", sm, " -> ", sequence);
            if ((!(sm instanceof GuardedSequenceMemento)) && (!sequence.current().equals(sm.getSequence().current()))) {
                out.error("Current value different for between updated sequence and memento: ", sequence.current(), " != ", sm.getSequence().current());
                return false;
            }
            return true;
            
        } catch (MemorizableException me) {
            out.error("Error occurred while using memento").error(me);
            return false;
        }
    }

    /**
      * Checks access rules to guarded mementos. 
      * 
      * @param  <E>      The type of values delivered by {@code sequence}.
      * @param  out      The log to use; never null.
      * @param  mementos The map to storing the created mementos; never null.
      * @param  sequence The sequence identifying the memento to use; never null.
      * @return True upon success, false if not.
      */
    private final static <E> boolean checkAccess(Log out, Map<MemorizableSequence<E>,SequenceMemento<E>> mementos, Sequence<E> sequence) {
        out.println("Testing: ", sequence);
        SequenceMemento<E> sm = mementos.get(sequence);
        // Not memorizable...
        if (sm == null) {
            return true;
        }
        boolean guarded = (sm instanceof GuardedSequenceMemento);
        try {
            out.println("Testing \"getSequence()\"...");
            Sequence<E> state = sm.getSequence();
            if (guarded) {
                out.error("Could acquire memento state from guarded memento: ", sm, " -> ", state);
                return false;
            } else {
                out.println("Testing ok - access");
            }
        } catch (Exception e) {
            if (!guarded) {
                out.error("Error occurred while using memento").error(e);
                return false;
            } else {
                out.println("Testing ok - no access");
            }
        }
        try {
            out.println("Testing \"current()\"...");
            E value = sm.current();
            if (guarded) {
                out.error("Could acquire memento state (value) from guarded memento: ", sm, " -> ", value);
                return false;
            } else {
                out.println("Testing ok - access");
            }
        } catch (Exception e) {
            if (!guarded) {
                out.error("Error occurred while using memento").error(e);
                return false;
            } else {
                out.println("Testing ok - no access");
            }   
        }
        try {
            // Test creation...
            out.println("Testing \"setSequence(Sequence)\"...");
            sm = guarded ? new GuardedSequenceMemento<E>() : new SequenceMemento<E>();
            sm.setSequence(sequence);
            if (guarded) {
                out.error("Could set memento state on guarded memento: ", sm, " -> ", sequence);
                return false;
            } else {
                out.println("Testing ok - access");
            }
        } catch (Exception e) {
            if (!guarded) {
                out.error("Error occurred while using memento").error(e);
                return false;
            } else {
                out.println("Testing ok - no access");
            }
        }
        return true;
    }
    
}
