/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.memento;

import static dk.rode.thesis.meta.util.Const.INFINITE;
import static dk.rode.thesis.meta.util.Const.SERIAL_VERSION_UID;
import static dk.rode.thesis.meta.util.Throwables.toThrowableType;

import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.AbstractSequence;
import dk.rode.thesis.meta.model.ReversibleSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.util.Const;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>range sequence</i> represents a positive {@code Integer} value
  * that is increased with each call to {@link #next()}, which
  * can be {@link #reverse() reversed} to deliver previous values 
  * as well unless unbounded. <p>
  * 
  * A start value can be supplied (default is zero), and an integer
  * sequence can also be found by supplying an end value. Default
  * range sequences are unbounded. <p>  
  * 
  * It can have its state restored either by using the Java Serialization
  * mechanism, or by explicitly using {@link SequenceMemento mementos}. <p>
  * 
  * A range sequence may be {@link Sequence#bounded() bounded}, 
  * but is {@link Sequence#consistent() consistent} and 
  * {@link Sequence#unique() unique}. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * The mementos created are {@link GuardedSequenceMemento} instances. However,
  * regular mementos can still be set. This could easily be remedied by an
  * {@code instanceof} check, though. 
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Originator") 
public class RangeSequence 
     extends AbstractSequence<Integer> 
  implements ReversibleSequence<Integer>, Serializable, MemorizableSequence<Integer> {

    /**
      * The serial version. 
      */
    final static long serialVersionUID = SERIAL_VERSION_UID;
    /**
      * Start number of this sequence. <p>
      * 
      * Invariant: start >= 0
      * 
      * @serial
      */
    private int start;
    /**
      * End number of this sequence. <p>
      *   
      * Invariant: end >= {@link #start} >= 0 unless {@linkplain Const#INFINITE infinite}.
      * 
      * @serial
      */
    private int end;
    /**
      * The current sequence number. <p>
      * 
      * Invariant: {@link #start} <= sequence <= {@link #end}
      * 
      * @serial
      */
    private int sequence;
    /**
      * True to return the next higher sequence value ("forward"), 
      * false to return previous values ("reverse")
      * 
      * @serial
      */
    private boolean direction;
    
    /**
      * No-arg constructor, which creates this sequence as unbounded, 
      * i.e. infinite, and starting at 0.
      */
    public RangeSequence() {
        this(0);
    }

    /**
      * Constructor, which creates this sequence as unbounded, 
      * i.e. infinite, starting at {@code start}.
      * 
      * @param  start The start value of this sequence; must be >= 0.
      * @throws IllegalArgumentException If {@code start} is negative.              
      */
    public RangeSequence(int start) {
        this(start, INFINITE);
    }
    
    /**
      * Constructor, which creates this sequence as bounded
      * unless {@code end} is {@linkplain Const#INFINITE infinite},
      * starting at {@code start}. <p>
      * 
      * @param  start The start value of this sequence; must be >= 0.
      * @param  end   The end value of this sequence; must be >= {@code start}
      *               unless infinite. 
      * @throws IllegalArgumentException If {@code start} is negative,
      *         or if {@code end} is not infinite and smaller than {@code start}.
      */
    public RangeSequence(int start, int end) {
        super(false);
        validate(this.start = start, this.end = end, this.sequence = this.start);
        this.direction = true; // forward
    }
    
    /**
      * Copy constructor.
      * 
      * @param  sequence The sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public RangeSequence(RangeSequence sequence) {
        super(sequence);
        this.start = sequence.start;
        this.end = sequence.end;
        this.sequence = sequence.sequence;
        this.direction = sequence.direction;
    }
    
    /**
      * 
      */
    public RangeSequence copy() {
        return new RangeSequence(this); 
    }

    /**
      * Validates the start and end values supplied 
      * as {@code start} and {@code end}, respectively. <p>
      * 
      * Furthermore, {@code sequence} is validated to make sure 
      * it is within the {@code start} and {@code end} bounds. <p>
      * 
      * Returns silently upon success, or throws an exception
      * otherwise.
      *   
      * @param  start    The start value; must be >= 0.
      * @param  end      The end value; must be >= {@code start} if not
      *                  {@linkplain Const#INFINITE infinite}.
      * @param  sequence A current sequence value to validate; must be 
      *                  within {@code start} and {@code end}, where 
      *                  {@code end} may be infinite.               
      * @throws IllegalArgumentException If {@code start} is negative;
      *         or if {@code end} is not infinite but smaller than {@code start};
      *         or if {@code sequence} is not within the {@code start} and 
      *         {@code end} bounds.
      * @see    #RangeSequence(int, int)
      * @see    #readObject(ObjectInputStream)         
      */
    private static void validate(int start, int end, int sequence) {
        if (start < 0) {
            throw new IllegalArgumentException("Start cannot be negative");
        } else if ((end != INFINITE) && (end < start)) {
            throw new IllegalArgumentException("End cannot be smaller than start");
        } else if ((start > sequence) || ((end != INFINITE) && (end < sequence))) {
            throw new IllegalArgumentException("Sequence value outside bounds");
        }
    }
    
    /**
      * If an end value was supplied at construction time, this
      * sequence is bounded, and true is returned. Otherwise false.
      *  
      * @return True if bounded, false if not. 
      */
    public boolean bounded() {
        return (this.end != INFINITE);
    }

    /**
      * Returns true.
      *  
      * @return True.
      */
    public boolean consistent() {
        return true;
    }

    /**
      * Returns true.
      *  
      * @return True. 
      */
    public boolean unique() {
        return true;
    }
    
    /**
      * 
      */
    public Integer current() {
        return this.sequence;
    }

    /**
      * 
      */
    public Integer next() {
        if (this.state != State.NORMAL) {
            this.state = State.NORMAL;
        }
        if (!this.bounded()) {
            this.sequence++;
            
        // Next...
        } else if (this.direction) {
            if (this.sequence + 1 > this.end) {
                this.sequence = this.start;
                this.state = State.RESTART;
                
            } else {
                this.sequence++;
            }
            
        // Previous...
        } else if (this.sequence - 1 < this.start) {
            this.sequence = this.end;
            this.state = State.RESTART;
        } else {
            this.sequence--;
        }
        
        return this.sequence;
    }

    /**
      * @throws IllegalStateException {@inheritDoc}
      */
    public Integer reverse() {
        if (!this.bounded()) {
            throw new IllegalStateException("Sequence is unbounded");
        }
        this.direction = !this.direction;
        LogFactory.getLog(this).println("Reversed: ", this.toString(SequencePolicy.IDENTITY));
        return this.sequence;
    }
    
    /**
      * Returns true if this range sequence is {@link #bounded() bounded},
      * false otherwise.
      * 
      * @return True if bounded, ergo reversible, false if not.
      */
    public boolean reversible() {
        return this.bounded();
    }
    
    /**
      * 
      */
    public void reset() {
        super.reset();
        this.sequence = this.start;
    }

    /**
      * 
      */
    public SequenceMemento<Integer> getMemorizableState() {
        return new GuardedSequenceMemento<Integer>(this); // check caller when updating!
    }

    /**
      * 
      */
    public void setMemorizableState(SequenceMemento<Integer> memento) throws MemorizableException {
        try {
            RangeSequence sequence = (RangeSequence)memento.getSequence();
            
            // Must have a matching lower bound...
            if (sequence.sequence < this.start) {
                throw new MemorizableException(memento);
                
            // If bounded, the sequence value cannot be larger than "end"...
            } else if (this.bounded()) {
                if (sequence.sequence > this.end) {
                    throw new MemorizableException(memento);
                }
                
            // If unbounded, the sequence cannot be reversed...
            } else if (!sequence.direction) {
                throw new MemorizableException(memento);
            }
            this.sequence = sequence.sequence;
            this.state = sequence.state;
            this.direction = sequence.direction;
            
            LogFactory.getLog(this).println("Updated ", this, " using ", memento);
            
        // To maintain memorizable sequence semantics...
        } catch (Exception e) {
            throw toThrowableType(e, MemorizableException.class);
        }
    }

    /**
      * Serializes this memorizable range sequence. <p>
      * 
      * The internal state is also serialized.
      * 
      * @param      out The output stream to write to; never null.
      * @throws     IOException In case the serialization fails.
      * 
      * @serialData Writes the {@link #start} and {@link #end} bounds to
      *             {@code out}, followed by the current {@link #sequence}
      *             number and the {@link #direction}. Last, the internal
      *             sequence {@link #state} inherited from the super
      *             class is written. 
      */
    private void writeObject(ObjectOutputStream out) throws IOException {
        out.writeInt(this.start);
        out.writeInt(this.end);
        out.writeInt(this.sequence);
        out.writeBoolean(this.direction);
        // Include state from super...
        out.writeObject(this.state);
        LogFactory.getLog(this).println("Serialised ", this);
    }

    /**
      * Deserializes this memorizable sequence. <p>
      * 
      * To protect against malicious attacks, all read attributes
      * are validated.
      * 
      * @param      in The input stream to read from; never null.
      * @throws     IOException In case the deserialization fails.
      * @throws     ClassNotFoundException If this class could not be loaded.
      * @see        #validate(int, int, int)
      *  
      * @serialData Reads the {@link #start} and {@link #end} bounds from
      *             {@code in} as well as the current {@link #sequence} value 
      *             and immediately validates them; upon success, the current 
      *             {@link #direction} is read and lastly, the internal sequence 
      *             {@link #state} inherited from the super class is read. 
      */
    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        // Read start and end and enforce invariants...
        try {
            validate(this.start = in.readInt(), this.end = in.readInt(), this.sequence = in.readInt());
            this.direction = in.readBoolean();
            // Include state from super...
            if ((this.state = (State)in.readObject()) == null) {
                throw new IllegalArgumentException("State cannot be null");
            }
        } catch (IllegalArgumentException e) {
            throw new InvalidObjectException(e.getMessage());
        }
        LogFactory.getLog(this).println("Deserialized ", this);
    }
    
}
