/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.log;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.concurrent.atomic.AtomicInteger;

import dk.rode.thesis.meta.util.Util;

/**
  * A <i>file log</i> logs to a given file based on the class
  * a given log is associated with, if any, or to the default
  * log represented by a single physical log file. <p>
  * 
  * The names of log files includes "todays" date. All logs are placed 
  * in the {@link Util#getLogDirectory() log} directory.
  *   
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class FileLog extends AppendableLog implements Log {
    
    /**
      * Constructor, which creates this log as active. <p>
      * 
      * The log will log line numbers if {@code lineNumber} is not null. <p>
      * 
      * All logs are placed in the {@link Util#getLogDirectory() log}
      * directory.
      * 
      * @param  append     True to append to an existing log, false to truncate.
      * @param  clazz      The class this log is associated with, if any;
      *                    can be null.
      * @param  lineNumber Line number values; can be null.                   
      * @throws NullPointerException If {@code directory} is null.
      * @throws Exception If the creation of the file log fails.
      * @see    #isActive()
      */
    FileLog(boolean append, Class<?> clazz, AtomicInteger lineNumber) throws Exception {
        super(create(append, clazz), true, clazz, true, lineNumber);
        if (append) {
            DateFormat format = new SimpleDateFormat("dd-MM-yyyy, HH:mm:ss");
            this.heading("Logging started at " + format.format(new Date()));
        }
    }
    
    /**
      * Creates a new physical log file to use for this logger. <p>
      * 
      * The name of the log file is based on the 
      * {@link AbstractLog#getPrefix(Class) prefix} used 
      * (based on {@code clazz}) and "todays" date,
      * such as {@code main.2007-08-30.log}, for example. <p>
      * 
      * All logs are placed in the {@link Util#getLogDirectory() log}
      * directory.
      * 
      * @param  append True to append to an existing log, false to truncate.
      * @param  clazz  The class the logger is associated with, if any; can
      *                be null.                   
      * @return An appendable representing the file output stream; never null.
      * @throws Exception If the creation fails.
      */
    private final static Appendable create(boolean append, Class<?> clazz) throws Exception {
        File dir = Util.getLogDirectory();
        String fileName = clazz == null ? "main" : clazz.getSimpleName().replaceAll("[\\W]+", "").toLowerCase();
        DateFormat format = new SimpleDateFormat("yyyy-MM-dd");
        File file;
        do {
            file = new File(dir, fileName + "." + format.format(new Date()) + ".log");
            
        } while (!append && file.exists());
        System.out.println((append ? "Created appending log: " : "Created truncating log: ") + file.getAbsolutePath());
        return new PrintStream(new FileOutputStream(file, append));
    }
        
}
