/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.log;

/**
  * A <b>very</b> simple logger used in this project. <p>
  * 
  * Logs are either associated with a specific class, or
  * a default global log. <p>
  * 
  * Logs are created by a {@link LogFactory} object.  
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public interface Log {
    
    /**
      * Returns the prefix of this log that will be prefixed to
      * all log messages starting at a new line.
      *  
      * @return The prefix.
      */
    public String getPrefix();
    
    /**
      * Logs each part of a heading supplied in {@code messages} 
      * with this log <i>as is</i>, no trailing new line or white space
      * inserted between heading parts, but with a single new line
      * character after the last part, underscored on the next log
      * line. <p>
      *  
      * A message that is null is logged as the string {@code null}.<p>
      * 
      * <b>Example</b> (default log): 
      * <pre>
      *   heading("This", "is ", " a ", "heading"):
      *   
      *   (newline)
      *   MAIN:xxx:123:: Thisis  a heading(newline)
      *   MAIN:xxx:123:: =================(newline)
      *   (newline)
      *   (newline)
      * </pre>
      * 
      * Where {@code xxx} is the name of the member invoking the
      * {@code heading} method and {@code 123} the line number.
      * 
      * @param  messages The message parts; can be null, corresponding to
      *                  a single new line being printed (no underscore).
      * @return This log; never null.                 
      */
    public Log heading(Object... messages);
    
    /**
      * Logs each part of an error message supplied in {@code messages} 
      * with this log <i>as is</i>, no trailing new line or white space
      * inserted between message parts, but with a single new line
      * character after the last part as well as a prefix
      * indicating the message represent an error message. <p>
      *  
      * A part that is null is logged as the string {@code null}. <p>
      * 
      * <b>Example</b> (default log): 
      * <pre>
      *   error("This", "is ", " an ", "error"):
      *   
      *   MAIN:xxx:123:: > ERROR - Thisis  an error(newline)
      * </pre>
      * 
      * Where {@code xxx} is the name of the member invoking the
      * {@code error} method and {@code 123} the line number.
      * 
      * @param  messages The message parts; can be null, corresponding to
      *                  a single new line being printed.
      * @return This log; never null.                 
      */
    public Log error(Object... messages);

    /**
      * Logs each part of a warning message supplied in {@code messages} 
      * with this log <i>as is</i>, no trailing new line or white space
      * inserted between message parts, but with a single new line
      * character after the last part as well as a prefix
      * indicating the message represent a warning message. <p>
      *  
      * A part that is null is logged as the string {@code null}. <p>
      * 
      * <b>Example</b> (default log): 
      * <pre>
      *   warn("This", "is ", " a", "warning"):
      *   
      *   MAIN:xxx:123:: > WARN - Thisis  awarning(newline)
      * </pre>
      * 
      * Where {@code xxx} is the name of the member invoking the
      * {@code warn} method and {@code 123} the line number.
      * 
      * @param  messages The message parts; can be null, corresponding to
      *                  a single new line being printed.
      * @return This log; never null.                 
      */
    public Log warn(Object... messages);
    
    /**
      * Logs each message part supplied in {@code messages} 
      * with this log <i>as is</i>, no trailing new line or 
      * white space inserted between messages. <p>
      * 
      * A part that is null is logged as the string {@code null}. <p>
      *  
      * <b>Example</b> (default log): 
      * <pre>
      *   print("This", "is ", "message #", new Integer(42), "!"):
      *   
      *   MAIN:xxx:123:: Thisis message #42!
      * </pre>
      * 
      * Where {@code xxx} is the name of the member invoking the
      * {@code print} method and {@code 123} the line number.
      * 
      * @param  messages The message parts; can be null, which will not
      *                  print anything.
      * @return This log; never null.
      */
    public Log print(Object... messages);
    
    /**
      * Logs each message part supplied in {@code messages} 
      * with this log <i>as is</i>, no trailing new line or 
      * white space inserted between messages, if and only
      * if {@code condition} is true. <p>
      * 
      * A part that is null is logged as the string {@code null}. <p>
      *  
      * <b>Example</b> (default log): 
      * <pre>
      *   i = 1;
      *   ..
      *   print(i > 0, "This", "is ", "message #", new Integer(42), "!"):
      *   
      *   MAIN:xxx:123:: Thisis message #42!
      *   
      *   ..
      *   
      *   i = 0;
      *   ..
      *   print(i > 0, "This", "is ", "message #", new Integer(42), "!"):
      *   
      *   (nothing logged)
      * </pre>
      * 
      * Where {@code xxx} is the name of the member invoking the
      * {@code print} method and {@code 123} the line number.
      * 
      * @param  condition The condition to use.
      * @param  messages  The message parts; can be null, which will not
      *                   print anything.
      * @return This log; never null.
      */
    public Log print(boolean condition, Object... messages);
    
    /**
      * Logs each message part supplied in {@code messages} 
      * with this log <i>as is</i>, no trailing new line or 
      * white space inserted between messages, but with a 
      * single new line character after the last part. <p>
      *  
      * A part that is null is logged as the string {@code null}. <p>
      * 
      * <b>Example</b> (default log): 
      * <pre>
      *   println("This", "is ", "message #", new Integer(42), "!"):
      *   
      *   MAIN:xxx:123:: Thisis message #42!(newline)
      * </pre>
      * 
      * Where {@code xxx} is the name of the member invoking the
      * {@code println} method and {@code 123} the line number.
      * 
      * @param  messages The message parts; can be null, corresponding to
      *                  a single new line being printed.
      * @return This log; never null.                 
      */
    public Log println(Object... messages);
    
    /**
      * Logs each message part supplied in {@code messages} 
      * with this log <i>as is</i>, no trailing new line or 
      * white space inserted between messages, if and only
      * if {@code condition} is true, but always with a 
      * single new line character after the last part. Hence,
      * if {@code condition} is false, the effect is 
      * equivalent to {@code println()}. <p>
      *  
      * A part that is null is logged as the string {@code null}. <p>
      * 
      * <b>Example</b> (default log): 
      * <pre>
      *   i = 1;
      *   ..
      *   println(i > 0, "This", "is ", "message #", new Integer(42), "!"):
      *   
      *   MAIN:xxx:123:: Thisis message #42!(newline)
      *   
      *   ..
      *   
      *   i = 0;
      *   ..
      *   println(i > 0, "This", "is ", "message #", new Integer(42), "!"):
      *   
      *   (newline logged only)
      * </pre>
      * Where {@code xxx} is the name of the member invoking the
      * {@code println} method and {@code 123} the line number.
      * 
      * @param  condition The condition to use.
      * @param  messages  The message parts; can be null, corresponding to
      *                   a single new line being printed.
      * @return This log; never null.                 
      */
    public Log println(boolean condition, Object... messages);
    
    /**
      * Returns true if this log is active, i.e. logs messages,
      * false if inactive.
      *  
      * @return True if active, false if inactive.
      */
    public boolean isActive();

    /**
      * If {@code active} is true, this log is set to active, i.e. to
      * log messages.
      *  
      * @param  active True to (re-)activate, false to disable.
      * @return This log; never null.
      */
    public Log setActive(boolean active);
    
    /**
      * If this log logs line numbers, the <b>current</b> line number is 
      * returned; otherwise 0.
      *  
      * @return The current line number, or null if line numbers are
      *         not logged.
      */
    public int getLineNumber();
    
}
