/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.reflect;

import static dk.rode.thesis.meta.util.Throwables.toRuntimeException;

import java.util.Iterator;

import dk.rode.thesis.meta.util.Util;

/**
  * A <i>caller class</i> represents the {@code java.lang.Class} 
  * instance of a given <i>calling context</i>. <p>
  * 
  * The calling context can be either a static initialisation block,
  * an instance initialisation block, a constructor, or a method. The
  * caller class thus represents the class declaring such a member. <p>
  * 
  * Caller classes are immutable. The actual {@code java.lang.Class} 
  * representing the calling context is available via the {@link #callerClass} 
  * attribute. The context that created a caller class instance is
  * stored in the {@link #calleeClass} attribute; such a context is
  * called the <i>immediate calling context</i> and is at stack depth 0. 
  * It corresponds to the member being called - the <i>callee</i>. <p>
  * 
  * <b>Example</b>:
  * <pre>
  *   ..
  *   public class Goo {
  *     public void goo {
  *       // Calling context at depth 2...
  *       new Foo().foo();
  *     }
  *   }
  *   ..
  *   public class Foo {
  *      public void foo {
  *         // Calling context at depth 1...
  *         new Bar().bar();
  *      }
  *   }
  *   ..
  *   public class Bar {
  *     public void bar {
  *       // Immediate calling context is the callee, i.e. the
  *       // class at depth 0 (class Bar)...
  *       CallerClass caller = new {@link #CallerClass()}; // = 1             
  *       System.out.println(caller.{@link #callerClass}); // prints "Foo" 
  *       System.out.println(caller.{@link #calleeClass}); // prints "Bar"
  *       
  *       caller = new {@link #CallerClass(int) CallerClass}(1);             
  *       System.out.println(caller.{@link #callerClass}); // prints "Foo" 
  *       System.out.println(caller.{@link #calleeClass}); // prints "Bar"
  *        
  *       caller = new {@link #CallerClass(int) CallerClass}(2);             
  *       System.out.println(caller.{@link #callerClass}); // prints "Goo" 
  *       System.out.println(caller.{@link #calleeClass}); // prints "Bar" 
  *     }
  *   }
  *   ..
  * </pre>
  * 
  * Notice that the no-arg constructor does <b>not</b> identify the 
  * immediate calling context as the {@link #callerClass}, but the 
  * calling context at stack depth 1. The immediate calling context 
  * is always available via the {@link #calleeClass} method. <p>
  * 
  * An {@link IdentifiedStackTraceElement#offset offset} can be supplied
  * at {@link #CallerClass(int) construction} time to specify a stack 
  * depth to identify the calling context. The offset is relative to
  * the immediate calling context. <p>
  * 
  * Stack traces are local to the thread in question.
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class CallerClass extends IdentifiedStackTraceElement implements Iterable<CallerClass> {

    /**
      * A <i>caller iterator</i> is used to iterator over {@code CallerClass} 
      * instances, or sub-classes of it. <p>
      * 
      * <b>Implementation notes</b>:<br>
      * This is an application of the {@linkplain dk.rode.thesis.iterator Iterator}
      * pattern specialised for sub-classes of {@link CallerClass}. The 
      * {@link #setNext()} method can on a fine day be considered an application 
      * of the {@linkplain dk.rode.thesis.factorymethod Factory Method} pattern.
      * 
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      *
      * @param <C> The specific (sub-) type of the caller class to iterate over.
      */
    protected static abstract class CallerIterator<C extends CallerClass> implements Iterator<C> {
        
        /**
          * The current caller set by {@link #setNext()}. <p>
          * 
          * Never null.
          */
        protected C caller;
        
        /**
          * Constructor.
          * 
          * @param  caller The caller to start from; can be null.
          */
        CallerIterator(C caller) {
            this.caller = caller;
        }
        
        /**
          * Returns true if additional callers are in the call-stack,
          * false if not.
          * 
          * @return True if more callers, false if not.
          */
        public boolean hasNext() {
            return (this.caller.offset + 1 < this.caller.stack.length);
        }
        
        /**
          * Returns the next caller.
          * 
          * @return The next caller; never null.
          */
        public C next() {
            this.setNext();
            return this.caller;
        }
        
        /**
          * Always throws an {@code UnsupportedOperationException}.
          *  
          * @throws UnsupportedOperationException Always thrown. 
          */
        public void remove() {
            throw new UnsupportedOperationException();
        }
        
        /**
          * Sets the {@link #caller next} caller class, or sub-class or it, to 
          * be retrieved by the next call to {@link #next()}.  
          */
        abstract void setNext();
    }
  
    /**
      * The <i>caller class</i> represents the class of the <i>calling
      * context</i>, relative to the {@link IdentifiedStackTraceElement#offset
      * offset} supplied at construction time. <p>
      * 
      * The caller class is the "other class" to test when methods 
      * like {@link #isAssignableToThisClass()} is invoked,
      * where {@link #calleeClass} represents "this class". <p>
      * 
      * Never null. 
      */
    public final Class<?> callerClass;
    /**
      * The <i>callee class</i> represents the class that created
      * this caller class instance, i.e. the class of the <i>immediate 
      * calling context</i>. <p>
      * 
      * The callee class is "this class" when methods like 
      * {@link #isAssignableToThisClass()} is invoked, where
      * {@link #callerClass} represents the other class to test. <p>
      *  
      * Never null.
      */
    public final Class<?> calleeClass;
    
    /**
      * No-arg constructor. <p>
      * 
      * The caller {@link #callerClass class} is the class of the calling 
      * context at stack depth 1 from the {@link #calleeClass class} of 
      * the immediate calling context, i.e. the calling context of the 
      * immediate calling context. <p> 
      * 
      * <b>Example</b>:
      * <pre>
      *   ..
      *   public void foo {
      *     this.bar();
      *   }
      * 
      *   public void bar {
      *     // Represents the class of foo, *not* of bar
      *     CallerClass caller = new CallerClass(); 
      *   }
      *   ..
      * </pre>
      */  
    public CallerClass() {
        // 0 is the depth of this class and +1 for this constructor...
        this(0, 1);
    }
    
    /**
      * Constructor. <p>
      * 
      * The caller {@link #callerClass class} is the class of the calling 
      * context at stack depth {@code offset} from the {@link #calleeClass 
      * class} of the immediate calling context. <p> 
      *  
      * Invoking this constructor with a value of zero corresponds to invoking 
      * the no-arg constructor.
      *   
      * @param  offset The stack offset; must be within bounds.
      * @throws IndexOutOfBoundsException If {@code offset} is illegal, i.e., does not
      *         represent a valid stack trace element. 
      */
    public CallerClass(int offset) {
        // 0 is the depth of this class and +1 for this constructor...
        this(0, offset + 1);
    }
    
    /**
      * Constructor. <p>
      * 
      * The caller {@link #callerClass class} is the class of the calling 
      * context at stack depth {@code offset} from the {@link #calleeClass 
      * class} of the immediate calling context. <p> 
      *  
      * Invoking this constructor with zero values corresponds to invoking 
      * the no-arg constructor.
      *   
      * @param  depth  The increase in stack depth added by this class or
      *                by sub-classes. The depth of this class is zero,
      *                the depth of an immediate sub-class is one, and
      *                so forth.
      * @param  offset The stack offset; must be within bounds.
      * @throws IndexOutOfBoundsException If {@code offset} is illegal, i.e., does not
      *         represent a valid stack trace element. 
      */
    protected CallerClass(int depth, int offset) {
        // +1 for this constructor...
        super(offset + 1);
        // Fetch specified caller class...
        this.callerClass = getClass(this.element);
        // Fetch the class of the calling context that
        // created this caller class. We add 3 to the depth 
        // because we know that the super-class adds three
        // stack frames!
        this.calleeClass = getClass(this.stack[depth + 3]);
    }
    
    /**
      * Constructor, which creates this caller using an
      * existing stack trace supplied from {@code caller}, 
      * but using a perhaps different offset supplied as 
      * {@code offset}. <p>
      * 
      * @param  caller The caller to supply the stack trace; 
      *                cannot be null.
      * @param  offset The stack offset; must be within bounds.
      * @throws NullPointerException If {@code caller} is null.
      * @throws IndexOutOfBoundsException If {@code offset} is illegal, i.e., 
      *         does not represent a valid stack trace element. 
      */
    protected CallerClass(CallerClass caller, int offset) {
        super(caller, offset);
        // Fetch specified caller class...
        this.callerClass = getClass(this.element);
        // The class of the calling context that
        // created this caller class have not changed,
        // so we simply reuse it...
        this.calleeClass = caller.calleeClass;
    }

    /**
      * Copy constructor. 
      * 
      * @param  caller The caller class to copy; cannot be null.
      * @throws NullPointerException If {@code caller} is null.
      */
    public CallerClass(CallerClass caller) {
        this(caller, caller.offset);
    }

    /**
      * Returns the class associated with the stack trace element
      * supplied as {@code element}.
      *   
      * @param  element The stack trace element; cannot be null.
      * @return The class; never null.
      * @throws NullPointerException If {@code element} is null.
      */
    protected final static Class<?> getClass(StackTraceElement element) {
        try {
            return Class.forName(element.getClassName());
        } catch (Throwable throwable) {
            // Should never happen...
            throw toRuntimeException(throwable);
        }
    }
    
    /**
      * Returns true if the {@link #callerClass class} of the calling context 
      * is defined in the same package as the {@link #calleeClass class}
      * representing the immediate calling context, false if not. <p> 
      * 
      * Example: 
      * <pre>
      *   ..
      *   
      *   package foo;
      *   
      *   public class Foo {
      *     public void foo() {
      *       this.foofoo();
      *     }
      *   
      *     public void foofoo() {
      *       CallerClass caller = new CallerClass();
      *       System.out.println(caller.isThisPackage());
      *     }
      *   }
      *   
      *   package bar;
      *   
      *   public class Bar {
      *     public void bar(foo.Foo foo) {
      *       foo.foofoo();
      *     }
      *   }
      *   
      *   ..
      *   
      *   Foo foo = new foo.Foo();
      *   Bar bar = new bar.Bar();
      *   
      *   foo.foo();    // prints "true"
      *   bar.bar(foo); // prints "false"   
      *   
      *   ..
      * </pre>
      *  
      * @return True if the class of the calling context is defined in the same 
      *         package as the class representing the immediate calling context, 
      *         false if not. 
      * @see    #isThisClass()         
      * @see    #isNestedClass()         
      * @see    #isSuperClass()
      * @see    #isSubClass()         
      * @see    #isAssignableToThisClass()         
      */
    public final boolean isThisPackage() {
        // Also handles default packages...
        return Util.equals(this.calleeClass.getPackage(), this.callerClass.getPackage());
    }
    
    /**
      * Returns true if the {@link #callerClass class} of the calling context 
      * is the same {@link #calleeClass class} representing the immediate 
      * calling context, false if not. <p> 
      *  
      * <b>Example</b>: 
      * <pre>
      *   ..
      *   public class Foo {
      *     public void foo() {
      *       this.foofoo();
      *     }
      *   
      *     public void foofoo() {
      *       CallerClass caller = new CallerClass();
      *       System.out.println(caller.isThisClass());
      *     }
      *   }
      *   
      *   public class Bar {
      *     public void bar(Foo foo) {
      *       foo.foofoo();
      *     }
      *   }
      *   
      *   ..
      *   
      *   Foo foo = new Foo();
      *   Bar bar = new Bar();
      *   
      *   foo.foo();    // prints "true"
      *   bar.bar(foo); // prints "false"   
      *   
      *   ..
      * </pre>
      *  
      * @return True if the class of the calling context is the same 
      *         class representing the immediate calling context, 
      *         false if not. 
      * @see    #isThisPackage()         
      * @see    #isNestedClass()
      * @see    #isSuperClass()
      * @see    #isSubClass()         
      * @see    #isAssignableToThisClass()         
      */
    public final boolean isThisClass() {
        return (this.calleeClass == this.callerClass);
    }

    /**
      * Returns true if the {@link #callerClass class} of the calling context 
      * is a nested or inner class of the {@link #calleeClass class} representing
      * the immediate calling context, false if not. <p> 
      *  
      * <b>Example</b>: 
      * <pre>
      *   ..
      *   public class Foo {
      *     public void foo() {
      *       CallerClass caller = new CallerClass();
      *       System.out.println(caller.isNestedClass());
      *     }
      *     
      *     public static class Bar {
      *       public void bar(Foo foo) {
      *         foo.foo();
      *       }
      *     }
      *   }
      *   
      *   public class Goo {
      *     public void goo(Foo foo) {
      *       foo.foo();
      *     }
      *   }
      *   ..
      *   
      *   Foo foo     = new Foo();
      *   Foo.Bar bar = new Foo.Bar();
      *   Goo goo     = new Goo();
      *   
      *   bar.bar(foo); // prints "true"
      *   goo.goo(foo); // prints "false"   
      *   
      *   ..
      * </pre>
      *  
      * @return True if the class of the calling context is a nested or inner 
      *         class of the class representing the immediate calling context, 
      *         false if not. 
      * @see    #isThisPackage()         
      * @see    #isThisClass()
      * @see    #isSuperClass()
      * @see    #isSubClass()
      * @see    #isAssignableToThisClass()         
      */
    public final boolean isNestedClass() {
        Class<?> enclosingClass = this.callerClass.getEnclosingClass();
        while (enclosingClass != null) {
            if (this.calleeClass == enclosingClass) {
                return true;
            }
            enclosingClass = enclosingClass.getEnclosingClass();
        }
        return false;
    }
    
    /**
      * Returns true if the {@link #callerClass class} of the calling context 
      * is a super-class of the {@link #calleeClass class} representing the 
      * immediate calling context, false if not. <p> 
      *  
      * <b>Example</b>: 
      * <pre>
      *   ..
      *   public abstract class Foo {
      *     public Foo() {
      *       this.foo();
      *     }
      *     public abstract void foo();
      *   }
      *   
      *   public class Bar extends Foo {
      *     public Bar() {
      *       super();
      *     }
      *     public void foo() {
      *       CallerClass caller = new CallerClass();
      *       System.out.println(caller.isSuperClass());
      *     }
      *   }
      *   
      *   public class Goo {
      *     public void goo(Foo foo) {
      *       foo.foo();
      *     }
      *   }
      *   ..
      *   
      *   Bar bar = new Bar(); // prints "true"
      *   Goo goo = new Goo();
      *   goo.goo(bar);        // prints "false"   
      *   
      *   ..
      * </pre>
      *  
      * @return True if the class of the calling context is a super-class of the 
      *         class representing the immediate calling context, false if not. 
      * @see    #isThisPackage()         
      * @see    #isThisClass()
      * @see    #isSubClass()
      * @see    #isNestedClass()
      * @see    #isAssignableToThisClass()         
      */
    public final boolean isSuperClass() {
        return Reflection.isSuperClass(this.calleeClass, this.callerClass); 
    }
    
    /**
      * Returns true if the {@link #callerClass class} of the calling context 
      * is a sub-class of the {@link #calleeClass class} representing the 
      * immediate calling context, false if not. <p> 
      *  
      * <b>Example</b>: 
      * <pre>
      *   ..
      *   public class Foo {
      *     public void foo() {
      *       CallerClass caller = new CallerClass();
      *       System.out.println(caller.isSubClass());
      *     }
      *   }
      *   
      *   public class Bar extends Foo {
      *     public void bar(Foo foo) {
      *       foo.foo();
      *     }
      *   }
      *   
      *   public class Goo {
      *     public void goo(Foo foo) {
      *       foo.foo();
      *     }
      *   }
      *   ..
      *   
      *   Foo foo = new Foo();
      *   Bar bar = new Bar();
      *   Goo goo = new Goo();
      *   
      *   bar.bar(foo); // prints "true"
      *   goo.goo(foo); // prints "false"   
      *   
      *   ..
      * </pre>
      *  
      * @return True if the class of the calling context is a sub-class of the 
      *         class representing the immediate calling context, false if not. 
      * @see    #isThisPackage()         
      * @see    #isThisClass()
      * @see    #isSuperClass()
      * @see    #isNestedClass()
      * @see    #isAssignableToThisClass()         
      */
    public final boolean isSubClass() {
        return Reflection.isSubClass(this.calleeClass, this.callerClass); 
    }
    
    /**
      * Returns true if the {@link #callerClass class} of the calling context 
      * is a class that is assignable to the {@link #calleeClass class} 
      * representing the immediate calling context, false if not. <p> 
      *  
      * <b>Example</b>: 
      * <pre>
      *   ..
      *   public class Foo implements FooBar {
      *     public void foo() {
      *       CallerClass caller = new CallerClass();
      *       System.out.println(caller.isAssignableToThisClass());
      *     }
      *   }
      *   
      *   public class Bar implements FooBar {
      *     public void bar(Foo foo) {
      *       foo.foo();
      *     }
      *   }
      *   
      *   public class Goo {
      *     public void goo(Foo foo) {
      *       foo.foo();
      *     }
      *   }
      *   ..
      *   
      *   Foo foo = new Foo();
      *   Bar bar = new Bar();
      *   Goo goo = new Goo();
      *   
      *   bar.bar(foo); // prints "true"
      *   goo.goo(foo); // prints "false"   
      *   
      *   ..
      * </pre>
      *  
      * @return True if the class of the calling context is a class that is 
      *         assignable to the class representing the immediate calling 
      *         context, false if not. 
      * @see    #isThisPackage()         
      * @see    #isThisClass()
      * @see    #isSuperClass()
      * @see    #isSubClass()
      */
    public final boolean isAssignableToThisClass() {
        return this.calleeClass.isAssignableFrom(this.callerClass); 
    }

    /**
      * Returns true if just one of the classes fetched from an object 
      * in {@code objects} is contained in the call-stack of this caller, 
      * false if not. <p>
      * 
      * {@code objects} may contain {@code Class} objects as well.
      *  
      * @param  objects The objects delivering the classes to test; cannot
      *                 contain null entries.
      * @return True if just one of the classes fetched from an object in 
      *         {@code objects} is contained in the call-stack, false if not.
      * @throws NullPointerException If {@code objects} contains null entries.
      * @see    #isPackageInCallStack(Object...)         
      */
    public boolean isClassInCallStack(Object... objects) {
        for (Object o : objects) {
            Class<?> c = o instanceof Class ? (Class<?>)o : o.getClass();
            for (int j = this.stack.length - 1; j >= this.offset; j--) {
                if (getClass(this.stack[j]).equals(c)) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
      * Returns true if just one of the packages fetched from an object 
      * in {@code objects} is contained in the call-stack of this caller, 
      * false if not. <p>
      * 
      * {@code objects} may contain {@code Class} and {@code Package} 
      * objects as well.
      *  
      * @param  objects The objects delivering the packages to test; cannot 
      *                 contain null entries.
      * @return True if just one of the packages fetched from an object in 
      *         {@code objects} is contained in the call-stack, false if not.
      * @throws NullPointerException If {@code objects} contains null entries.
      * @see    #isClassInCallStack(Object...)         
      */
    public boolean isPackageInCallStack(Object... objects) {
        for (Object o : objects) {
            Package p = null;
            if (o instanceof Package) {
                p = (Package)o;
            } else if (o instanceof Class) {
                p = ((Class<?>)o).getPackage();
            } else {
                p = o.getClass().getPackage();
            }
            for (int j = this.stack.length - 1; j >= this.offset; j--) {
                if (getClass(this.stack[j]).getPackage().equals(p)) {
                    return true;
                }
            }
        }
        return false;
    }
      
    /**
      * Returns the caller representing the calling context of 
      * <i>this</i> caller, or null if no such caller exist. <p>
      * 
      * The returned caller will represents the same immediate
      * calling context as this caller, i.e. the {@link #calleeClass} 
      * is equal to that of this caller.
      *  
      * @return The caller representing the calling context of this 
      *         caller, or null if no such caller exist (end of call-stack).
      * @see    #getCallers()         
      */
    public CallerClass getCaller() {
        if (this.stack.length > this.offset) {
            return new CallerClass(this, this.offset + 1);
        }
        return null;
    }
    
    /**
      * Returns a new iterator to iterate over all <b>preceding</b> 
      * callers in the call-stack. <p> 
      * 
      * Each returned caller will represents the same immediate
      * calling context as this caller, i.e. the {@link #calleeClass} 
      * is equal to that of this caller.
      * 
      * The {@code remove()} method of the returned iterator will 
      * always throw an {@code UnsupportedOperationException}.
      * 
      * @return A new iterator; never null.
      * @see    #getCaller()
      * @see    #iterator()
      */
    public Iterator<? extends CallerClass> getCallers() {
        return this.iterator();
    }
    
    /**
      * 
      */
    public Iterator<CallerClass> iterator() {
        return new CallerIterator<CallerClass>(this) {
            protected void setNext() {
                this.caller = this.caller.getCaller();
            }
        };
    }
    
    /**
      * Returns the <b>first</b> caller in the call-stack where the  
      * {@link #callerClass class} of the calling context 
      * is a sub-class of the {@link #calleeClass class} representing 
      * the immediate calling context. <p> 
      * 
      * @return The first caller representing a class that is 
      *         a sub-class, or null.
      * @see    #getNonSubClassCaller()
      */
    public CallerClass getSubClassCaller() {
        // This caller...
        if (this.isSubClass()) {
            return this;
        }
        // Callers to this caller...
        for (CallerClass caller : this) {
            if (caller.isSubClass()) {
                return caller;
            }
        }
        return null;
    }

    /**
      * Returns the <b>first</b> caller in the call-stack where the  
      * {@link #callerClass class} of the calling context 
      * is a not a sub-class of the {@link #calleeClass class} representing 
      * the immediate calling context. <p> 
      * 
      * @return The first caller representing a class that is not 
      *         a sub-class, or null.
      * @see    #getSubClassCaller()
      */
    public CallerClass getNonSubClassCaller() {
        // This caller...
        if (!this.isSubClass()) {
            return this;
        }
        // Callers to this caller...
        for (CallerClass caller : this) {
            if (!caller.isSubClass()) {
                return caller;
            }
        }
        return null;
    }

    /**
      * Returns the <b>first</b> caller in the call-stack where the  
      * {@link #callerClass class} of the calling context 
      * is assignable to the {@link #calleeClass class} representing 
      * the immediate calling context. <p> 
      * 
      * This may be the same class as the {@link #calleeClass class} 
      * of the immediate calling context represented by this caller.
      * 
      * @return The first caller representing a class that is 
      *         assignable, or null.
      * @see    #getNonAssignableCaller()
      */
    public CallerClass getAssignableCaller() {
        // This caller...
        if (this.isAssignableToThisClass()) {
            return this;
        } 
        // Callers to this caller...
        for (CallerClass caller : this) {
            if (caller.isAssignableToThisClass()) {
                return caller;
            }
        }
        return null;
    }
    
    /**
      * Returns the <b>first</b> caller in the call-stack where the  
      * {@link #callerClass class} of the calling context 
      * is not assignable to the {@link #calleeClass class} representing 
      * the immediate calling context. <p> 
      * 
      * @return The first caller representing a class that is 
      *         not assignable, or null.
      * @see    #getAssignableCaller()
      */
    public CallerClass getNonAssignableCaller() {
        // This caller...
        if (!this.isAssignableToThisClass()) {
            return this;
        }
        // Callers to this caller...
        for (CallerClass caller : this) {
            if (!caller.isAssignableToThisClass()) {
                return caller;
            }
        }
        return null;
    }

    /**
      * Returns the <b>first</b> caller in the call-stack where the  
      * {@link #callerClass class} of the calling context 
      * is identical to the {@link #calleeClass class} representing 
      * the immediate calling context. <p> 
      * 
      * @return The first caller representing the same class, or null.
      * @see    #getNonClassCaller()
      */
    public CallerClass getClassCaller() {
        // This caller...
        if (this.isThisClass()) {
            return this;
        }
        // Callers to this caller...
        for (CallerClass caller : this) {
            if (caller.isThisClass()) {
                return caller;
            }
        }
        return null;
    }
    

    /**
      * Returns the <b>first</b> caller in the call-stack where the  
      * {@link #callerClass class} of the calling context 
      * is not identical to the {@link #calleeClass class} representing 
      * the immediate calling context. <p> 
      * 
      * @return The first caller representing a different class, or null.
      * @see    #getClassCaller()
      */
    public CallerClass getNonClassCaller() {
        // This caller...
        if (!this.isThisClass()) {
            return this;
        }
        // Callers to this caller...
        for (CallerClass caller : this) {
            if (!caller.isThisClass()) {
                return caller;
            }
        }
        return null;
    }
    
    /**
      * Returns the hash code of the caller.
      * 
      * @return The hash code.
      */
    public int hashCode() {
        return this.callerClass.hashCode();
    }
    
    /**
      * Returns true if {@code object} is an instance of
      * this class representing the <b>same</b> {@link #callerClass caller}
      * and {@link #calleeClass context} class in the <b>same</b> 
      * call-stack, false if not. 
      * 
      * @param  object The object to test; can be null.
      * @return True if equal, false if not. 
      */
    public boolean equals(Object object) {
        if (super.equals(object)) {
            // No need to test the caller class since that will be the
            // same in case all other attributes are equal (assuming 
            // same class loader)...
            CallerClass cc = (CallerClass)object;
            return this.calleeClass == cc.calleeClass;
        }
        return false;
    }
  
}
