/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.reflect;

import java.lang.reflect.Constructor;
import java.lang.reflect.Modifier;
import java.lang.reflect.Type;
import java.util.Arrays;

import dk.rode.thesis.meta.log.LogFactory;

/**
  * Represents the <i>instantiable type</i> realised as the type 
  * parameter {@code T}, where {@code T} may be a generic type. <p>
  *
  * New instances of {@code T} can be created using the
  * {@link #newInstance(Object...)} method. Hence, interfaces
  * and abstract classes cannot be used as the <i>raw type</i>
  * of {@code T}. <p>
  * 
  * The constructor to create new instances of {@code T} is
  * fetched ignoring primitive types, e.g. a formal parameter type
  * of {@code Integer} will match both {@code Integer} and {@code int}. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    TypeLiteral
  * 
  * @param  <T> The type represented by this instantiable type literal.
  */
public abstract class InstantiableTypeLiteral<T> extends TypeLiteral<T> {

    /**
      * The constructor used to create new instances of the
      * type represented by the type parameter {@code T}. <p>
      * 
      * The constructor is identified by its formal parameter 
      * types supplied at construction time. <p>
      * 
      * Never null.
      */
    private final Constructor<?> constructor;

    /**
      * Constructor, which creates this instantiable type literal to represent
      * the type realised by the type parameter {@code T}. <p>
      * 
      * New instances of {@code T} can be {@link #newInstance(Object...)
      * created} using the declared constructor having the formal parameter 
      * types supplied {@code parameterTypes}. <p>
      * 
      * The constructor is made accessible if not already.
      *        
      * @param  parameterTypes The formal parameter types for the 
      *                        declared constructor to use.
      * @throws IllegalArgumentException If this class is not generic, if
      *         {@code T} has no raw type, or if the raw raw represents 
      *         a non instantiable class.
      * @throws NoSuchMethodException If no such constructor is declared in 
      *         the raw type of {@code T}. 
      */
    protected InstantiableTypeLiteral(Class<?>... parameterTypes) 
        throws NoSuchMethodException {
        super();
        this.constructor = getConstructor(this.type, parameterTypes); // type set in super
    }

    /**
      * Constructor, which creates this instantiable type literal to represent
      * the type represented by the type literal {@code typeLiteral}. <p>
      * 
      * New instances of {@code T} can be {@link #newInstance(Object...)
      * created} using the declared constructor having the formal parameter 
      * types supplied {@code parameterTypes}. <p>
      * 
      * The constructor is made accessible if not already.
      * 
      * @param  typeLiteral    The type literal; cannot be null.
      * @param  parameterTypes The formal parameter types for the 
      *                        declared constructor to use.
      * @throws IllegalArgumentException If this class is not generic, if
      *         {@code T} has no raw type, or if the raw raw represents 
      *         a non instantiable class.
      * @throws NoSuchMethodException If no such constructor is declared in 
      *         the raw type of {@code T}. 
      */
    private InstantiableTypeLiteral(TypeLiteral<T> typeLiteral, Class<?>... parameterTypes) 
        throws NoSuchMethodException {
        super(typeLiteral);
        this.constructor = getConstructor(this.type, parameterTypes);
    }

    /**
      * Copy constructor, which creates this instantiable type literal 
      * based on the generic type and constructor represented by 
      * {@code typeLiteral}
      * 
      * @param  typeLiteral The instantiable type literal to copy; cannot be null.
      * @throws NullPointerException If {@code typeLiteral} is null.
      */
    public InstantiableTypeLiteral(InstantiableTypeLiteral<T> typeLiteral) {
        super(typeLiteral);
        this.constructor = typeLiteral.constructor;
    }
    
    /**
      * Returns the constructor having the formal parameter types
      * supplied as {@code parameterTypes} in the raw type
      * fetched from {@code type}. <p> 
      * 
      * The constructor is made accessible if not already.
      * 
      * @param  type           The type; cannot be null. 
      * @param  parameterTypes The formal parameter types.
      * @return The constructor; never null.
      * @throws NullPointerException If {@code type} is null.
      * @throws IllegalArgumentException If {@code T} has no raw type,
      *         or if the raw type represents an interface or abstract class.
      * @throws NoSuchMethodException If no such constructor is declared in 
      *         the raw type.
      * @see    Reflection#getConstructor(Class, Class...)         
      */
    private final static Constructor<?> getConstructor(Type type, Class<?>... parameterTypes) 
        throws NoSuchMethodException {
        if (!isInstatiable(type)) {
            throw new IllegalArgumentException("Type is not instantiable");
        }
        // Fetch the constructor ignoring primitives...
        Constructor<?> constructor = Reflection.getConstructor(getRawType(type), parameterTypes);
        if (!Modifier.isPublic(constructor.getModifiers())) {
            constructor.setAccessible(true);
        }
        return constructor;
    }
    
    /**
      * Returns an instantiable type literal representing the type 
      * {@code S} supplied via the class literal {@code clazz}. <p>
      * 
      * The declared constructor to use is identified by the formal
      * parameter types supplied as {@code parameterTypes}. The 
      * constructor is made accessible if not already.
      * 
      * @param  <S>            The type.
      * @param  clazz          The class literal representing the type; 
      *                        cannot be null.
      * @param  parameterTypes The formal parameter types for the declared 
      *                        constructor to use.
      * @return A instantiable type literal representing the type {@code S}; 
      *         never null.
      * @throws NullPointerException If {@code clazz} is null.
      * @throws IllegalArgumentException If {@code S} is a non-instantiable
      *         type.
      * @throws NoSuchMethodException If no such constructor is declared in
      *         the raw type of {@code S}.
      * @see    #create(Type, Class...)
      * @see    #create(TypeLiteral, Class...)
      * @see    #isInstatiable(Type)                   
      */
    public static <S> InstantiableTypeLiteral<S> create(Class<S> clazz, Class<?>... parameterTypes) 
        throws NoSuchMethodException {
        return new InstantiableTypeLiteral<S>(parameterTypes){/*magic*/};
    }
    
    /**
      * Returns a instantiable type literal representing the type 
      * supplied as {@code type}.
      * 
      * The declared constructor to use is identified by the formal
      * parameter types supplied as {@code parameterTypes}. The 
      * constructor is made accessible if not already.
      * 
      * @param  type           The type; cannot be null. 
      * @param  parameterTypes The formal parameter types for the declared 
      *                        constructor to use.
      * @return A instantiable type literal representing the type; 
      *         never null.
      * @throws NullPointerException If {@code type} is null.
      * @throws IllegalArgumentException If {@code type} represents a 
      *         non-instantiable type.
      * @throws NoSuchMethodException If no such constructor is declared in
      *         the raw type of {@code type}.
      * @see    #create(Class, Class...)
      * @see    #create(TypeLiteral, Class...)                   
      * @see    #isInstatiable(Type)                   
      */
    public static InstantiableTypeLiteral<?> create(Type type, Class<?>... parameterTypes)
        throws NoSuchMethodException {
        return create(create(type), parameterTypes);
    }
    
    /**
      * Returns an instantiable type literal representing the type 
      * {@code S} supplied via the type literal {@code type}. <p>
      * 
      * The declared constructor to use is identified by the formal
      * parameter types supplied as {@code parameterTypes}. The 
      * constructor is made accessible if not already.
      * 
      * @param  <S>            The type.
      * @param  typeLiteral    The type literal representing the type; 
      *                        cannot be null.
      * @param  parameterTypes The formal parameter types for the declared 
      *                        constructor to use.
      * @return A instantiable type literal representing the type {@code S}; 
      *         never null.
      * @throws NullPointerException If {@code typeLiteral} is null.
      * @throws IllegalArgumentException If {@code typeLiteral} represents a 
      *         non-instantiable type.
      * @throws NoSuchMethodException If no such constructor is declared in
      *         the raw type of {@code S}.
      * @see    #create(Class, Class...)
      * @see    #create(Type, Class...)
      * @see    #isInstatiable(TypeLiteral)                   
      */
    public static <S> InstantiableTypeLiteral<S> create(TypeLiteral<S> typeLiteral, Class<?>... parameterTypes) 
        throws NoSuchMethodException {
        return new InstantiableTypeLiteral<S>(typeLiteral, parameterTypes){/*magic*/};
    }
    
    /**
      * Returns true if the raw type represented by {@code type}
      * is instantiable, false if not. <p>
      * 
      * If no raw type exist based on {@code type}, or if the raw type
      * is abstract or an interface, false is returned; otherwise
      * true. <p>
      * 
      * This method does not test for explicit constructors.
      *  
      * @param  type The type to test; cannot be null.
      * @return True if instantiable, false if not.
      * @throws NullPointerException If {@code type} is null.
      */
    public static boolean isInstatiable(Type type) {
        Class<?> clazz = getRawType(type);
        return !((clazz == null) || (clazz.isInterface()) || (Modifier.isAbstract(clazz.getModifiers())));
    }

    /**
      * Returns true if the raw type represented by {@code type}
      * is instantiable, false if not. <p>
      * 
      * If no raw type exist based on {@code type}, or if the raw type
      * is abstract or an interface, false is returned; otherwise
      * true. <p> 
      * 
      * This method does not test for explicit constructors.
      *  
      * @param  type The type to test; cannot be null.
      * @return True if instantiable, false if not.
      * @throws NullPointerException If {@code type} is null.
      */
    public static boolean isInstatiable(TypeLiteral<?> type) {
        return isInstatiable(type.getType());
    }
    
    /**
      * Returns a new instance of the type represented by
      * this instantiable type literal. <p> 
      *  
      * @param  arguments The arguments to supply to the constructor. 
      * @return A new instance of the type represented by this
      *         instantiable type literal; never null.
      * @throws Exception If the construction fails.
      */
    public T newInstance(Object... arguments) throws Exception {
        // Cast known to succeed since the constructor describes (raw type of) T...
        @SuppressWarnings("unchecked") T object = 
            (T)this.constructor.newInstance(arguments);
        LogFactory.getLog(this).println("Created: ", arguments == null ? "null" : Arrays.asList(arguments), " -> ", object);
        return object;
    }

    /**
      * Return the constructor represented by this instantiable 
      * type literal.
      *  
      * @return The constructor; never null.
      * @see    #getType()
      */
    public Constructor<?> getConstructor() {
        return this.constructor;
    }
    
    /**
      * Returns true of {@code object} is an instantiable type literal
      * representing the same {@link #getType() type} and constructor 
      * as this instantiable type literal, false if not.
      * 
      * @param  object The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object object) {
        if (!(object instanceof InstantiableTypeLiteral)) {
            return false;
        }
        InstantiableTypeLiteral<?> itl = (InstantiableTypeLiteral<?>)object;
        return this.type.equals(itl) && this.constructor.equals(itl.constructor); 
    }
    
}
