/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.reflect;

import static dk.rode.thesis.meta.util.Throwables.toRuntimeException;

import java.lang.annotation.Annotation;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import dk.rode.thesis.meta.util.KeyValue;
import dk.rode.thesis.meta.util.Primitive;
import dk.rode.thesis.meta.util.Util;

/**
  * Various <i>reflection functionality</i>. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * Though all methods are declared static, instances of this class 
  * cannot be created, and hence it is in effect an application of
  * the {@linkplain dk.rode.thesis.singleton Singleton} pattern.
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a> 
  */
public class Reflection {
    
    /**
      * Private off-limit constructor.
      */
    private Reflection() {
        throw new AssertionError(); // guard against reflection...
    }
    
    /**
      * Returns the {@code Class} object associated with the class or
      * interface with the name supplied as {@code className}. <p>
      * 
      * Unlike the standard {@code Class.forName(String)} method, this 
      * method trims {@code className} before trying the load the class 
      * and it also handles primitive class names, e.g. {@code int} for
      * {@code Integer.TYPE} etc. 
      * 
      * @param  className The class name; cannot be null.
      * @return The corresponding {@code Class} object; never null.
      * @throws NullPointerException If {@code className} is null.
      * @throws ClassNotFoundException If no such class could be
      *         found.
      */
    public final static Class<?> forName(CharSequence className) throws ClassNotFoundException {
        String name = className.toString().trim();
        try {
            return Class.forName(name);
        } catch (ClassNotFoundException cnfe) {
            // Check primitives...
            for (Class<?> clazz : Primitive.getPrimitiveTypes()) {
                if (clazz.getName().equals(name)) {
                    return clazz;
                }
            }
            throw cnfe;
        }
    }
    
    /**
      * Returns the <i>caller class</i> of the <i>calling context
      * of the immediate calling context</i> that invoked this method. <p>
      * 
      * The calling context can be either a static initialisation block,
      * an instance initialisation block, a constructor, or a method. The
      * caller class thus represents the class declaring such a member. <p>
      *  
      * <b>Example</b>:
      * <pre>
      *   ..
      *   public void foo {
      *     // Calling context
      *     this.bar();
      *   }
      * 
      *   public void bar {
      *     // Immediate calling context
      *     Class&lt;?> callerClass = Reflection.getCallerClass(); // Represents the class of foo, *not* of bar
      *     .. 
      *   }
      *   ..
      * </pre>
      * 
      * @return The caller class of the calling context; never null.
      * @see    CallerClass
      * @see    Caller
      */
    public final static Class<?> getCallerClass() {
        return new CallerClass(1).callerClass;
    }
    
    /**
      * Checks whether or not {@code caller} can access {@code member} based 
      * on Java's visibility rules.<p>
      * 
      * <b>Caution</b>: the access check is performed <b>without</b> 
      * the use of a security manager, and is <b>only</b> based on the 
      * access modifiers and packages of of {@code caller} and {@code member}! <p>
      *
      * @param  caller The caller class; cannot be null.
      * @param  member The member to test; cannot be null.
      * @return True if {@code caller} has access, false if not.
      * @throws NullPointerException If either parameter is null.
      * @see    #hasAccess(Class, Class)           
      */
    public final static boolean hasAccess(Class<?> caller, Member member) {
        Class<?> clazz = member.getDeclaringClass();
        return hasAccess(caller, clazz, clazz.getModifiers()) &&
               hasAccess(caller, clazz, member.getModifiers());
    }

    /**
      * Checks whether or not {@code caller} can access {@code clazz} 
      * based on Java's visibility rules.<p>
      * 
      * <b>Caution</b>: the privilege check is performed <b>without</b> 
      * the use of a security manager, and is <b>only</b> based on the 
      * access modifiers and packages of {@code caller} and 
      * {@code object}! <p>
      *
      * @param  caller The caller class; cannot be null.
      * @param  clazz  The clazz to test; cannot be null.
      * @return True if {@code caller} has access, false if not.
      * @throws NullPointerException If either parameter is null.
      * @see    #hasAccess(Class, Member)           
      */
    public final static boolean hasAccess(Class<?> caller, Class<?> clazz) {
        return hasAccess(caller, clazz, clazz.getModifiers());
    }
    
    /**
      * Checks whether or not the access modifiers supplied as {@code modifiers}
      * allows {@code caller} to access to the <i>owner</i> of the modifiers declared
      * in, or being, {@code clazz}. <p>
      * 
      * An <i>owner</i> will thus be a member, e.g. method, or a class, e.g. {@code clazz}
      * it self.
      *  
      * @param  caller    The caller class; cannot be null.
      * @param  clazz     The clazz to test; cannot be null.
      * @param  modifiers The access modifiers.
      * @return True if {@code modifiers} allows {@code caller} to access the owner
      *         of the modifiers, false if not.
      * @throws NullPointerException If either {@code caller} or {@code clazz} are null.
      */
    private final static boolean hasAccess(Class<?> caller, Class<?> clazz, int modifiers) {
        // Public...
        if (Modifier.isPublic(modifiers)) {
            return true;
            
        // Protected...
        } else if (Modifier.isProtected(modifiers)) {
            // Same package (also handles equal classes)?...
            if (Util.equals(caller.getPackage(), clazz.getPackage())) {
                return true;
            }
            // Sub-class?...
            return isSubClass(clazz, caller);
           
        // Private...
        } else if (Modifier.isPrivate(modifiers)) {
            return caller.equals(clazz);
            
        // Package private (no access modifier)...
        } else {
            return Util.equals(caller.getPackage(), clazz.getPackage());
        }
    }
    
    /**
      * Determines if a given accessible non-abstract method is overriding a 
      * similar method in a super class based on Java's visibility rules. <p>
      * 
      * The overriding method, e.g. the method defined by {@code name}
      * and the formal parameter types supplied as {@code types} overriding
      * a similar method in a super class, need not be declared in {@code clazz},
      * but may be declared in any super class of it, providing it is accessible
      * to {@code clazz}.
      * 
      * @param  clazz The initial class to start the search for the method; cannot be null.
      * @param  name  Method name; cannot be null.
      * @param  types Formal types of the method parameters, if any; can be empty.
      * @return True if such a method is overriding a similar method from a
      *         super class, false if not; the overriding method need not be declared in
      *         {@code clazz}, but can also be declared in any super class of it. 
      * @throws NullPointerException If {@code clazz} or {@code name} are null.
      * @see    #getOverriding(Class, String, Class...)
      * @see    #hasAccess(Class, Member)         
      */
    public final static boolean isOverriding(Class<?> clazz, String name, Class<?>... types) {
        KeyValue<?,?> classes = getOverriding(clazz, name, types);
        return ((classes != null) && (classes.getValue() != null));
    }

    /**
      * Returns a (key,value) placeholder, where the <i>key</i> is the first super class of 
      * {@code clazz}, or {@code clazz} it self, declaring a non-abstract method with 
      * the name supplied as {@code name} and the formal parameter types supplied as {@code types},
      * providing it is accessible to {@code clazz}, and where the <i>value</i> is the first super 
      * class of that class declaring a similar method. Hence, the class stored as the key declares 
      * the identified overriding method, and the class stored as the value declares the similar 
      * method being overridden. <p>
      * 
      * If the value of the returned (key,value) placeholder is null, the method does not override 
      * any similar method from a super class. If the value is not null, the class stored as the 
      * value will be a regular super class of the class stored as the key. <p>
      * 
      * If the returned placeholder is null, no such accessible method could be found in 
      * {@code clazz}, or in any of its super classes. <p>
      * 
      * Java's normal visibility rules are applied in the check.
      * 
      * @param  <T>   The type of the initial class.
      * @param  clazz The initial class to start the search for the method; cannot be null.
      * @param  name  Method name; cannot be null.
      * @param  types Formal types of the method parameters, if any; can be empty.
      * @return A (key,value) placeholder storing the relevant classes; can be null if no such
      *         method can be found in any super class, or {@code clazz} it self. The value will
      *         be null if the specified method does not override any similar method from a 
      *         super class.
      * @throws NullPointerException If {@code clazz} or {@code name} are null.
      * @see    #isOverriding(Class, String, Class...)         
      * @see    #hasAccess(Class, Member)         
      */
    public final static <T> KeyValue<Class<? super T>,Class<? super T>> getOverriding(Class<T> clazz, String name, Class<?>... types) {
        
        // Note:
        // We do not check on the @Override annotation, since as of Java 6, it is allowed
        // to it on methods implementing a method declared in an interface as well. Since
        // we are only interested in methods overriding actual behaviour, we in any case
        // have to test super classes for such non-abstract methods. Furthermore, the 
        // @Override annotation does not take accessibility relative to "clazz" into account!
        
        KeyValue<Class<? super T>,Class<? super T>> classes = null;
        Class<? super T> c = clazz, first = clazz;
        do {
            try {
                Method method = c.getDeclaredMethod(name, types);
                // If here, such a method is declared in the current class...
                int modifiers = method.getModifiers();
                // Static or abstract methods are never overridden...
                if (Modifier.isStatic(modifiers) || Modifier.isAbstract(modifiers)) {
                    continue;
                }
                boolean access = hasAccess(first, method);
                if ((classes != null) && (access)) {
                    return classes.setValue(c);
                } else if (access) {
                    classes = new KeyValue<Class<? super T>,Class<? super T>>(first = c);
                }
            } catch (Exception e) {
                // Intentionally left blank...
            }
            
        } while ((c = c.getSuperclass()) != null);
        
        // Value may be null = not overriding...
        return classes;
    }
    
    /**
      * Returns the method with the name supplied as {@code name} that 
      * is declared in the class supplied as {@code clazz} accepting 
      * the formal parameter types supplied as {@code parameterTypes}, 
      * ignoring primitive types. <p>
      * 
      * In case several methods match, the method declared first in 
      * {@code clazz} is returned.
      *  
      * @param  clazz          The class; cannot be null.
      * @param  name           The method name; cannot be null.
      * @param  parameterTypes The formal parameter types for the method
      *                        to fetch.
      * @return The (first) matching method; never null.
      * @throws NullPointerException If {@code clazz} or {@code name} are null.
      * @throws NoSuchMethodException If no such method can be found.
      */
    public final static Method getMethod(Class<?> clazz, String name, Class<?>... parameterTypes) throws NoSuchMethodException {
        // Try and fetch directly...
        try {
            return clazz.getDeclaredMethod(name, parameterTypes);
        } catch (NoSuchMethodException nsme) {
            
            // Check assignable from...
            loop:
            for (Method method : clazz.getDeclaredMethods()) {
                if (method.getName().equals(name)) {
                    Class<?>[] types = method.getParameterTypes();
                    if (types.length != parameterTypes.length) {
                        continue;
                    }
                    for (int i = 0; i < types.length; i++) {
                        if (!Primitive.isAssignableFromIgnorePrimitives(types[i], parameterTypes[i])) {
                            continue loop;
                        }
                    }
                    // Match...
                    return method;
                }
            }
            // No match - re-throw...
            throw nsme;
        }
    }
    
    /**
      * Returns the methods declared in the class supplied as {@code clazz}
      * annotated with the annotation type supplied as {@code annotation}.
      *  
      * @param  <A>        The type of annotation.
      * @param  clazz      The class; cannot be null.
      * @param  annotation The class representing the annotation type; cannot be null. 
      * @return An array containing the annotated methods; never null, but
      *         can be empty.
      * @throws NullPointerException If either argument is null.
      * @see    #getConstructors(Class, Class)         
      */
    public static <A extends Annotation> Method[] getMethods(Class<?> clazz, Class<A> annotation) {
        if (annotation == null) {
            throw new NullPointerException("Annotation cannot be null");
        }
        List<Method> methods = new ArrayList<Method>();
        for (Method m : clazz.getDeclaredMethods()) {
            Annotation a = m.getAnnotation(annotation);
            if (a != null) {
                methods.add(m);
            }
        }
        return methods.toArray(new Method[methods.size()]);
    }

    /**
      * Returns the constructors declared in the class supplied as {@code clazz}
      * annotated with the annotation type supplied as {@code annotation}.
      *  
      * @param  <A>        The type of annotation.
      * @param  clazz      The class; cannot be null.
      * @param  annotation The class representing the annotation type; cannot be null. 
      * @return An array containing the annotated constructors; never null, but
      *         can be empty.
      * @throws NullPointerException If either argument is null.
      * @see    #getMethods(Class, Class)         
      */
    public static <A extends Annotation> Constructor<?>[] getConstructors(Class<?> clazz, Class<A> annotation) {
        if (annotation == null) {
            throw new NullPointerException("Annotation cannot be null");
        }
        List<Constructor<?>> constructors = new ArrayList<Constructor<?>>();
        for (Constructor<?> c : clazz.getDeclaredConstructors()) {
            Annotation a = c.getAnnotation(annotation);
            if (a != null) {
                constructors.add(c);
            }
        }
        return constructors.toArray(new Constructor[constructors.size()]);
    }
    
    /**
      * Returns a list of all super-classes of {@code clazz},
      * excluding {@code clazz}, where the last index is always
      * be {@code Object.class}.
      *  
      * @param  <T>   The type of {@code clazz}. 
      * @param  clazz The class to fetch the super-classes for;
      *               cannot be null.
      * @return A list containing the super-classes; never null,  
      *         but will be empty if {@code clazz.equals(Object.class)}
      *         is true.
      * @throws NullPointerException If {@code clazz} is null.         
      */
    public final static <T> List<Class<? super T>> getSuperClasses(Class<T> clazz) {
        // No need to use a set, since we know that no class will
        // appear twice in the collection...
        List<Class<? super T>> classes = new ArrayList<Class<? super T>>(4);
        Class<? super T> c = clazz;
        while ((c = c.getSuperclass()) != null) {
            classes.add(c);
        }
        return classes;
    }

    /**
      * Returns true if and only if {@code subClass} is a sub-class
      * of {@code clazz}, false otherwise. <p> 
      * 
      * For this method to return true, {@code subClass} must at some 
      * point extend {@code clazz}, directly or indirectly. Implementing 
      * common interfaces is not enough.
      * 
      * @param  clazz    The parent class; cannot be null.
      * @param  subClass The sub-class; cannot be null.
      * @return True if {@code subClass} is a sub-class of {@code clazz},
      *         false if not.
      * @throws NullPointerException If either argument is null.         
      */
    public final static boolean isSubClass(Class<?> clazz, Class<?> subClass) {
        // Will only return true if subClass is a sub-class of
        // clazz - not if they have common interfaces...
        return getSuperClasses(subClass).contains(clazz);
    }

    /**
      * Returns true if and only if {@code superClass} is a super-class
      * of {@code clazz}, false otherwise. <p> 
      * 
      * For this method to return true, {@code clazz} must at some 
      * point extend {@code superClass}, directly or indirectly. Implementing 
      * common interfaces is not enough.
      * 
      * @param  clazz      The parent class; cannot be null.
      * @param  superClass The super-class; cannot be null.
      * @return True if {@code superClass} is a super-class of {@code clazz},
      *         false if not.
      * @throws NullPointerException If either argument is null.         
      */
    public final static boolean isSuperClass(Class<?> clazz, Class<?> superClass) {
        // Will only return true if subClass is a sub-class of
        // clazz - not if they have common interfaces...
        return getSuperClasses(clazz).contains(superClass);
    }
    
    /**
      * Returns true if each value in {@code values} is assignable to
      * the corresponding type at the same index in {@code types},
      * respectively, false if not. <p>
      * 
      * If {@code objects} is null or contains null values, this method will
      * always return false. <p>
      * 
      * The comparison is done ignoring primitives.
      *  
      * @param  types  The types; cannot be null.
      * @param  values The values; cannot be null.
      * @return True if all values are instances of the corresponding types,
      *         false if not.
      * @throws NullPointerException If {@code types} is null.
      */
    public final static boolean isInstance(Class<?>[] types, Object... values) {
        return isInstance(false, types, values);
    }
    
    /**
      * Returns true if each value in {@code values} is assignable to
      * the corresponding type at the same index in {@code types},
      * respectively, false if not. <p>
      * 
      * If {@code objects} is null and {@code nullValuesMatch} is true
      * and {@code types} is empty, true will be returned; false otherwise.
      * If {@code nullValuesMatch} is true, a null value will match the
      * corresponding type in {@code types} at the same index. <p>
      * 
      * The comparison is done ignoring primitives.
      *
      * @param  nullValuesMatch True if null values in {@code values} will
      *                         count as a match to the corresponding type,
      *                         false if not.
      * @param  types           The types; cannot be null.
      * @param  values          The values; can be null.
      * @return True if all values are instances of the corresponding types,
      *         false if not.
      * @throws NullPointerException If {@code types} is null.
      */
    public final static boolean isInstance(boolean nullValuesMatch, Class<?>[] types, Object... values) {
        if (values == null) {
            return ((nullValuesMatch) && (types.length == 0));
        } else if (types.length != values.length) {
            return false;
        }
        for (int i = 0; i < types.length; i++) {
            Object value = values[i];
            if (value == null) {
                if (nullValuesMatch) {
                    continue;
                }
                return false;
                
            } else if (!Primitive.isAssignableFromIgnorePrimitives(types[i], value.getClass())) {
                return false;
            }
        }
        return true;
    }
    
    /**
      * Returns the constructor declared in the class supplied as 
      * {@code clazz} accepting the formal parameter types supplied
      * as {@code parameterTypes}, ignoring primitive types. <p>
      * 
      * In case several constructors match, the constructor declared
      * first in {@code clazz} is returned.
      *  
      * @param  clazz          The class; cannot be null.
      * @param  parameterTypes The formal parameter types for the constructor
      *                        to fetch.
      * @return The (first) matching constructor; never null.
      * @throws NullPointerException If {@code clazz} is null.
      * @throws NoSuchMethodException If no such constructor can be found.
      */
    public final static Constructor<?> getConstructor(Class<?> clazz, Class<?>... parameterTypes) throws NoSuchMethodException {
        // Try and fetch directly...
        try {
            return clazz.getDeclaredConstructor(parameterTypes);
        } catch (NoSuchMethodException nsme) {
            
            // Check assignable from...
            loop:
            for (Constructor<?> constructor : clazz.getDeclaredConstructors()) {
                Class<?>[] types = constructor.getParameterTypes();
                if (types.length != parameterTypes.length) {
                    continue;
                }
                for (int i = 0; i < types.length; i++) {
                    if (!Primitive.isAssignableFromIgnorePrimitives(types[i], parameterTypes[i])) {
                        continue loop;
                    }
                }
                // Match...
                return constructor;
            }
            // No match - re-throw...
            throw nsme;
        }
    }
    
    /**
      * Returns the copy constructor declared in {@code clazz}, if any. <p>
      * 
      * A copy constructor of {@code clazz} is defined as a constructor 
      * accepting a single parameter of the exact same as {@code clazz}. <p>
      * 
      * The constructor will be made accessible, if not already. 
      *  
      * @param  <T>   The type of {@code class}.
      * @param  clazz The class; cannot be null.
      * @return The copy constructor, or null if no copy constructor is 
      *         declared in {@code clazz}.
      * @throws NullPointerException If {@code clazz} is null.         
      */
    public final static <T> Constructor<T> getCopyConstructor(Class<T> clazz) {
        try {
            Constructor<T> constructor = clazz.getDeclaredConstructor(clazz);
            if (!Modifier.isPublic(constructor.getModifiers())) {
                constructor.setAccessible(true);
            }
            return constructor;
            
        } catch (Exception e ){
            // intentionally left blank...
        }
        return null;
    }
    
    
    /**
      * Returns a collection of the classes and/or interfaces the 
      * {@code Class} supplied as {@code clazz} can be assigned
      * to. <p>
      * 
      * The contained classes either represent super-classes or 
      * implemented interfaces, including those implemented by 
      * super-classes. <p>
      * 
      * The classes and interfaces are added to {@code classes} in a depth 
      * first manner, super-classes precedes sub-classes, but the
      * collection type may or may not respect the insertion order. 
      * Furthermore, a given interface may be included more than one time in 
      * the returned collection, again depending on the type of collection.
      *    
      * @param  <C>     The actual collection type to use.
      * @param  clazz   The class to fetch the assignable classes for; 
      *                 cannot be null. 
      * @param  classes The collection to add to classes and interfaces to; 
      *                 cannot be null.               
      * @return The {@code classes} collection; never null, and never empty
      *         (will always at least contain {@code Object}).
      * @throws NullPointerException If either argument is null.
      */
    public final static <C extends Collection<Class<?>>> C getAssignableTo(Class<?> clazz, C classes) {
        // Super class...
        Class<?> superClass = clazz.getSuperclass();
        if (superClass != null) {
            getAssignableTo(superClass, classes);
        }
        // Interfaces... Why the "Class<T>.getInterfaces()" methods returns
        // Class<?> and not Class<? super T> is beyond me. Must be a legacy 
        // issue or something... Surely, any interface implemented by a given 
        // class, T, will fit the bound <? super T>, or am I missing something?
        for (Class<?> interfaze : clazz.getInterfaces()) {
            getAssignableTo(interfaze, classes);
        }
        // Duplicate check depends on the type of collection used...
        classes.add(clazz);
        return classes;
    }
    
    /**
      * Returns an ordered set of the classes and/or interfaces the 
      * {@code Class} supplied as {@code clazz} can be assigned
      * to. <p>
      * 
      * The contained classes either represent super-classes or 
      * implemented interfaces, including those implemented by 
      * super-classes. <p>
      * 
      * The returned set lists the classes and interfaces in a depth 
      * first manner, i.e., super-classes precedes sub-classes. A given
      * interface is only included once in the set, corresponding
      * to the first time visited.
      *    
      * @param  clazz The class to fetch the assignable classes for; 
      *               cannot be null. 
      * @return An ordered set of the assignable classes/interfaces; never null,
      *         and never empty (will always contain at least {@code Object}).
      * @throws NullPointerException If {@code clazz} is null.
      */
    public final static Set<Class<?>> getAssignableTo(Class<?> clazz) {
        return getAssignableTo(clazz, new LinkedHashSet<Class<?>>());
    }
    
    /**
      * Returns an ordered set of the interfaces the {@code Class} supplied 
      * as {@code clazz} can be assigned to, including interfaces
      * implemented by super-classes of {@code clazz}. <p>
      * 
      * The returned set lists the interfaces in a depth first
      * manner, i.e., interfaces from super-classes precedes interfaces
      * from sub-classes. The order per class is the declaration order. 
      * A given interface is only included once in the set, corresponding 
      * to the first time visited.
      *    
      * @param  clazz The class to fetch the assignable interfaces for; 
      *               cannot be null. 
      * @return An ordered set of the assignable interfaces; never null,
      *         but can be empty.
      * @throws NullPointerException If {@code clazz} is null.
      */
    public final static Set<Class<?>> getInterfaces(Class<?> clazz) {
        return getInterfaces(clazz, new LinkedHashSet<Class<?>>());
    }
    
    /**
      * Returns an ordered set of the interfaces the {@code Class} supplied 
      * as {@code clazz} can be assigned to, including interfaces
      * implemented by super-classes of {@code clazz}. <p>
      * 
      * The ordering of the interfaces is determined by the collection
      * supplied as {@code interfaces}, but the interfaces will be
      * added in a depth-first manner, i.e., interfaces from super-classes 
      * will be added before interfaces from sub-classes. The order per class 
      * is the declaration order. <p>
      * 
      * If {@code interfaces} is a non-set collection, duplicates may occur.
      *    
      * @param  <C>        The actual collection type to use.
      * @param  clazz      The class to fetch the assignable interfaces for; 
      *                    cannot be null. 
      * @param  interfaces The collection to add the interfaces to; 
      *                    cannot be null.               
      * @return The {@code interfaces} collection; never null, but can be empty.
      * @throws NullPointerException If either argument is null.
      */
    public final static <C extends Collection<Class<?>>> C getInterfaces(Class<?> clazz, C interfaces) {
        getAssignableTo(clazz, interfaces);
        // Remove all non-interface classes...
        for (Iterator<Class<?>> i = interfaces.iterator(); i.hasNext(); ) {
            if (!i.next().isInterface()) {
                i.remove();
            }
        }
        return interfaces;
    }

    /**
      * Returns all declared methods in {@code clazz}, or in
      * a super-class or implemented interface. The methods
      * are returned in a depth-first order, i.e. methods
      * from super-classes or interfaces precede those from
      * sub-classes ({@code clazz}). <p>
      * 
      * A unique method is only included once, but overridden 
      * versions may be included as well.
      * 
      * @param  clazz The class to fetch the methods for; cannot be null.
      * @return The list of methods; never null, but can be empty if
      *         a {@code clazz} represents a <i>marker</i> interface.
      * @throws NullPointerException If {@code clazz} is null.           
      */
    public final static List<Method> getMethods(Class<?> clazz) {
        List<Method> methods = new ArrayList<Method>(); 
        // Find all relevant classes, and for each class fetch
        // all declared methods; each class will only be included
        // one time in the collection of classes, because we 
        // know that "getAssignableTo" uses a linked hash set by
        // default...
        for (Class<?> c : getAssignableTo(clazz)) {
            // asList does not copy!
            methods.addAll(Arrays.asList(c.getDeclaredMethods()));
        }
        return methods;
    }
    
    /**
      * Returns all declared fields in {@code clazz}, or in
      * a super-class or implemented interface. The fields
      * are returned in a depth-first order, i.e. fields
      * from super-classes or interfaces precede those from
      * sub-classes ({@code clazz}). <p>
      * 
      * @param  clazz The class to fetch the fields for; cannot be null.
      * @return The list of fields; never null, but can be empty.
      * @throws NullPointerException If {@code clazz} is null.           
      */
    public final static List<Field> getFields(Class<?> clazz) {
        List<Field> fields = new ArrayList<Field>(); 
        // Find all relevant classes, and for each class fetch
        // all declared fields; each class will only be included
        // one time in the collection of classes, because we 
        // know that "getAssignableTo" uses a linked hash set by
        // default...
        for (Class<?> c : getAssignableTo(clazz)) {
            // asList does not copy!
            fields.addAll(Arrays.asList(c.getDeclaredFields()));
        }
        return fields;
    }
    
    /**
      * Returns the result of <i>invoking</i> the accessible
      * member supplied as {@code member}. <p>
      * 
      * The member can either be a method, constructor, or
      * field. <p> 
      * 
      * If the member is a non-static method, constructor, or non-static
      * field, {@code target} will be used as the target instance. <p>
      * 
      * The member will be made accessible if not already. <p>
      * 
      * Checked exceptions are suppressed to runtime exceptions.
      *
      * @param  <E>       The type of member.
      * @param  target    The target if {@code member} is non static;
      *                   can be null if {@code member} is static.
      * @param  member    The member to invoke; cannot be null.
      * @param  arguments The arguments for {@code member}; can be null
      *                   if {@code member} allows it. Unused if 
      *                   {@code member} is a field.
      * @return The result of invoking {@code member}; can be null.
      * @throws NullPointerException If {@code member} is null, or if
      *         {@code target} and/or {@code arguments} is null but
      *         not allowed by {@code member}.
      * @throws RuntimeException If the invocation fails.
      */
    public final static <E extends AccessibleObject & Member> Object invoke(Object target, E member, Object... arguments) {
        try {
            int modifiers = member.getModifiers();
            if (!Modifier.isPublic(modifiers)) {
                if (!member.isAccessible()) {
                    member.setAccessible(true);
                }
            }
            
            // Method...
            if (member instanceof Method) {
                return ((Method)member).invoke(Modifier.isStatic(modifiers) ? null : target, arguments);
                
            // Constructor...
            } else if (member instanceof Constructor) {
                return ((Constructor<?>)member).newInstance(arguments);
                
            // Field...
            } else if (member instanceof Field) {
                return ((Field)member).get(Modifier.isStatic(modifiers) ? null : target);
                
            // Should never happen (only other known member is Class, which is
            // not an accessible object)...
            } else {
                throw new IllegalArgumentException("Illegal member type: " + member);
            }
            
        // To runtime, if not already...
        } catch (Exception e) {
            throw toRuntimeException(e);
        }
    }

    /**
      * Returns the first <i>reachable</i> <b>named</b> class or interface
      * that can identify {@code clazz}. <p>
      * 
      * If {@code clazz} is null, then null is returned. If 
      * {@code clazz} is an anonymous class, the first non-anonymous
      * super class is used <b>or</b> the first and only 
      * interface implemented.
      *  
      * @param  clazz The class; can be null.
      * @return The class to use, or null.
      */
    public final static Class<?> getClass(Class<?> clazz) {
        while (clazz != null && clazz.isAnonymousClass()) {
            Class<?>[] interfaces = clazz.getInterfaces();
            if (interfaces.length == 1) {
                return interfaces[0];
            } else {
                clazz = clazz.getSuperclass();
            }
        }
        return clazz;
    }
    
}
