/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.reflect.proxy;

import static dk.rode.thesis.meta.util.Util.add;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Set;

import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.reflect.Reflection;
import dk.rode.thesis.meta.util.Primitive;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>reference handler</i> is the standard implementation of 
  * the {@link Reference} interface. <p> 
  * 
  * <b>Implementation notes</b>:<br>
  * The copy method cannot be a varargs method. The mutator methods 
  * are only matched on their name, regardless of declaring class. <p>
  * 
  * There are potential issues with {@code equals} because the original
  * class information is lost for proxies, i.e. the class becomes some
  * sub-class of {@link java.lang.reflect.Proxy java.lang.reflect.Proxy}.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    ReferenceInvocationHandler
  * @see    ProxyFactory#getSharedObject(Object, String, Class[], String...)
  * 
  * @param  <T> The type of the referenced object.
  */
public class ReferenceHandler<T> implements Reference<T> {

    /**
      * The proxy factory used. <p>
      * 
      * Never null.
      */
    private final ProxyFactory factory;
    /**
      * The actual referenced object. <p>
      * 
      * Never null.
      */
    private T reference;
    /**
      * The number of references to {@link #reference}.
      */
    private int referenceCount;
    /**
      * The <i>copy</i> method. <p>
      * 
      * Never null.
      */
    private final Method copyMethod;
    /**
      * The names of the <i>mutator</i> methods that will force
      * a copy of {@link #reference} before executed. <p>
      * 
      * The names do not include information about the declaring
      * class! <p>
      * 
      * Never null, and never empty.
      */
    private final Set<String> mutatorMethods;
    /**
      * All interfaces implemented by {@link #reference}, including
      * the adapted {@link Reference} interface at index zero. <p>
      * 
      * Never null.
      */
    private final Class<?>[] interfaces;
    
    /**
      * Constructor, which creates this reference to manage the referenced
      * object supplied as {@code reference}. <p>
      * 
      * The reference count is set to 1.
      * 
      * @param  factory        The proxy factory to create the proxies; cannot 
      *                        be null.
      * @param  reference      The referenced object; cannot be null.
      * @param  copyMethod     The name of the no-arg <i>copy</i> method to use; 
      *                        cannot be null. Must be declared in the class of 
      *                        {@code reference}, but will be made accessible 
      *                        if not already. 
      * @param  mutatorMethods The names of the <i>mutator</i> methods that will 
      *                        force a copy of {@code reference} before executed;
      *                        cannot be null or empty.
      * @throws NullPointerException If either argument is null.
      * @throws IllegalArgumentException If {@code mutatorMethods} is empty.
      * @throws Exception If no such copy method exists.        
      */
    public ReferenceHandler(ProxyFactory factory, T reference, String copyMethod, String... mutatorMethods) throws Exception {
        this(factory, reference, copyMethod, null, mutatorMethods);
    }
    
    /**
      * Constructor, which creates this reference to manage the referenced
      * object supplied as {@code reference}. <p>
      * 
      * The reference count is set to 1.
      * 
      * @param  factory        The proxy factory to create the proxies; cannot 
      *                        be null.
      * @param  reference      The referenced object; cannot be null.
      * @param  copyMethod     The name of the <i>copy</i> method to use; cannot 
      *                        be null. Must be declared in the class of 
      *                        {@code reference}, but will be made accessible 
      *                        if not already. 
      * @param  parameterTypes The formal parameter types of {@code copyMethod},
      *                        if any, ignoring primitive types; can be null.                        
      * @param  mutatorMethods The names of the <i>mutator</i> methods that will 
      *                        force a copy of {@code reference} before executed;
      *                        cannot be null or empty.
      * @throws NullPointerException If {@code factory}, {@code reference}, or
      *         {@code copyMethod} are null.
      * @throws IllegalArgumentException If {@code mutatorMethods} is empty, or
      *         if the copy method is a vararg method.
      * @throws Exception If no such copy method exists.        
      */
    public ReferenceHandler(ProxyFactory factory, T reference, String copyMethod, Class<?>[] parameterTypes, String... mutatorMethods) throws Exception {
        if ((this.factory = factory) == null) {
            throw new NullPointerException("Proxy factory cannot be null");
        } 
        this.copyMethod = 
            Reflection.getMethod(reference.getClass(), copyMethod, parameterTypes);
        if (this.copyMethod.isVarArgs()) {
            throw new IllegalArgumentException("Cannot handle varargs copy method");
        }
        int modifiers = this.copyMethod.getModifiers();
        if (!Modifier.isPublic(modifiers)) {
            this.copyMethod.setAccessible(true);
        }
        this.reference = reference;
        
        add(this.mutatorMethods = new HashSet<String>(), mutatorMethods);
        if (this.mutatorMethods.size() == 0) {
            throw new IllegalArgumentException("Mutator methods cannot be null");
        }
        // Order of interfaces becomes important because "reference" also
        // have a copy method! We adapt "reference" to support the 
        // Reference interface as the first interface...
        LinkedHashSet<Class<?>> interfaces = new LinkedHashSet<Class<?>>();
        interfaces.add(Reference.class);
        // Fetch all interfaces from "reference"...
        Reflection.getInterfaces(reference.getClass(), interfaces);
        interfaces.toArray(this.interfaces = new Class<?>[interfaces.size()]);
        
        this.referenceCount = 1;
        
        LogFactory.getLog(this).println("Created reference: ", this);
    }

    /**
      * Constructor, which creates this reference to manage a <b>copy</b> of 
      * the referenced object managed by {@code handler}. <p> 
      * 
      * The reference count is set to 1.
      * 
      * @param  handler   The reference handler to copy; cannot be null.
      * @throws NullPointerException If {@code handler} is null.
      * @throws Exception If the copying fails.
      */
    private ReferenceHandler(ReferenceHandler<T> handler) throws Exception {
        // Copy actual object...
        Object copy = Reflection.invoke(handler.reference, handler.copyMethod);
        // This is potentially dangerous, but "cast(copy)" ensures that we
        // will throw an exception if the copy is not assignable to
        // the type of the original reference. The copy can still be a 
        // sub-type of T, though!
        @SuppressWarnings("unchecked") T reference = 
            (T)handler.reference.getClass().cast(copy);
        this.reference = reference;
        this.copyMethod = handler.copyMethod;
        this.factory = handler.factory;
        this.mutatorMethods = handler.mutatorMethods;
        this.interfaces = handler.interfaces;
        this.referenceCount = 1;
        
        LogFactory.getLog(this).println("Copied existing reference: ", this);
    }

    /**
      * 
      */
    public T getReference() {
        return this.reference;
    }
    
    /**
      * 
      */
    public synchronized T addReference() {
        this.referenceCount++;
        LogFactory.getLog(this).println("Increased reference count to: ", this.referenceCount, " -> ", this);
        // Create a new proxy using this reference handler as well...
        return this.toProxy();
    }
    
    /**
      * 
      */
    public boolean copyMethod(Method method) {
        return ((this.copyMethod.getName().equals(method.getName())) &&
                (Primitive.isAssignableFromIgnorePrimitives(this.copyMethod.getParameterTypes(), method.getParameterTypes())));
    }
    
    /**
      * 
      */
    public synchronized Reference<T> copyIfMutator(Method method) throws Exception {
        if (this.referenceCount > 1) {
            if (this.mutatorMethods.contains(method.getName())) {
                this.referenceCount--;
                LogFactory.getLog(this).println("Decreased reference count to: ", this.referenceCount, " -> ", this);
                // Create a new reference handler to use, which will copy
                // the referenced object...
                return new ReferenceHandler<T>(this);
            }
        }
        return this;
    }

    /**
      * Creates a new proxy instance using this reference handler. <p>
      * 
      * This method does not alter the reference count of this
      * reference.
      * 
      * @return A new proxy instance; never null.
      */
    public T toProxy() {
        // We do not register the created proxy in the proxy factory, because
        // we change the actual reference on copying...
        return this.factory.getProxy((T)null, // NetBeans complains otherwise  
                                     new ReferenceInvocationHandler(this), 
                                     this.interfaces);
    }
    
    /**
      * Returns the string representation of this handler.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        return sb.append("; reference: ").append(this.reference)
                 .append("; count: ").append(this.referenceCount)
                 .append("; copy-method: ").append(this.copyMethod.getName())
                 .append("; mutator-methods: ").append(this.mutatorMethods)
                 .append(']').toString();
    }
}
