/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.meta.test;

import static dk.rode.thesis.meta.reflect.Reflection.isOverriding;

import java.io.Closeable;
import java.io.IOException;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Proxy;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import dk.rode.thesis.bridge.MemorizableSequenceAbstraction;
import dk.rode.thesis.bridge.SequenceAbstraction;
import dk.rode.thesis.bridge.SequenceValueArrayList;
import dk.rode.thesis.bridge.SequenceValueHashSet;
import dk.rode.thesis.bridge.SequenceValueLinkedHashSet;
import dk.rode.thesis.bridge.SequenceValueRange;
import dk.rode.thesis.bridge.SequenceValueTreeSet;
import dk.rode.thesis.decorator.AppenderDecorator;
import dk.rode.thesis.decorator.DuplexDecorator;
import dk.rode.thesis.decorator.UppercaseDecorator;
import dk.rode.thesis.facade.FibonacciSequence;
import dk.rode.thesis.facade.RandomSequence;
import dk.rode.thesis.flyweight.CharacterFactory;
import dk.rode.thesis.memento.MemorizableEnglishAlphabetSequence;
import dk.rode.thesis.memento.MemorizableSequence;
import dk.rode.thesis.memento.RangeSequence;
import dk.rode.thesis.memento.SequenceMemento;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.DigitSequence;
import dk.rode.thesis.meta.model.EnglishAlphabetSequence;
import dk.rode.thesis.meta.model.HexDigitSequence;
import dk.rode.thesis.meta.model.LongSequence;
import dk.rode.thesis.meta.model.PrimeSequence;
import dk.rode.thesis.meta.model.ReversibleSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.model.Sequence.State;
import dk.rode.thesis.meta.util.Arguments;
import dk.rode.thesis.meta.util.Util;
import dk.rode.thesis.observer.DateSequence;
import dk.rode.thesis.observer.ObservableSequence;
import dk.rode.thesis.prototype.CountdownSequence;
import dk.rode.thesis.prototype.SymbolSequence;
import dk.rode.thesis.proxy.SequenceProxyFactory;
import dk.rode.thesis.proxy.SynchronisedSequence;
import dk.rode.thesis.singleton.DanishAlphabetSequence;
import dk.rode.thesis.singleton.MutatedSimpsonsFamilySequence;
import dk.rode.thesis.singleton.NorwegianAlphabetSequence;
import dk.rode.thesis.singleton.SimpsonsAndBouvierFamilySequence;
import dk.rode.thesis.singleton.SimpsonsFamilySequence;
import dk.rode.thesis.singleton.SmileySequence;
import dk.rode.thesis.state.ReversiblePrimeSequence;
import dk.rode.thesis.state.StateableSequence;
import dk.rode.thesis.state.StepSequence;
import dk.rode.thesis.strategy.SequencePolicy;
import dk.rode.thesis.templatemethod.FileSequence;
import dk.rode.thesis.templatemethod.NegativeSequence;
import dk.rode.thesis.templatemethod.ZipSequence;
import dk.rode.thesis.visitor.TypeVisitableSequence;
import dk.rode.thesis.visitor.ValueVisitableSequence;

/**
  * <b>Sequence integrity</b> tests.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class IntegrityTests implements Testable {

    /**
      * Executes the <b>sequence integrity</b> tests.
      * 
      * @param  args The arguments to supply to {@link #test(Log, Arguments)},
      *              if any; can be null.
      * @throws Exception If the sequence integrity tests fail.            
      */
    public static void main(String[] args) throws Exception {
        new IntegrityTests().test(LogFactory.getLog(), new Arguments(args));
    }
    
    /**
      * 
      */
    public boolean test(Log out, Arguments arguments) throws Exception {
        boolean ok = true;
    
        // Number of iterations to advance each sequence while testing it...
        int iterations = 10000;
        // Number of maximum errors per tested sequence before we bail out...
        int maxErrors = 100;
        
        out.heading("Sequence integrity tests...");

        // Various factories...
        
        CharacterFactory cf = new CharacterFactory();
        SequenceProxyFactory spf = new SequenceProxyFactory();
        DateFormat df = new SimpleDateFormat("dd.MM.yyyy");
        
        // Turn of verbose logging for dynamic proxies...
        LogFactory.getLog(InvocationHandler.class).setActive(false);
        
        // A number of different sequences have been implemented, many
        // in different packages to illustrate aspects of a given
        // pattern. Their specific use are tested in the respective
        // packages!
        
        // Sequence values for sequence abstractions...
        Integer[] values = {1, -3, 2, 5, 3, 4, -1, -4, 2, 3, 4, -5, 4, 3, 2, -4, 1, 0, -1, -2};
        List<Integer> list = Arrays.asList(values); 

        // Sorted according to natural ordering, with duplicates...
        ArrayList<Integer> sorted = new ArrayList<Integer>(list);
        Collections.sort(sorted, null);
        
        // Created sequence types
        Sequence<?>[] sequences = {

            // Bridge package...
            
            // bounded, non-unique, consistent
            new SequenceAbstraction<Integer>(
                new SequenceValueArrayList<Integer>(Integer.class, new ArrayList<Integer>(list))
            ),
            // bounded, non-unique, consistent, sorted
            new SequenceAbstraction<Integer>(
                new SequenceValueArrayList<Integer>(Integer.class, new ArrayList<Integer>(sorted), true, true)
            ),
            // bounded, unique, inconsistent 
            new SequenceAbstraction<Integer>(
                new SequenceValueHashSet<Integer>(Integer.class, new HashSet<Integer>(list))
            ),
            // bounded, unique, consistent
            new SequenceAbstraction<Integer>(
                new SequenceValueLinkedHashSet<Integer>(Integer.class, new LinkedHashSet<Integer>(list))
            ),
            // bounded, unique, consistent
            new SequenceAbstraction<Integer>(
                new SequenceValueTreeSet<Integer>(Integer.class, new TreeSet<Integer>(list))                    
            ),
            // bounded, unique, consistent
            new SequenceAbstraction<Integer>(new SequenceValueRange(-5, 5)),
            // unbounded, unique, consistent
            new SequenceAbstraction<Integer>(new SequenceValueRange(-5)),
            
            // The following 7 are identical to the previous 7 except memorizable...
            new MemorizableSequenceAbstraction<Integer>(
                new SequenceValueArrayList<Integer>(Integer.class, new ArrayList<Integer>(list))
            ),
            new MemorizableSequenceAbstraction<Integer>(
                new SequenceValueArrayList<Integer>(Integer.class, new ArrayList<Integer>(sorted), true, true)
            ),
            new MemorizableSequenceAbstraction<Integer>(
                new SequenceValueHashSet<Integer>(Integer.class, new HashSet<Integer>(list))
            ),
            new MemorizableSequenceAbstraction<Integer>(
                new SequenceValueLinkedHashSet<Integer>(Integer.class, new LinkedHashSet<Integer>(list))
            ),
            new MemorizableSequenceAbstraction<Integer>(
                new SequenceValueTreeSet<Integer>(Integer.class, new TreeSet<Integer>(list))                    
            ),
            new MemorizableSequenceAbstraction<Integer>(new SequenceValueRange(-5, 5)), 
            new MemorizableSequenceAbstraction<Integer>(new SequenceValueRange(-5)), 
            
            // Decorator package...
            new AppenderDecorator(new EnglishAlphabetSequence(), "<<", ">>"),
            new DuplexDecorator(new EnglishAlphabetSequence()),
            new UppercaseDecorator(new EnglishAlphabetSequence()),
            
            // Facade package...
            new FibonacciSequence(),
            new PrimeSequence(100),
            new RandomSequence(10),
            
            // Flyweight package...
                
            // unique sentence                
            cf.getSentence("This-is_a|small~test!"),            
            // not unique sentence (space, , and very)...
            cf.getSentence("This is, however, a very, very long test!"), 
            // unique word...
            cf.getWord("Rode"),                                           
            // not unique word (n)...
            cf.getWord("Gunni"),                                
            
            // Memento package...
            new MemorizableEnglishAlphabetSequence(),
            new RangeSequence(100),     // unbounded
            new RangeSequence(100,110), // bounded
            
            // Model package (common sequences)...
            new DigitSequence(),
            new EnglishAlphabetSequence(),
            new HexDigitSequence(),
            new LongSequence(10),
            
            // Observer package...
            new DateSequence(df.parse("01.01.2008")), // unbounded             
            new DateSequence(df.parse("01.01.2008"),  // bounded
                             df.parse("10.05.2008")),  
                             
            // Prototype package...
            CountdownSequence.getInstance(),
            new SymbolSequence(),
            
            // Proxy package... Cannot use non-resettable or immutable
            // sequences since they intentionally break the invariants
            // we are testing here...
            new SynchronisedSequence<Long>(new LongSequence(20)),
            spf.getSynchronisedSequence(new LongSequence(30)),
            spf.getLoggableSequence(new LongSequence(30)),
            
            // Singleton package...
            DanishAlphabetSequence.Instance,
            NorwegianAlphabetSequence.getInstance(),
            SmileySequence.getSequence(),
            SimpsonsFamilySequence.getFamily(),
            SimpsonsAndBouvierFamilySequence.getFamily(),
            new MutatedSimpsonsFamilySequence(),
            
            // State package...
            new ReversiblePrimeSequence(100),
            StepSequence.Factory.create(13,-11),
            
            // Template Method package...
            new NegativeSequence(),         // unbounded
            new NegativeSequence(-17),      // bounded
            new FileSequence(Util.getDataDirectory(), null), // all files  
            new ZipSequence(Util.getDataDirectory(), "foobar.zip")
        };
        
        // Print sequence types...
        out.heading("Sequence types and variants");
        out.println("(excluding most adapters, decorators, proxies, and composites)");
        out.println();
        
        for (int i = 0; i < sequences.length; i++) {
            Sequence<?> sequence = sequences[i];
            out.println("Sequence ", i, ": ", sequence.toString(SequencePolicy.VERBOSE));
        }

        // Print sequence types...
        out.heading("Testing sequences' state");
        out.println("Using ", iterations, " iterations per sequence");
        out.println("Using ", iterations * 2, " iterations per reversible sequence");
        out.println();
        out.println("Bailing out after ~", maxErrors, " errors per sequence!");
        out.println();
        
        // True to enable class related log output, if any (verbose)...
        boolean log = false;
        
        for (int i = 0; i < sequences.length; i++) {
            ok = test(out, log, i, sequences[i], iterations, maxErrors) && ok;
        }
        
        out.heading("Sequence integrity tests ", ok ? "ok" : "failed", "!");
        
        return ok;
    }

    /**
      * Performs an integrity test of the {@code Sequence} supplied 
      * as {@code sequence}, and returns true upon success or false 
      * if the test failed.
      *  
      * @param  <E>        The type of values delivered by {@code sequence}.
      * @param  out        The log to use; never null.
      * @param  log        True to enable verbose logging for the type of 
      *                    {@code sequence}. 
      * @param  index      The index of {@code sequence} among all tested sequence
      *                    instances.
      * @param  sequence   The sequence to test; never null.
      * @param  iterations Number of times to invoke {@code next()} on {@code sequence}.
      * @param  maxErrors  Maximum number of allowed errors before bailing out of
      *                    the test.
      * @return True if the test succeeded, false if it failed.
      */
    private <E> boolean test(Log out, boolean log, int index, Sequence<E> sequence, int iterations, int maxErrors) {
        // Update state of individual logs associated with the type of "sequence",
        // if any...
        LogFactory.getLog(sequence).setActive(log);
        int i = -1, errors = 0;
        try {
            out.println();
            out.println("Testing sequence ", index, ": ", sequence.toString(SequencePolicy.VERBOSE), " -> ", sequence);
            
            // PROPERTIES
            if (sequence.bounded()) {
                out.println("- is bounded");
            } else {
                out.println("- is unbounded");
            }
            if (sequence.consistent()) {
                out.println("- is consistent");
            } else {
                out.println("- is inconsistent");
            }
            if (sequence.unique()) {
                out.println("- is unique");
            } else {
                out.println("- is not unique");
            }
            SequenceMemento<E> memento = null;
            boolean memorizable = (sequence instanceof MemorizableSequence);
            if (memorizable) {
                out.println("- is memorizable");
            }
            if (sequence instanceof StateableSequence) {
                out.println("- is stateable");
            }
            boolean reversible = (sequence instanceof ReversibleSequence);
            if (reversible) {
                out.println("- is reversible");
                // next() #iterations, and previous() #iterations 
                iterations *= 2; 
            }
            if (sequence instanceof ObservableSequence) {
                out.println("- is observable");
            }
            if (sequence instanceof ValueVisitableSequence) {
                out.println("- is value visitable");
            }
            if (sequence instanceof TypeVisitableSequence) {
                out.println("- is type visitable");
            }
            if (sequence instanceof Closeable) {
                out.println("- is closeable");
            }
            if (Proxy.isProxyClass(sequence.getClass())) {
                out.println("- is dynamic proxy");
            }
            out.println();
            
            // START PHASE: neither reset() or next() has been invoked, but 
            // current() may have (i.e. in toString)...
            State state = sequence.state(), lastState = null;
            if (state != State.START) {
                out.error("Sequence has illegal initial state: ", state);
                errors++;
            }
            
            // A sequence always have a value...
            E initialValue = sequence.current();
            if (initialValue == null) {
                out.error("Sequence has no initial value");
                errors++;
            }

            int reset = iterations / 8;
            int memorise = iterations / 5;
            int reverse = iterations / 3;

            out.println("- Reset every ", reset, " iterations");
            if (memorizable) {
                out.println("- Memorised OR updated every ", memorise, " iterations");
            }
            if (reversible) {
                out.println("- Reversed, if possible, every ", reverse, " iterations");
            }
            out.println();
            
            // Set containing the unique values for unique sequences
            // in between restart and reset...
            Set<E> unique = new HashSet<E>();
            unique.add(initialValue);
            
            // ADVANCING: state can never be anything besides NORMAL or
            // RESTART for bounded sequences...
            for (i = 0; errors <= maxErrors && i < iterations; i++) {
                // Get (previous) current value...
                E current = sequence.current();
                // Value...
                E value = sequence.next();
                if (value == null) {
                    out.error("Null value at iteration ", i);
                    errors++;
                    
                // We have advanced the sequence, so next() cannot be
                // equal to old current if unbounded. If bounded, this
                // is allowed if the sequence contains a single value.
                // However, the invariant used here is that all sequences
                // tested have more than one element; hence, current()
                // must not equals next() (in that order)...
                } else if ((Util.equals(value, current)) && (sequence.unique())) {
                    out.error("Next value equals previous current value at iteration ", i, ": ", value);
                    errors++;
                }
                
                // State...
                state = sequence.state();
                // Get current value - must not change state...
                current = sequence.current();
                if (!value.equals(current)) {
                    out.error("Current value different from the last returned value at iteration ", i);
                    errors++;
                }
                
                // State must not have changed based on a call to current()...
                if (state != sequence.state()) {
                    out.error("current() invocation changed internal state at iteration ", i);
                    errors++;
                }
                // State must be valid...
                if ((state == State.START) || 
                    (state == State.RESET) ||
                    (state == State.RESTART) && ((!sequence.bounded()) && (!sequence.consistent()))) {
                    out.error("Illegal state at iteration ", i, ": ", state);
                    errors++;
                }
                
                // Check consistent properties... 
                if ((sequence.consistent()) && (!reversible)) {
                    if ((state == State.RESET) || (state == State.RESTART)) {
                        if (!Util.equals(value, initialValue)) {
                            out.error("Initial value not reset at iteration ", i, ": ", state, " -> ", initialValue, " != ", value);
                            errors++;
                        }
                    }
                }
                // Reset unique values...
                if (state == State.RESTART) {
                    unique.clear();
                }
                
                // Check unique values...
                if (sequence.unique()) {
                    if (unique.contains(value)) {
                        out.error("Unique sequence delivered duplicate value at iteration ", i, ": ", value);
                        errors++;
                    }
                    unique.add(value);
                }
                
                if (state.equals(lastState)) {
                    if (state != State.NORMAL) {
                        out.error("Erroneous state at iteration ", i, ": ", state);
                        errors++;
                    }
                }
                
                lastState = state;
                
                // Reset...
                if (i % reset == reset - 1) {
                    unique.clear();
                    sequence.reset();
                    
                    // State should reflect reset...
                    if (State.RESET != sequence.state()) {
                        out.error("Illegal state at iteration ", i, " after reset: ", state);
                        errors++;
                    }
                    
                    
                // Memorizable...
                } else if ((memorizable) && (i % memorise == memorise - 1)) {
                    MemorizableSequence<E> ms = (MemorizableSequence<E>)sequence;
                    if (memento != null) {
                        ms.setMemorizableState(memento);
                        memento = null;
                        
                        // We do not know which state we set, so we have to clear
                        // the unique map...
                        if (sequence.unique()) {
                            unique.clear();
                        }
                    } else {
                        memento = ms.getMemorizableState();
                    }
                }
                // Reverse sequence, if possible...
                if ((reversible) && (i % reverse == reverse - 1)) {
                    ReversibleSequence<E> rs = (ReversibleSequence<E>)sequence;
                    if (rs.reversible()) {
                        rs.reverse();
                        if (!Util.equals(current, rs.current())) {
                            out.error("Current value changed after reversed at iteration ", i, ": ", current, " != ", rs.current());
                            errors++;
                        }
                        unique.clear();
                    }
                }
            }
            
            // COPY (prototype)...
            Sequence<E> copy = sequence.copy();
            boolean overridden = ((isOverriding(sequence.getClass(), "equals", Object.class)) && (isOverriding(sequence.getClass(), "hashCode")));
            if (overridden) {
                out.warn("\"hashCode()\" and \"equals(Object)\" overridden");
            }
            if (copy == null) {
                out.error("Copy returned null");
                errors++;
                
            } else if (copy == sequence) {
                out.warn("Copy same instance as sequence: ", copy.toString(SequencePolicy.VERBOSE));
                
            } else if (!copy.equals(sequence)) {
                if (overridden) {
                    out.warn("Copy not equivalent: ", copy.toString(SequencePolicy.VERBOSE), " (copy) != ", sequence.toString(SequencePolicy.VERBOSE));
                }
            }
            
            // CONSISTENT...
            if ((sequence.consistent()) && (sequence != copy)) {
                for (int j = 0; errors <= maxErrors && j < iterations; j++) {
                    if (sequence.state() != copy.state()) {
                        out.error("Sequence not consistent, different state between it and copy: ", sequence.state(), " != ", copy.state());
                        errors++;
                    }
                    if (!Util.equals(sequence.current(), copy.current())) {
                        out.error("Sequence not consistent, different current value between it and copy: ", sequence.current(), " != ", copy.current());
                        errors++;
                    }
                    if (!Util.equals(sequence.next(), copy.next())) {
                        out.error("Sequence not consistent, different next value between it and copy: ", sequence.current(), " != ", copy.current());
                        errors++;
                    }
                    // Reset...
                    if (i % reset == reset - 1) {
                        sequence.reset();
                        copy.reset();
                    }
                }
            }
            
        } catch (Throwable t) {
            out.error("Exception thrown ", i > -1 ? "at iteration " + i : "before iteration", ": ", t);
            errors++;
        }

        // close...
        if (sequence instanceof Closeable) {
            try {
                ((Closeable)sequence).close();
            } catch (IOException ioe) {
                out.error("Could not close closeable sequence: ", ioe);
                errors++;
            }
        }
        
        if (errors >= maxErrors) {
            out.warn("Bailing out because of maximum number of errors reached: ", errors, " > ", maxErrors);
        }
        if (errors == 0) {
            out.println("Test succeeded!");
        } else {
            out.error("Test failed: ", errors, " error(s)!");
        }
        return errors == 0;
    }
        
}
