/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.util;

import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>placeholder</i> for a key with an associated value, where a non-null
  * key or value can only be set once. <p>
  * 
  * This class is not thread-safe!
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  *
  * @param  <K> The type of the key.
  * @param  <V> The type of value associated with the key.
  */
public class KeyValue<K,V> {

    /**
      * The key. <p>
      * 
      * Can be null if not set.
      */
    private K key;
    /**
      * The value, which can only be non-null if a {@link #key} is also set. <p>
      * 
      * Can be null if not set.
      */
    private V value;
    
    /**
      * Constructor.
      * 
      * @param key The key; can be null.
      */
    public KeyValue(K key) {
        this(key, null);
    }
    
    /**
      * Constructor.
      * 
      * @param  key   The key; can be null.
      * @param  value The value; can be null, and not allowed to be non-null if 
      *               {@code key} is null.
      * @throws IllegalStateException If {@code key} is null and {@code value}
      *         is not.              
      */
    public KeyValue(K key, V value) {
        this.setKey(key);
        this.setValue(value);
    }
        
    /**
      * Copy constructor.
      * 
      * @param  keyValue The (key,value) placeholder to copy; cannot be null.
      * @throws NullPointerException If {@code keyValue} is null.              
      */
    public KeyValue(KeyValue<? extends K,? extends V> keyValue) {
        if (keyValue == null) {
            throw new NullPointerException("Key value cannot be null");
        }
        this.key = keyValue.key;
        this.value = keyValue.value;
    }
    
    /**
      * Returns a new (key,value) placeholder with a value of null and a key
      * value matching the key value of this (key,value) placeholder.
      *  
      * @return A new (key,value) placeholder; never null.
      */
    public KeyValue<K,V> copyKey() {
        return new KeyValue<K,V>(this.key);
    }
    
    /**
      * Returns the key of this (key,value) placeholder.
      *  
      * @return The key; can be null if not set.
      */
    public K getKey() {
        return this.key;
    }
    
    /**
      * Sets the key of this (key,value) placeholder to {@code key}.
      * 
      * @param  key The key to set; can be null.
      * @return This (key,value) placeholder; never null.
      * @throws IllegalStateException If a non-null key has already been 
      *         set. 
      */
    public KeyValue<K,V> setKey(K key) {
        if (this.key != null) {
            throw new IllegalStateException("A key is already set");
        }
        this.key = key;
        return this;
    }
    
    /**
      * Returns the value of this (key,value) placeholder.
      *  
      * @return The value; can be null if not set, but never null if non-null a 
      *         {@link #getKey() key} is set. 
      */
    public V getValue() {
        return this.value;
    }
    
    /**
      * Sets the value of this (key,value) placeholder to {@code value}.
      * 
      * @param  value The value to set; can be null.
      * @return This (key,value) placeholder; never null.
      * @throws IllegalStateException If a non-null key has not been set, or if a non-null
      *         value has been set.
      */
    public KeyValue<K,V> setValue(V value) {
        if (this.key == null) {
            throw new IllegalStateException("No key has been set");
        } else if ((this.value != null) && (this.value != value)) {
            throw new IllegalStateException("A value is already set");
        }
        this.value = value;
        return this;
    }
    
    /**
      * Returns the hash code of this (key,value) placeholder.
      * 
      * @return The hash code. 
      */
    public int hashCode() {
        return this.key == null ? 0 : this.key.hashCode();
    }
    
    /**
      * Returns true if {@code o} is a (key,value) placeholder
      * having a key that is equivalent to the key of this (key,value) 
      * placeholder, regardless of the associated values, false if not.
      * 
      * @param  o The object to test; can be null.
      * @return True if equal, false if not.
      */
    public boolean equals(Object o) {
        if (!(o instanceof KeyValue)) {
            return false;
        }
        return this.key.equals(((KeyValue<?,?>)o).key);
    }
    
    /**
      * Returns the string representation of this (key,value) placeholder.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.IDENTITY_OPEN.toString(this);
        if (this.key != null) {
            sb.append("; key: ").append(this.key);
        }
        if (this.value != null) {
            sb.append("; value: ").append(this.value);
        }
        return sb.append(']').toString();
    }
    
}
