/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.meta.util;

import static dk.rode.thesis.meta.util.Const.EMPTY;

import java.util.Locale;

/**
  * Various <i>string utilities</i>. <p>
  * 
  * When ever possible, the supplied char sequence arguments are used
  * as is in order to avoid copying into strings. Supplied char sequence
  * arguments are <b>never</b> modified!
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
public class Strings {

    /**
      * Private off-limit constructor.
      */
    private Strings() {
        throw new AssertionError(); // guard against reflection...
    }

    /**
      * Returns true if the char at index {@code index} in the
      * char sequence {@code cs} is escaped, false if not. <p>
      * 
      * The escape char used is backslash (\).
      * 
      * @param  cs     The char sequence; cannot be null.
      * @param  index  The index to check.
      * @return True if escaped, false if not.
      * @throws NullPointerException If {@code cs} is null.
      * @throws IndexOutOfBoundsException If {@code index} is illegal.
      */
    public final static boolean isEscaped(final CharSequence cs, int index) {
        return isEscaped(cs, '\\', index);
    }
    
    /**
      * Returns true if the char at index {@code index} in the
      * char sequence {@code cs} is escaped, false if not. <p>
      * 
      * The escape char used is {@code escape}.
      * 
      * @param  cs     The char sequence; cannot be null.
      * @param  escape The escape char.
      * @param  index  The index to check.
      * @return True if escaped, false if not.
      * @throws NullPointerException If {@code cs} is null.
      * @throws IndexOutOfBoundsException If {@code index} is illegal.
      */
    public final static boolean isEscaped(final CharSequence cs, char escape, int index) {
        if ((index < 0) || (index > cs.length() - 1)) {
            throw new IndexOutOfBoundsException(Integer.toString(index));
        }
        // First char cannot be escaped...
        if (index == 0) {
            return false;
        }
        int i = 0;
        while (--index > 0) {
            char c = cs.charAt(index);
            if (c == escape) {
                i++;
            } else {
                break;
            }
        }
        return ((i > 0) && (i % 2 == 1)); // odd number = escaped
    }

    /**
      * Returns the number of times the char {@code c}
      * is present in the char sequence {@code cs}.
      *
      * @param  cs The char sequence; cannot be null.
      * @param  c  The character to test.
      * @return The number of times {@code c} is present in {@code cs}.
      * @throws NullPointerException If {@code cs} is null.
      */
    public final static int count(final CharSequence cs, char c) {
        int count = 0;
        for (int i = 0; i < cs.length(); i++) {
            if (cs.charAt(i) == c) {
                count++;
            }
        }
        return count;
    }

    /**
      * Finds the first index where {@code c} is present in {@code cs}, 
      * or -1 if {@code c} cannot be found. <p>
      *  
      * @param  cs The char sequence; cannot be null.
      * @param  c  The char to find.
      * @return The first index of {@code c}, or -1 if not found.
      * @throws NullPointerException If {@code cs} is null.
      */
    public final static int indexOf(final CharSequence cs, char c) {
        return indexOf(cs, c, 0);
    }
    
    /**
      * Finds the first index where {@code c} is present in {@code cs}
      * <b>after</b> the index supplied, or -1 if {@code c} cannot be
      * found. <p>
      * 
      * {@code index} is allowed to be larger than {@code cs.length()},
      * in which case this method returns -1. 
      *  
      * @param  cs    The char sequence; cannot be null.
      * @param  c     The char to find.
      * @param  index The index to start at; cannot be negative.
      * @return The first index of {@code c}, or -1 if not found.
      * @throws NullPointerException If {@code cs} is null.
      * @throws IndexOutOfBoundsException If {@code index} is negative.
      */
    public final static int indexOf(final CharSequence cs, char c, int index) {
        for (int i = index; i < cs.length(); i++) {
            if (cs.charAt(i) == c) {
                return i;
            }
        }
        return -1;
    }

    /**
      * Finds the first index where {@code c} is present in {@code cs}, 
      * regardless of case, or -1 if {@code c} cannot be found. <p>
      *  
      * @param  cs The char sequence; cannot be null.
      * @param  c  The char to find.
      * @return The first index of {@code c}, regardless of case, or -1 if 
      *         not found.
      * @throws NullPointerException If {@code cs} is null.
      */
    public final static int indexOfIgnoreCase(final CharSequence cs, char c) {
        return indexOfIgnoreCase(cs, c, 0);
    }
    
    /**
      * Finds the first index where {@code c} is present in {@code cs}
      * <b>after</b> the index supplied, regardless of case, or -1 if 
      * {@code c} cannot be found. <p>
      * 
      * {@code index} is allowed to be larger than {@code cs.length()},
      * in which case this method returns -1. 
      *  
      * @param  cs    The char sequence; cannot be null.
      * @param  c     The char to find.
      * @param  index The index to start at; cannot be negative.
      * @return The first index of {@code c}, regardless of case, or -1 if 
      *         not found.
      * @throws NullPointerException If {@code cs} is null.
      * @throws IndexOutOfBoundsException If {@code index} is negative.
      */
    public final static int indexOfIgnoreCase(final CharSequence cs, char c, int index) {
        return indexOfIgnoreCase(cs, Character.toString(c), index);
    }
    
    /**
      * Finds the first index where {@code token} is present in {@code cs}, 
      * regardless of case, or -1 if {@code token} cannot be found. <p>
      *  
      * The default locale is used to perform the case conversion.
      * 
      * @param  cs    The char sequence; cannot be null.
      * @param  token The char sequence to find; cannot be null.
      * @return The first index of {@code token}, regardless of case, or -1 if 
      *         not found.
      * @throws NullPointerException If either argument is null.
      */
    public final static int indexOfIgnoreCase(final CharSequence cs, final CharSequence token) {
        return indexOfIgnoreCase(cs, token, 0);
    }

    /**
      * Finds the first index where {@code token} is present in {@code cs}, 
      * regardless of case, or -1 if {@code token} cannot be found. <p>
      *  
      * The supplied locale is used to perform the case conversion.
      * 
      * @param  cs     The char sequence; cannot be null.
      * @param  token  The char sequence to find; cannot be null.
      * @param  locale The locale to use; can be null, in which case
      *                the default locale is used.
      * @return The first index of {@code token}, regardless of case, or -1 if 
      *         not found.
      * @throws NullPointerException If {@code cs} or {@code token} are null. 
      */
    public final static int indexOfIgnoreCase(final CharSequence cs, final CharSequence token, final Locale locale) {
        return indexOfIgnoreCase(cs, token, 0, locale);
    }
    
    /**
      * Finds the first index where {@code token} is present in {@code cs}
      * <b>after</b> the index supplied, regardless of case, or -1 if 
      * {@code token} cannot be found. <p>
      * 
      * {@code index} is allowed to be larger than {@code cs.length()},
      * in which case this method returns -1. 
      *  
      * The default locale is used to perform the case conversion.
      * 
      * @param  cs    The char sequence; cannot be null.
      * @param  token The char sequence to find; cannot be null.
      * @param  index The index to start at; cannot be negative.
      * @return The first index of {@code token}, regardless of case, or -1 if 
      *         not found.
      * @throws NullPointerException If {@code cs} or {@code token} are null.
      * @throws IndexOutOfBoundsException If {@code index} is negative.
      */
    public final static int indexOfIgnoreCase(final CharSequence cs, final CharSequence token, int index) {
        return indexOfIgnoreCase(cs, token, index, Locale.getDefault());
    }
    
    /**
      * Finds the first index where {@code token} is present in {@code cs}
      * <b>after</b> the index supplied, regardless of case, or -1 if 
      * {@code token} cannot be found. <p>
      * 
      * {@code index} is allowed to be larger than {@code cs.length()},
      * in which case this method returns -1. 
      *  
      * The supplied locale is used to perform the case conversion.
      * 
      * @param  cs     The char sequence; cannot be null.
      * @param  token  The char sequence to find; cannot be null.
      * @param  index  The index to start at; cannot be negative.
      * @param  locale The locale to use; can be null, in which case
      *                the default locale is used.
      * @return The first index of {@code token}, regardless of case, or -1 if 
      *         not found.
      * @throws NullPointerException If {@code cs} or {@code token} are null.
      * @throws IndexOutOfBoundsException If {@code index} is negative.
      */
    public final static int indexOfIgnoreCase(final CharSequence cs, CharSequence token, int index, Locale locale) {
        if (locale == null) {
            locale = Locale.getDefault();
        }
        return cs.toString().toLowerCase(locale).indexOf(token.toString().toLowerCase(locale), index);
    }
    
    /**
      * Splits the char sequence {@code cs} on the char {@code c}. <p>
      *  
      * Note, that this method does not always yield the same result
      * as {@code string.split()} using the same split value (a single
      * char)! The {@code string.split()} method does not include 
      * "empty parts in the end of the string" in the returned array, 
      * for example:
      * <pre>
      * String foo = "##a#b##c###";
      * 
      * String[] bar = foo.split("#");          // [, , a, b, , c]     = 6 elements
      *          bar = Strings.split(foo, '#'); // [, , a, b, , c, , ] = 8 elements   
      * </pre>
      * 
      * @param  cs The char sequence; cannot be null.
      * @param  c  The split char.
      * @return An array containing the parts; never null, and never empty.
      * @throws NullPointerException If {@code cs} is null.
      */
    public final static CharSequence[] split(final CharSequence cs, char c) {
        CharSequence[] array = new CharSequence[count(cs, c) + 1];
        int count = 0, last = 0, index = 0;
        while ((index = indexOf(cs, c, last)) != -1) {
            array[count++] = cs.subSequence(last, index);
            last = ++index;
        }
        // Remember last part...
        array[count] = cs.subSequence(last, cs.length()); // -1?
        return array;
    }

    /**
      * Prefixes the string value represented by {@code cs} with {@code c}
      * {@code length} times.
      *  
      * @param  cs     The char sequence; cannot be null.
      * @param  c      The char to prefix.
      * @param  length The number of times to prefix {@code c}.
      * @return A new prefixed char sequence; never null.
      * @throws NullPointerException If {@code cs} is null.
      * @throws IllegalArgumentException If {@code length} is less than zero.
      */
    public final static CharSequence prefix(final CharSequence cs, char c, int length) {
        if (length < 0) {
            throw new IllegalArgumentException("Length cannot be negative");
        }
        StringBuilder sb = new StringBuilder(cs.length() + length);
        for (int i = 0; i < length; i++) {
            sb.append(c);
        }
        return sb.append(cs);
    }

    /**
      * Postfixes the string value represented by {@code cs} with {@code c}
      * {@code length} times.
      *  
      * @param  cs     The char sequence; cannot be null.
      * @param  c      The char to postfix.
      * @param  length The number of times to postfix {@code c}.
      * @return A new prefixed char sequence; never null.
      * @throws NullPointerException If {@code cs} is null.
      * @throws IllegalArgumentException If {@code length} is less than zero.
      */
    public final static CharSequence postfix(final CharSequence cs, char c, int length) {
        if (length < 0) {
            throw new IllegalArgumentException("Length cannot be negative");
        }
        StringBuilder sb = new StringBuilder(cs.length() + length);
        sb.append(cs);
        for (int i = 0; i < length; i++) {
            sb.append(c);
        }
        return sb;
    }
    
    /**
      * Lower-cases the first char in the char sequence {@code cs}
      * and returns the modified string using the default locale. <p>
      * 
      * The returned string might not be the same length as {@code cs}
      * as explained under {@code String.toLowerCase(Locale)}.
      *  
      * @param  cs The char sequence; cannot be null.
      * @return A char sequence representing the modification; never null.
      * @throws NullPointerException If {@code cs} is null.
      */
    public final static CharSequence lowerCaseFirst(CharSequence cs) {
        return lowerCaseFirst(cs, Locale.getDefault());
    }
    
    /**
      * Lower-cases the first char in the char sequence {@code cs}
      * and returns the modified string using the supplied locale. <p>
      * 
      * The returned string might not be the same length as {@code cs}
      * as explained under {@code String.toLowerCase(Locale)}.
      *  
      * @param  cs     The char sequence; cannot be null.
      * @param  locale The locale to use; can be null, in which case the
      *                default locale is used.
      * @return A char sequence representing the modification; never null.
      * @throws NullPointerException If {@code cs} is null.
      */
    public final static CharSequence lowerCaseFirst(CharSequence cs, Locale locale) {
        int length = cs.length();
        if (length == 0) {
            return EMPTY;
        }
        if (locale == null) {
            locale = Locale.getDefault();
        }
        // Lower case may contain more than one character...
        String lowerCase = String.valueOf(cs.charAt(0)).toLowerCase(locale);
        if (length == 1) {
            return lowerCase;
        }
        return new StringBuilder(cs).replace(0, 1, lowerCase);
    }

    /**
      * Upper-cases the first char in the char sequence {@code cs}
      * and returns the modified string using the default locale. <p>
      * 
      * The returned string might not be the same length as {@code cs}
      * as explained under {@code String.toUpperCase(Locale)}.
      *  
      * @param  cs The char sequence; cannot be null.
      * @return A char sequence representing the modification; never null.
      * @throws NullPointerException If {@code cs} is null.
      */
    public final static CharSequence upperCaseFirst(CharSequence cs) {
        return upperCaseFirst(cs, Locale.getDefault());
    }
    
    /**
      * Upper-cases the first char in the char sequence {@code cs}
      * and returns the modified string using the supplied locale. <p>
      * 
      * The returned string might not be the same length as {@code cs}
      * as explained under {@code String.toUpperCase(Locale)}.
      *  
      * @param  cs     The char sequence; cannot be null.
      * @param  locale The locale to use; can be null, in which case the
      *                default locale is used.
      * @return A char sequence representing the modification; never null.
      * @throws NullPointerException If {@code cs} is null.
      */
    public final static CharSequence upperCaseFirst(CharSequence cs, Locale locale) {
        int length = cs.length();
        if (length == 0) {
            return EMPTY;
        }
        if (locale == null) {
            locale = Locale.getDefault();
        }
        // Upper case may contain more than one character...
        String upperCase = String.valueOf(cs.charAt(0)).toUpperCase(locale);
        if (length == 1) {
            return upperCase;
        }
        return new StringBuilder(cs).replace(0, 1, upperCase);
    }
    
    /**
      * Returns true if {@code cs} only contains white space
      * characters, false if not.
      * 
      * @param  cs The char sequence to test; cannot be null.
      * @return True if empty, false if not.
      * @throws NullPointerException If {@code cs} is null.
      */
    public final static boolean empty(CharSequence cs) {
        for (int i = 0; i < cs.length(); i++) {
            if (!Character.isWhitespace(cs.charAt(i))) {
                return false;
            }
        }
        return true;
    }
    
}
