/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.observer;

import java.util.Collection;

import dk.rode.thesis.meta.Participant;

/**
  * An <i>observable object</i> is an object that has a 
  * collection of associated <i>observers</i> to be notified in
  * case the object changes its internal state, or <i>aspect</i>. <p>
  *  
  * This interface only supplies the logic for 
  * {@link #addObserver(Object) adding}, {@link #removeObserver(Object) 
  * removing}, and {@link #getObservers() fetching} associated
  * observers, but no notification method to notify the 
  * observers. <p>
  * 
  * Instead, the type of observers is supplied as the type
  * parameter {@code O} and implementing classes must supply
  * the notification mechanism. This allows for greater flexibility,
  * for example asynchronous notification, as well as the use of
  * different methods. <p> 
  * 
  * <b>Implementation notes</b>:<br>
  * This is essentially an application of the 
  * {@linkplain dk.rode.thesis.composite Composite} pattern. <p>
  * 
  * It is not allowed to implement the same generic interface more 
  * than once by a given type because erasure will cause the signatures 
  * to conflict. Hence, an observable type should instead use a 
  * <i>broader</i> type of {@code O} to allow different unique observer 
  * types. Internally, composition could then be used via stand-alone
  * {@code Observable} implementations, such as the 
  * {@link ObserverManager} class, each handling a specific type
  * of observers.  
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    AspectObservable
  * 
  * @param  <O> The type of observers associated with this observable. 
  */
@Participant("Subject")
public interface Observable<O> {

    /**
      * Returns the observers currently associated with this observable 
      * sequence. <p>
      * 
      * No specific order is maintained. <p>
      * 
      * Modifying the returned collection will not affect this observable.
      *  
      * @return The collection of observers; never null, but can be empty.
      */
    public Collection<O> getObservers();
      
    /**
      * Adds the observer supplied as {@code observer} to this observable,
      * if not already. <p>
      * 
      * @param  observer The observer to add; cannot be null.
      * @return True if {@code observer} was added, false if not.
      * @throws NullPointerException If {@code observer} is null.
      * @throws IllegalArgumentException If {@code observer} is not 
      *         a valid observer to add to this observable (optional).
      * @see    #removeObserver(Object)         
      */
    public boolean addObserver(@Participant("ConcreteObserver") O observer);
    
    /**
      * Removes the observer supplied as {@code observer} from this 
      * observable, if already added. <p>
      * 
      * Trying to remove an observer that is not associated with
      * this observable has no effect. 
      * 
      * @param  observer The observer to remove; cannot be null.
      * @return True if {@code observer} was removed, false if not.
      * @throws NullPointerException If {@code observer} is null.
      * @see    #addObserver(Object)         
      */
    public boolean removeObserver(Object observer);

}
