/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.observer;

import static dk.rode.thesis.meta.util.Util.newIdentityHashSet;
import static dk.rode.thesis.meta.util.Throwables.toRuntimeException;

import java.lang.ref.WeakReference;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import dk.rode.thesis.chainofresponsibility.Handler;
import dk.rode.thesis.chainofresponsibility.HandlerChain;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.ParticipantUsage;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.reflect.Executor;
import dk.rode.thesis.meta.reflect.ExecutorProcessor;
import dk.rode.thesis.meta.reflect.Reflection;
import dk.rode.thesis.meta.util.ErrorHandler;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * An <i>observer manager</i> is a stand-alone manager storing 
  * any type of observer that will be notified with methods
  * declared in its class that are annotated with a matching 
  * {@link Executor} annotation. Registered observers subscribe
  * to <i>all or nothing</i>. A <i>real subject</i> implementation
  * can use this manager as a component, delegating the notification
  * and work to it. The subject, if any, must be supplied as one of
  * the arguments to the {@link #notifyObservers(Object...)} method. <p>   
  * 
  * When observers are {@link #addObserver(Object) added}, 
  * the matching methods will be found: methods annotated with
  * the {@code Executor} annotation declaring a 
  * {@link Executor#value() context} to which the 
  * {@link #getOwnerType() owner type} of the manager is 
  * assignable to. The methods represent the notification mechanism
  * for the observer in question and will be executed, using the
  * observer object as the <i>target</i> if not static, when 
  * {@link #notifyObservers(Object...)} is invoked. <p>
  * 
  * The formal parameter types of applicable methods must be 
  * supplied at construction time, and matching arguments
  * supplied to the {@code notifyObserver} method. The formal
  * parameter types ignore primitive types, i.e. {@code Integer.class}
  * and {@code Integer.TYPE} can both represent {@code int},
  * for example. <p>
  * 
  * Any manager must use an error {@link ErrorHandler handler}
  * to determine how errors during notification should be
  * handled, for example if the notification should continue
  * in case of errors or stop immediately. For convenience, the 
  * {@link NotificationPolicy} enumeration defines two such default 
  * error handlers. <p>
  * 
  * The observer manager is not thread-safe! It does not enforce
  * synchronisation, neither when adding or removing observers, nor
  * when observers are notified. It could be potentially dangerous
  * or least time consuming to notify observers from a synchronised 
  * context, holding a lock on the manager itself. If synchronisation
  * is required, the context using the manager must enforce it. <p>
  *  
  * <b>Implementation notes</b>:<br>
  * An observer manager can be seen as a simple version of the
  * <i>Change Manager</i> object described by Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.299]
  * that allows the subject to be supplied as an argument to
  * the {@link #notifyObservers(Object...)} method. <p>
  * 
  * By using annotations and reflection, any method with a matching
  * signature can be used to notify the observer. This includes static 
  * and/or private methods if the security manager allows it. <p>
  * 
  * Methods annotated with the {@code Executor} annotation must 
  * declare the applicable context, i.e. classes allowed to use
  * it. This ensures that different observer registries can 
  * use the same annotation in the same class. <p>
  * 
  * Observers are only stored as <i>weak references</i>, and the 
  * manager will not prevent ordinary garbage collection of the
  * observers. Hence, anonymous non-referenced observers are <b>not</b> 
  * a good idea for this manager as they will be garbage collected
  * without warning. <p>
  *  
  * As always, reflection can cause numerous runtime errors! Hence, 
  * the observer manager utilises error handlers to determine how 
  * errors should be processed in case of notification errors. <p>
  * 
  * An alternative to annotations is to use a regular interface to
  * specify both the observer type and the means (method) to perform the
  * notification. The {@linkplain dk.rode.thesis.chainofresponsibility Chain of Responsibility}
  * pattern utilises such an approach to identify its {@link Handler handlers} 
  * (observers) in the {@link HandlerChain} class. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see  Executor
  */
@ParticipantUsage(value = "Observer", type = Method.class)
@Participant({"Subject", "ConcreteSubject"})
public class ObserverManager implements Observable<Object> {

    /**
      * Default <i>error handlers</i> for observer registries.  
      *
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    public enum NotificationPolicy implements ErrorHandler {
        
        /**
          * Any exception thrown during notification to an observer is 
          * handled, and the notification will thus <b>continue</b> to
          * notify the remaining observers. <p>
          * 
          * Errors cannot be handled.
          */
        ContinueOnError {
            public boolean handle(Throwable throwable) {
                return (throwable instanceof Exception);
            }
        },
        
        /**
          * Any exception/error thrown during notification to an observer will
          * cause the notification to <b>stop</b>, and remaining observers
          * are <b>not</b> notified.
          */
        StopOnError {
            public boolean handle(Throwable throwable) {
                return false;
            }
        };
    }
    
    /**
      * An <i>error handler</i> used to handle errors thrown during
      * notification to a given observer, which will decide if a given 
      * type of error will cause the overall notification to fail. <p>
      * 
      * Default handler is {@link NotificationPolicy#StopOnError}. <p>
      * 
      * Never null.
      */
    private final ErrorHandler errorHandler;
    /**
      * The <i>owner type</i> of this manager. <p>
      * 
      * For an observer to be registered to this manager, this
      * type must be assignable to the class specified in 
      * {@link Executor#value()} for each annotated 
      * method. <p>
      * 
      * Never null.
      */
    private final Class<?> ownerType;
    /**
      * The formal parameter types declared for <b>all</b> methods
      * to be used to notify registered observers. <p>
      * 
      * Primitive types are ignored, i.e. the parameter types may
      * contain both {@code Integer.class} and {@code Integer.TYPE}
      * to represent an {@code int}. <p>
      *  
      * Never null, but can be empty.
      */
    private final Class<?>[] parameterTypes;
    /**
      * The number of registered observers. <p>
      * 
      * A given observer may be included more than once if it is 
      * notified with more than one annotated method.
      */
    private int size;
    /**
      * The observers registered in this manager. <p>
      * 
      * Keys are the methods to perform the notification and the 
      * associated value a list of observers to be notified using
      * that method, stored as weak references. <p> 
      * 
      * Never null, but can be empty.
      */
    private final Map<Method,List<WeakReference<Object>>> observers;

    /**
      * Constructor. <p>
      *
      * The error handling policy used is {@link NotificationPolicy#StopOnError},
      * i.e. the notification will stop in case the notification of
      * an observer fails.
      * 
      * @param  ownerType      The <i>owner type</i> of this manager; 
      *                        cannot be null.
      * @param  parameterTypes The formal parameter types required by all
      *                        annotated methods used to notify observers
      *                        stored in this manager, ignoring primitive 
      *                        types; can be null.
      * @throws NullPointerException If {@code ownerType} is null.
      */
    public ObserverManager(Class<?> ownerType, Class<?>... parameterTypes) {
        this(NotificationPolicy.StopOnError, ownerType, parameterTypes);
    }

    /**
      * Constructor. <p>
      *
      * @param  errorHandler   The error handler determining if the notification
      *                        should continue in case of errors; cannot be null.
      * @param  ownerType      The <i>owner type</i> of this manager; 
      *                        cannot be null.
      * @param  parameterTypes The formal parameter types required by all
      *                        annotated methods used to notify observers
      *                        stored in this manager, ignoring primitive 
      *                        types; can be null.
      * @throws NullPointerException If {@code errorHandler} or {@code ownerType} 
      *         are null.
      */
    public ObserverManager(ErrorHandler errorHandler, Class<?> ownerType, Class<?>... parameterTypes) {
        if ((this.errorHandler = errorHandler) == null) {
            throw new NullPointerException("Error handler cannot be null");
        } else if ((this.ownerType = ownerType) == null) {
            throw new NullPointerException("Type cannot be null");
        }
        this.parameterTypes = parameterTypes == null ? new Class<?>[0] : parameterTypes;
        this.observers = new LinkedHashMap<Method,List<WeakReference<Object>>>(); 
    }

    /**
      * Copy constructor.
      * 
      * @param  manager The manager to copy; cannot be null.
      * @throws NullPointerException If {@code manager} is null.
      */
    public ObserverManager(ObserverManager manager) {
        this.ownerType = manager.ownerType;
        this.errorHandler = manager.errorHandler;
        // Assumed immutable...
        this.parameterTypes = manager.parameterTypes;
        // We can simply reuse the same weak references!
        this.observers = new LinkedHashMap<Method,List<WeakReference<Object>>>(manager.observers);
        this.size = manager.size;
    }
    
    /**
      * Returns the <i>owner type</i> of this manager. <p>
      * 
      * All methods associated with a registered observer 
      * in this manager is annotated with a context 
      * {@link Executor#value() type} this type is assignable to.
      * 
      * @return The owner type of this manager; never null.
      */
    public Class<?> getOwnerType() {
        return this.ownerType;
    }
    
    /**
      * 
      */
    public Collection<Object> getObservers() {
        // Create an identity hash set to store unique observers...
        Set<Object> identities = newIdentityHashSet();
        // A set storing methods to remove because garbage collection has
        // removed all associated observers. Declared as a separate set to 
        // avoid concurrent modification exceptions when traversing the 
        // observer map below...
        Set<Method> remove = new HashSet<Method>();
        for (Entry<Method,List<WeakReference<Object>>> entry : this.observers.entrySet()) {
            Method method = entry.getKey();
            List<WeakReference<Object>> observers = entry.getValue();
            for (int i = observers.size() - 1; i >= 0; i--) {
                Object observer = observers.get(i).get();
                // Remove garbage collected observers...
                if (observer == null) {
                    observers.remove(i);
                    this.size--;
                    LogFactory.getLog(this).warn("Removed garbage collected observer: ", method);
                } else {
                    identities.add(observer);
                }
            }
            if (observers.size() == 0) {
                remove.add(method);
            }
        }
        for (Method method : remove) {
            this.observers.remove(method);
        }
        return identities;
    }

    
    /**
      * Returns the observers and their associated notification
      * methods, regardless if static or not, registered to this 
      * manager. <p>
      * 
      * Modifying the returned map will not affect this manager.
      *  
      * @return A map containing the methods as keys and the associated
      *         observers as the values; never null, but can be empty.
      */
    public Map<Method,List<Object>> getObserverMethods() {
        Map<Method,List<Object>> allObservers = new HashMap<Method,List<Object>>();
        // A set storing methods to remove because garbage collection has
        // removed all associated observers. Declared as a separate set to 
        // avoid concurrent modification exceptions when traversing the 
        // observer map below...
        Set<Method> remove = new HashSet<Method>();
        for (Entry<Method,List<WeakReference<Object>>> entry : this.observers.entrySet()) {
            Method method = entry.getKey();
            List<WeakReference<Object>> localObservers = entry.getValue();
            List<Object> observers = new ArrayList<Object>(localObservers.size());
            for (int i = localObservers.size() - 1; i >= 0; i--) {
                Object observer = localObservers.get(i).get();
                // Remove garbage collected observer...
                if (observer == null) {
                    localObservers.remove(i);
                    this.size--;
                    LogFactory.getLog(this).warn("Removed garbage collected observer: ", method);
                } else {
                    observers.add(observer);
                }
            }
            if (localObservers.size() == 0) {
                remove.add(method);
            } else {
                allObservers.put(method, observers);
            }
        }
        for (Method method : remove) {
            this.observers.remove(method);
        }
        return allObservers;
    }
    
    /**
      * Adds the observer supplied as {@code observer} to this
      * manager, associated with the methods annotated in the class 
      * of {@code object} with a context {@link Executor#value() type} 
      * the {@link #getOwnerType() owner type} of this manager is
      * assignable to, if any. <p>
      *   
      * {@code observer} is used as the target for all non-static 
      * methods registered by this invocation. <p>
      * 
      * Adding {@code observer} multiple times has no effect. 
      * 
      * @param  observer The observer; cannot be null.
      * @return True if {@code observer} was added as an observer,
      *         false if not. 
      * @throws NullPointerException If {@code observer} is null.
      * @see    #removeObserver(Object)
      */
    public boolean addObserver(@Participant("ConcreteObserver") Object observer) {
        return this.addObserver(observer.getClass(), observer);
    }
    
    /**
      * Adds the observer supplied as {@code observer} to this
      * manager, associated with the methods annotated in the class 
      * supplied as  {@code clazz} with a context {@link Executor#value() type} 
      * the {@link #getOwnerType() owner type} of this manager is
      * assignable to, if any. <p>
      *   
      * {@code observer} is used as the target for all non-static 
      * methods registered by this invocation. <p> 
      * 
      * Adding {@code observer} multiple times has no effect.
      *  
      * @param  <T>      The type of {@code clazz}.
      * @param  clazz    The class literal supplying the type to fetch the
      *                  methods from; cannot be null.
      * @param  observer The observer; cannot be null.
      * @return True if {@code observer} was added as an observer,
      *         false if not. 
      * @throws NullPointerException If either argument is null.
      * @see    #removeObserver(Class, Object)
      */
    public <T> boolean addObserver(Class<? extends T> clazz, @Participant("ConcreteObserver") T observer) {
        if (observer == null) {
            throw new NullPointerException("Object cannot be null");
        }
        // Fetch methods annotated with proper executors...
        List<Method> methods = 
            ExecutorProcessor.getMethodExecutors(clazz, this.ownerType, this.parameterTypes);
        if (methods.size() == 0) {
            return false;
        }
        boolean added = false;
        for (Method method : methods) {
            // "method" is now confirmed a method to use to notify the
            // observer supplied as "observer"!
            List<WeakReference<Object>> observers = this.observers.get(method);
            if (observers == null) {
                this.observers.put(method, observers = new ArrayList<WeakReference<Object>>(3));
            }
            // Compare on identity! We traverse the list backwards so we
            // can remove garbage collected observers without fear of
            // excessive copying...
            boolean add = true;
            // Linear, but short lists because methods include type
            // information, hence good hash code function...
            for (int i = observers.size() - 1; i >= 0; i--) {
                Object o = observers.get(i).get();
                // Remove garbage collected observers...
                if (o == null) {
                    observers.remove(i);
                    this.size--;
                    LogFactory.getLog(this).warn("Removed garbage collected observer: ", method);
                    continue;
                    
                } else if (o == observer) {
                    add = false;
                    break;
                }
            }
            if (add) {
                observers.add(new WeakReference<Object>(observer));
                this.size++;
                LogFactory.getLog(this).println("Added: ", method, " -> ", observer);
            }
            added |= add;
        }
        return added;
    }

    /**
      * Removes the observer supplied as {@code observer} from this
      * manager. <p>
      * 
      * If {@code object} is not associated with this manager, this
      * method does nothing.
      * 
      * @param  observer The observer to remove; cannot be null.
      * @return True if {@code observer} was removed, false if not.
      * @throws NullPointerException If {@code observer} is null.
      * @see    #addObserver(Object)
      */
    public boolean removeObserver(Object observer) {
        return this.removeObserver(observer.getClass(), observer);
    }

    /**
      * Removes the observer supplied as {@code observer} from this
      * manager. <p>
      * 
      * If {@code object} is not associated with this manager, this
      * method does nothing.
      * 
      * @param  <T>    The type of {@code clazz}.
      * @param  clazz  The class literal supplying the type; cannot be null.
      * @param  observer The observer to remove; cannot be null.
      * @return True if {@code observer} was removed, false if not.
      * @throws NullPointerException If either argument is null.
      * @see    #addObserver(Class, Object)
      */
    public <T> boolean removeObserver(Class<? extends T> clazz, T observer) {
        boolean removed = false;
        for (Method m : ExecutorProcessor.getMethodExecutors(clazz, this.ownerType, this.parameterTypes)) {
            removed = this.remove(m, observer) || removed;
        }
        return removed;
    }
    
    /**
      * Removes the specific notification method supplied as {@code method}
      * from the observer supplied as {@code observer}. <p>
      * 
      * If {@code method} is the last method associated with {@code observer},
      * {@code observer} will also be removed. 
      * 
      * @param  method   The specific method to remove; cannot be null.
      * @param  observer The observer; cannot be null.
      * @return True if {@code method} was removed, false if not.
      * @throws NullPointerException If either argument is null.
      */
    private boolean remove(Method method, Object observer) {
        if (method == null) {
            throw new NullPointerException("Method cannot be null");
        } else if (observer == null) {
            throw new NullPointerException("Object cannot be null");
        }
        List<WeakReference<Object>> observers = this.observers.get(method);
        if (observers == null) {
            return false;
        }
        // Invariant: static methods also have an observer associated, but
        //            the observer is unused when executed (still used for
        //            garbage collection tests)...
        
        boolean removed = false;
        // Linear, but short lists because methods include type
        // information, hence good hash code function...
        for (int i = observers.size() - 1; i >= 0; i--) {
            Object o = observers.get(i).get();
            // Remove garbage collected observers...
            if (o == null) {
                observers.remove(i);
                this.size--;
                LogFactory.getLog(this).warn("Removed garbage collected observer: ", method);
                continue;
                
            } else if (o == observer) {
                observers.remove(i);
                removed = true;
                LogFactory.getLog(this).println("Removed: ", method, " -> ", observer);
                break;
            }
        }
        if (observers.size() == 0) {
            this.observers.remove(method);
        }
        return removed;
    }

    /**
      * Notifies all observers stored in this manager, using 
      * the arguments for each method supplied as {@code arguments}. <p>
      * 
      * In case the notification to a given observer (per method) fails, 
      * the error handler supplied at construction time will try and
      * handle the error. If the error is manageable, the notification
      * will continue; otherwise an exception is thrown. <p>
      * 
      * No specific order of notification among observers is enforced.
      * 
      * @param  arguments The arguments to supply to each observer.
      * @return True if the notification caused errors that could be
      *         handled by this manager, false if no errors. In case of
      *         errors that cannot be handled, a runtime exception is thrown.
      * @throws IllegalArgumentException If {@code arguments} does not
      *         match the formal observer parameter types specified at
      *         construction time.             
      * @throws RuntimeException If the notification failed in a manner
      *         that could not be handled by this manager.  
      */
    public boolean notifyObservers(Object... arguments) {
        // Check arguments...
        if (!Reflection.isInstance(true, this.parameterTypes, arguments)) {
            throw new IllegalArgumentException("Illegal arguments: " + arguments);
        }
        boolean errors = false;
        Log out = LogFactory.getLog(this);
        // A set storing methods to remove because garbage collection has
        // removed all associated observers. Declared as a separate set to 
        // avoid concurrent modification exceptions when traversing the 
        // observer map below...
        Set<Method> remove = new HashSet<Method>();
        // As methods are only registered when fetched from a given
        // object, we know that the object can be used as the target
        // when the methods are invoked (if not static)...
        for (Entry<Method,List<WeakReference<Object>>> entry : this.observers.entrySet()) {
            Method method = entry.getKey();
            List<WeakReference<Object>> observers = entry.getValue();
            for (int i = observers.size() - 1; i >= 0; i--) {
                Object observer = observers.get(i).get();
                // Remove garbage collected observers...
                if (observer == null) {
                    observers.remove(i);
                    this.size--;
                    out.warn("Removed garbage collected observer: ", method);
                    continue;
                   
                // Call method...
                } else {
                    try {
                        out.println("Invoking: ", method, " -> ", observer);
                        Reflection.invoke(observer, method, arguments);
                        
                    } catch (Exception e) {
                        // Use error handler strategy to determine what to do...
                        if (this.errorHandler.handle(e)) {
                            out.warn("Handled observer error").warn(e);
                            errors = true;
                            
                        } else {
                            out.error("Unhandled observer error").error(e);
                            throw toRuntimeException(e);
                        }
                    }
                }
            }
            if (observers.size() == 0) {
                remove.add(method);
            }
        }
        
        // Can safely remove methods now...
        for (Method method : remove) {
            this.observers.remove(method);
        }
        return errors;
    }
    
    /**
      * Return the number of observers associated with this manager.
      * 
      * @return The number of observers.
      */
    public int size() {
        return this.size;
    }

    /**
      * Returns the string representation of this manager.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        StringBuilder sb = (StringBuilder)ObjectPolicy.ID_OPEN.toString(this);
        return sb.append("; owner-type: ").append(this.ownerType.getSimpleName())
                 .append("; observers: ").append(this.size).append(']').toString(); 
    }
    
}
