/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.observer;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.meta.model.Sequence.State;
import dk.rode.thesis.meta.reflect.Executor;
import dk.rode.thesis.meta.reflect.Reflection;
import dk.rode.thesis.meta.util.Util;
import dk.rode.thesis.state.ReversiblePrimeSequence;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>probe sequence observer</i> prints the internal attributes that
  * has changed for a given {@link Sequence sequence} since the last time 
  * the observer was notified about that sequence, based on equality
  * using {@code equals(Object)} (and {@code hashCode()}). <p>
  * 
  * <b>Implementation notes</b>:<br>
  * A probe observer uses the {@link Executor} annotation to specify 
  * the observer methods, and can thus be used with the {@link ObserverManager}
  * type or any {@link AnnotatedObserversSequence} implementation. It cannot,
  * however, be used an observer for sequences expecting a  
  * {@link SequenceObserver} type.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("ConcreteObserver")
public class ProbeSequenceObserver { // no standard interface, annotations are used instead 

    /**
      * The map of currently probed sequences. <p>
      * 
      * Never null.
      */
    private Map<Sequence<?>,Sequence<?>> sequences;
    
    /**
      * No-arg constructor.
      */
    public ProbeSequenceObserver() {
        this.sequences = new HashMap<Sequence<?>,Sequence<?>>();
    }

    /**
      * Probes the sequence supplied as {@code sequence}, and logs
      * all attributes that has changed value (based on {@code equals(Object)})
      * since the last time this method was invoked with the same
      * sequence (based on {@code equals(Object)}) as the first argument. <p>
      *  
      * Any context may use this probe method.
      * 
      * @param  sequence  The sequence to probe; cannot be null.
      * @param  state     The current state of {@code sequence}; can be null.
      * @throws Exception If the probing fails.
      */
    @Executor // can be executed by all
    void internalProbe(Sequence<?> sequence, State state) throws Exception {
        if (sequence == null) {
            throw new NullPointerException("Sequence cannot be null");
        }
        Log out = LogFactory.getLog(this);

        // Last copy of "sequence"...
        Sequence<?> old = this.sequences.get(sequence);
        if (old != null) {
            boolean first = true;
            List<Field> fields = Reflection.getFields(sequence.getClass());
            for (Field field : fields) {
                Object newValue = Reflection.invoke(sequence, field);
                if (newValue instanceof Sequence) {
                    internalProbe((Sequence<?>)newValue, state);
                } else {
                    Object oldValue = Reflection.invoke(old, field);
                    if (!Util.equals(oldValue, newValue)) {
                        if (first) {
                            first = false;
                            out.println("Internal probe: ", ObjectPolicy.ID.toString(sequence));
                        }
                        out.println("- Attribute changed: ", field.getName(), ": ", oldValue, " -> ", newValue);
                    }
                }
            }
        }
        // Update "old" by making a copy to save the current values...
        this.sequences.put(sequence, sequence.copy());
    }

    /**
      * Probe method that will only be invoked by an observable 
      * {@link ReversiblePrimeSequence} instance (decorated by
      * the {@link AnnotatedObserversSequenceDecorator} type).
      * 
      * @param  sequence The sequence; never null.
      * @param  state    The sequence state; never null.
      */
    @Executor(ReversiblePrimeSequence.class) // can only be executed by a reversible prime sequence, or sub-class of it
    static void externalProbe(Sequence<?> sequence, State state) {
        LogFactory.getLog(ProbeSequenceObserver.class).println("External probe: ", ObjectPolicy.ID.toString(sequence));
    }
    
    /**
      * Returns the string representation of this observer. 
      *
      * @return The string representation; never null.
      */
    public String toString() {
        return ObjectPolicy.IDENTITY.asString(this);
    }
    
}
