/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.prototype;

import static dk.rode.thesis.meta.util.Throwables.toRuntimeException;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.util.LinkedHashSet;
import java.util.Set;

import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.reflect.Reflection;
import dk.rode.thesis.meta.reflect.TypeLiteral;
import dk.rode.thesis.meta.reflect.proxy.ProxyFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>prototype factory</i> creates various <i>dynamic proxies</i> 
  * that allows any object to become {@link Copyable copyable}, 
  * if not already, if it supplies a 
  * {@link Reflection#getCopyConstructor(Class) copy} constructor. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * This factory utilises the {@linkplain dk.rode.thesis.proxy Proxy} pattern
  * to create prototypical objects. The factory defers the actual proxy 
  * creation to the {@link ProxyFactory} class, which allows creation of 
  * proxies based on any type. The {@code ProxyFactory} could have been 
  * inherited, but composition is chosen instead to keep the
  * interface of this class clean and focused on copyable objects. <p>
  * 
  * Notice that the created proxies cannot be {@link StrictCopyable} 
  * because the original type is not copyable. <p>
  * 
  * The prototype factory does <b>not</b> correspond to a 
  * <b>Prototype Manager</b> as described by Gamma et al. [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.121] because
  * prototypes are created on the fly. No fixed pool of predetermined
  * prototype objects is used by this factory.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    ProxyFactory
  */
public class PrototypeFactory {

    /**
      * The singleton proxy factory used to create the proxy objects
      * representing {@link Copyable copyable} objects. <p>
      * 
      * Never null.
      */
    private final static ProxyFactory factory = new ProxyFactory();
    /**
      * The {@link Copyable#copy()} method defined in the
      * {@code Copyable} interface. <p>
      * 
      * Never null.
      */
    private final Method copy;
    
    /**
      * No-arg constructor.
      */
    public PrototypeFactory() {
        try {
            this.copy = Copyable.class.getMethod("copy");
            
        } catch (Throwable t) {
            LogFactory.getLog(PrototypeFactory.class).error("Cannot fetch copy() method").error(t);
            throw toRuntimeException(t);
        }
    }

    /**
      * Returns a prototypical object representing {@code object} 
      * that can be copied by invoking {@link Copyable#copy()} on it
      * <b>if and only if</b> the class of {@code object} supplies 
      * a <i>copy constructor</i> and/or directly implements {@code Copyable}. 
      * If a prototypical object cannot be created, null is returned. <p>
      * 
      * To invoke the {@code copy()} method, {@code object} must 
      * first be cast into the {@code Copyable} type. It will not
      * be possible to cast to a generic type without generating 
      * compiler warnings because of type erasure. <p>
      *  
      * A copy constructor of {@code type} is defined as a constructor 
      * accepting a single parameter of the exact same type as {@code type}. <p>
      * 
      * If {@code object} implements {@code Copyable}, {@code object}
      * is simply returned. <p>
      * 
      * The returned prototypical object can be cast into any interface 
      * implemented by {@code object}, though <b>not</b> the actual type
      * (class) itself, represented by {@code type}. Trying to cast the 
      * returned object into the actual type (class) of {@code object} 
      * will thus cause a class cast exception to be thrown!
      * 
      * @param  <T>    The type of {@code object}.
      * @param  type   A class literal representing the type of {@code object};
      *                cannot be null. 
      * @param  object The object to make prototypical; cannot be null.
      * @return The prototypical object, or null.
      * @throws NullPointerException Either argument is null.
      * @see    #getPrototypicalObject(TypeLiteral, Object)
      */
    public <T> T getPrototypicalObject(Class<T> type, final T object) {
        // Does object directly implement copyable? If so, no need to
        // create a proxy...
        if (object instanceof Copyable) {
            return object;
        }
        
        // Check copy constructor...
        final Constructor<T> constructor = Reflection.getCopyConstructor(type);
        if (constructor == null) {
            return null;
        }
        Set<Class<?>> interfaces = new LinkedHashSet<Class<?>>();
        // Tag with copyable as the *first* interface...
        interfaces.add(Copyable.class);
        // ...and implement all interfaces of object...
        Reflection.getInterfaces(type, interfaces);
        
        InvocationHandler handler = new InvocationHandler() {
            public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
                // Intercept call to "copy()"...
                if (PrototypeFactory.this.copy.equals(method)) {
                    // Use copy constructor...
                    return constructor.newInstance(object); // could also have recursively created copyable proxies here...
                }
                // Forward all other calls to "object"...
                return Reflection.invoke(object, method, args);
            }
        };
        return factory.getProxy(object, handler, interfaces.toArray(new Class<?>[interfaces.size()]));
    }

    /**
      * Returns a prototypical object representing {@code object} 
      * that can be copied by invoking {@link Copyable#copy()} on it
      * <b>if and only if</b> the class of {@code object} supplies 
      * a <i>copy constructor</i> and/or directly implements {@code Copyable}. 
      * If a prototypical object cannot be created, null is returned. <p>
      * 
      * A copy constructor of {@code type} is defined as a constructor 
      * accepting a single parameter of the exact same type as {@code type}. <p>
      * 
      * If {@code object} implements {@code Copyable}, {@code object}
      * is simply returned. <p>
      * 
      * The returned instance is cast to a copyable instance matching
      * the exact type parameter {@code E}, even if it contains nested
      * type parameters. <p> 
      *  
      * The returned prototypical object can be cast into any interface 
      * implemented by {@code object}, though <b>not</b> the actual type
      * (class) itself, represented by {@code type}. Trying to cast the 
      * returned object into the actual type (class) of {@code object} 
      * will thus cause a class cast exception to be thrown!
      * 
      * @param  <T>    The type of {@code object}.
      * @param  type   A type literal representing the type of {@code object};
      *                cannot be null. 
      * @param  object The object to make prototypical; cannot be null.
      * @return The prototypical object, or null.
      * @throws NullPointerException Either argument is null.
      * @see    #getPrototypicalObject(Class, Object)
      */
    @SuppressWarnings("unchecked") // Copyable<T> known to be implemented...
    public <T> Copyable<T> getPrototypicalObject(TypeLiteral<T> type, T object) {
        return (Copyable<T>)this.getPrototypicalObject(type.asClass(), object);
    }
    
    /**
      * Returns the string representation of this factory.
      * 
      * @return The string representation; never null.
      */
    public String toString() {
        return ObjectPolicy.ID.asString(this);
    }
    
}
