/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.prototype;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.ArraySequence;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>cloneable symbol sequence</i> can be cloned by using the 
  * {@link #copy()} method, or by using Java's built-in clone
  * mechanism by calling {@link #clone()}. <p>
  * 
  * The following symbols are {@linkplain #getAllowedSymbols() allowed}:
  * <pre>
  *   ! " #  % & / ( ) = ? < > @  $  { [ ] } + | ~ : * 
  * </pre>
  * 
  * A symbol sequence thus acts as a prototype for similar sequences 
  * containing the same symbols, in order. <p> 
  *  
  * A symbol sequence is {@link Sequence#bounded() bounded}, 
  * {@link Sequence#consistent() consistent}, and 
  * {@link Sequence#unique() unique}. 
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("ConcretePrototype")
public class SymbolSequence extends ArraySequence<CharSequence> {

    /**
      * The various symbols allowed in a symbol sequence. <p>
      * 
      * The allowed symbols are: 
      * 
      * <pre>
      *   ! " #  % & / ( ) = ? < > @  $  { [ ] } + | ~ : * 
      * </pre>
      * 
      * Trying to construct a symbol sequence with other symbols
      * is an error.
      */
    private final static CharSequence[] SYMBOLS = {  
        "!", "\"", "#", "", "%", "&", "/", "(", ")", "=", "?", "<", ">", 
        "@", "",  "$", "", "{", "[", "]", "}", "+", "|", "~", ":", "*" 
    };
    
    /**
      * Validates the symbols supplied in {@code symbols}. <p>
      * 
      * If each char in {@code symbols} is contained in
      * {@link #SYMBOLS}, this method returns silently with
      * a {@code CharSequence} array corresponding to the
      * symbols. Otherwise an exception is thrown.
      * 
      * @param  symbols The symbols to validate; cannot be null.
      * @return A corresponding {@code CharSequence} array; never null.
      * @throws NullPointerException If {@code symbols} is null.
      * @throws IllegalArgumentException If {@code symbols} contains
      *         an illegal symbol.
      */
    private final static CharSequence[] validate(char[] symbols) {
        CharSequence[] array = new CharSequence[symbols.length]; 
        loop:
        for (int i = 0; i < symbols.length; i++) {
            char c = symbols[i];
            for (CharSequence cs : SYMBOLS) {
                if (cs.charAt(0) == c) {
                    array[i] = String.valueOf(c);
                    continue loop;
                }
            }
            throw new IllegalArgumentException("Illegal symbol: " + c);
        }
        return array;
    }
    
    /**
      * No-arg constructor. <p>
      * 
      * All allowed {@linkplain #getAllowedSymbols() allowed} symbols are used.
      */
    public SymbolSequence() {
        super(SYMBOLS);
    }

    /**
      * Constructor. <p>
      * 
      * The symbols supplied as {@code symbols} are used in this
      * symbol sequence.
      * 
      * @param  symbols The symbols from the {@linkplain #getAllowedSymbols() allowed} 
      *                 symbols to use in this sequence; cannot be null or be empty.
      * @throws NullPointerException If {@code symbols} is null.                 
      * @throws IllegalArgumentException If {@code symbols} is empty, or
      *         if it contains symbols not found in the {@linkplain #getAllowedSymbols() allowed}
      *         symbols.                 
      */
    public SymbolSequence(char... symbols) {
        super(validate(symbols));
    }
    
    /**
      * Returns the various symbols allowed in any symbol sequence. <p>
      * 
      * The allowed symbols are: 
      * 
      * <pre>
      *   ! " #  % & / ( ) = ? < > @  $  { [ ] } + | ~ : * 
      * </pre>
      * 
      * Trying to construct a symbol sequence with other symbols
      * is an error. <p>
      * 
      * Modifying the returned array will only have local effect.
      * 
      * @return The allowed symbols; never null.
      */
    public final static CharSequence[] getAllowedSymbols() {
        CharSequence[] symbols = new CharSequence[SYMBOLS.length];
        System.arraycopy(symbols, 0, SYMBOLS, 0, SYMBOLS.length);
        return symbols;
    }
    
    /**
      * Copy constructor. <p>
      * 
      * This sequence starts at the current index of {@code sequence}.
      * 
      * @param  sequence The sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public SymbolSequence(SymbolSequence sequence) {
        super(sequence);
    }
    
    /**
      * 
      */
    public SymbolSequence copy() {
        return new SymbolSequence(this);
    }

    /**
      * Clones this symbol sequence and returns the clone (copy). 
      *
      * @return A clone (copy) of this sequence; never null.
      */
    public SymbolSequence clone() {
        try {
            // Extend Super pattern...
            SymbolSequence sequence = (SymbolSequence)super.clone();
            sequence.index = this.index;
            sequence.state = this.state;
            return sequence;
            
        // This should never happen since this class is cloneable!
        } catch (CloneNotSupportedException cnse) {
            throw new InternalError(cnse.getMessage());
        }
    }
    
}
