/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */
package dk.rode.thesis.singleton;

import java.lang.reflect.Type;

import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.reflect.TypeLiteral;

/**
  * A <i>loadable singleton registry</i> is a registry that offers
  * the same functionality as {@link StatefullSingletonRegistry}, but
  * also allows loading of the actual singleton classes if not already
  * loaded using the {@link #getInstance(String)} method. <p>
  * 
  * All loaded singleton classes will be loaded using the class loader
  * that loaded this class, or by a specific class loader supplied at
  * construction time. Only singleton classes matching the bound 
  * supplied as {@code T} can be loaded, regardless if {@code T} is 
  * generic or not. <p> 
  *  
  * The class is abstract in order to preserve the generic information, and 
  * must be created in the following manner:
  * <pre>
  *   LoadableSingletonRegistry&lt;Foo&lt;? extends CharSequence>> registry = 
  *     new {@link #LoadableSingletonRegistry() LoadableSingletonRegistry}&lt;Foo&lt;? extends CharSequence>>(){}; // default: {@link StatelessSingletonRegistry} is used
  * 
  *   ClassLoader classLoader = ..;
  *   {@link SingletonRegistry}&lt;Foo&lt;? extends CharSequence>> aggregate = ..;
  *   LoadableSingletonRegistry&lt;Foo&lt;? extends CharSequence>> registry = 
  *     new {@link #LoadableSingletonRegistry(ClassLoader, SingletonRegistry) LoadableSingletonRegistry}&lt;Foo&lt;? extends CharSequence>>(classLoader, aggregate){};
  * </pre>
  * In the above example, only singleton types conforming to {@code Foo<? extends CharSequence>} 
  * can be loaded. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * This class <b>cannot</b> load singleton types implemented using enumerations, because
  * generic information from super-classes is used and all enumerations extend  
  * {@link java.lang.Enum java.lang.Enum}. They can still be acquired using the 
  * {@link #getInstance(Class)} method, though.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <T> The super type of all singleton types handled by
  *             this registry. Use {@code Object} if any type
  *             of singleton type can be handled.
  */
public abstract class LoadableSingletonRegistry<T> extends StatefullSingletonRegistry<T> {

    /**
      * The class loader used by this registry. <p>
      *  
      * Default class loader is the class loader that loaded this class. <p>
      *  
      * Never null. 
      */
    private final ClassLoader classLoader;
    
    /**
      * Constructor. <p>
      * 
      * An instance of the {@link StatelessSingletonRegistry} type
      * will be used to acquire the singleton instances. <p>
      * 
      * The class loader that loaded this class will be used by this
      * registry.
      */
    protected LoadableSingletonRegistry() {
        this(LoadableSingletonRegistry.class.getClassLoader());
    }
    
    /**
      * Constructor. <p>
      * 
      * An instance of the {@link StatelessSingletonRegistry} type
      * will be used to acquire the singleton instances.
      * 
      * @param  classLoader The class loader to use; cannot be null.
      * @throws NullPointerException If {@code classLoader} is null.
      */
    protected LoadableSingletonRegistry(ClassLoader classLoader) {
        this(classLoader, new StatelessSingletonRegistry<T>());
    }
    
    /**
      * Constructor. <p>
      * 
      * @param  classLoader The class loader to use; cannot be null.
      * @param  registry    The registry that will be used to acquire the
      *                     singleton instances; cannot be null.              
      * @throws NullPointerException If either argument is null. 
      */
    protected LoadableSingletonRegistry(ClassLoader classLoader, SingletonRegistry<T> registry) {
        super(registry);
        if ((this.classLoader = classLoader) == null) {
            throw new NullPointerException("Class loader cannot be null"); 
        }
    }
    
    /**
      * Loads the singleton type represented by the class name supplied
      * as {@code className} using the class loader that loaded this
      * registry, and then invokes {@link #getInstance(Class)} using
      * the loaded singleton type to return the singleton instance. <p>
      * 
      * A cast is required to obtain the exact singleton type.
      * 
      * @param  className The class name of the singleton type; cannot 
      *                   be null.
      * @return The singleton instance; never null.
      * @throws NullPointerException If {@code className} is null.
      * @throws ClassNotFoundException If {@code className} is illegal.
      * @throws ClassCastException If {@code className} represents a type
      *         that is not an instance of {@code T}, including if 
      *         {@code T} is generic.
      * @throws SingletonException If the class represented by {@code className}
      *         is not a singleton type handled by this registry, or if the 
      *         type cannot be created when first requested.
      */
    public T getInstance(String className) throws ClassNotFoundException {
        // Class literal as a Type representing the class just loaded...
        Type clazz = this.classLoader.loadClass(className);
        // Type literal representing "clazz"...
        TypeLiteral<?> type = TypeLiteral.create(clazz); // Type
        // Class literal representing T...
        TypeLiteral<?> handledTypes = TypeLiteral.create(this.getClass()); // Class<?>
        // If this succeeds without a class cast exception, the types conform...
        handledTypes.asType(type);
        // Yes!
        LogFactory.getLog(this).println("Loaded singleton class: ", className);
        // Now safe to cast, as the (generic) type of clazz is a conforming sub-class
        // of T that can be handled by this registry. The aggregate registry may still 
        // throw an exception, though, if the singleton type cannot be acquired...
        @SuppressWarnings("unchecked") Class<T> c = (Class<T>)clazz;
        return this.getInstance(c);
    }
    
}

