/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.singleton;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.Log;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.ArraySequence;
import dk.rode.thesis.meta.reflect.Caller;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>Simpsons family sequence</i> is a singleton sequence representing
  * the primary family members in the one and only <i>The Simpsons</i> family. <p>
  * 
  * Only a single instance of this class is created, on demand only, and 
  * is available via the {@link #getFamily()} method. However, <b>sub-classing
  * is allowed</b> using the supplied {@link #SimpsonsFamilySequence(String...) 
  * sub-class} and/or {@link #SimpsonsFamilySequence(SimpsonsFamilySequence) copy} 
  * constructors. If any other context than a sub-class tries to invoke these 
  * constructors, a {@link SingletonError} is thrown! Hence, the singleton 
  * invariant is still enforced even though sub-classing is allowed, and 
  * sub-classes may or may not be singletons. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * The {@code SimpsonsFamilySequence} utilises a static inner class to store the
  * actual singleton instance. Unlike the {@link DanishAlphabetSequence} class, 
  * the instance is therefore <b>not</b> created at the class loading time, but 
  * the first time {@link #getFamily()} is invoked, i.e. lazily initialised. 
  * Therefore, for example, accessing the static field {@link #SIZE} will 
  * <b>not</b> cause the singleton instance to be created. The initialisation 
  * utilises standard static initialisation (of the inner class), and the 
  * {@code getFamily()} method need not be synchronised. This corresponds
  * to the <i>initialise-on-demand holder class</i> idiom [<a href="http://www.rode.dk/thesis/Bibliography.htm#Bloch01" target="thesis-bibliography">Bloch01</a>, p.194].<p>
  * 
  * The check to see if it is a sub-class that is invoking a 
  * sub-class constructor is done using a {@link Caller} instance. 
  * It utilises Java's reflection capabilities. Acquiring a sub-class 
  * instance before the {@code getFamily()} method has been invoked 
  * will <b>not</b> cause the singleton instance to be created. Sub-class
  * constructors are declared protected, but even if they had been public, the 
  * sub-class check would still work. But, by declaring them protected, we only 
  * have to worry about objects in the same package illegally invoking the 
  * constructors. However, using reflection, even private members can be 
  * accessed if the security manager allows it, but the sub-class check 
  * will <b>still</b> work if reflection is used. There is furthermore a check
  * in the private constructor that ensures that {@code SimpsonsFamilySequence}
  * instances cannot be created using reflection (the same goes for the
  * other singleton types declared in this package!). The {@code SimpsonsFamilySequence} class
  * is thus a <a href="http://www.rode.dk/thesis/ComparativeEvaluation.htm#section-7.1.2.1" target="thesis"><i>Guarded Type</i></a>. <p> 
  * 
  * Though not implemented, this class could implement {@link java.io.Serializable
  * java.io.Serializable}. If so, the serialization should override the serialization 
  * methods {@code writeReplace()} and {@code readResolve()}, the latter always returning
  * the singleton instance. It does not implement {@link java.lang.Cloneable java.lang.Cloneable}, 
  * either (to throw an exception), because otherwise sub-classes cannot become cloneable.
  * It is ok for a sub-class to invoke {@code super.clone()} because that will correspond
  * to {@code Object.clone()}, which will return an instance of the sub-class, not
  * this sequence. This is illustrated in the {@link MutatedSimpsonsFamilySequence} sub-class. <p>
  *   
  * Any singleton instance in Java is only a singleton per <i>class loader</i>
  * if created based on and utilising intrinsic class features. <p>
  * 
  * Though the standard {@link ArraySequence} {@code hashCode} and
  * {@code equals(Object)} works for singleton sub-classes of it, 
  * faster versions of the two are implemented here to take advantage of
  * the singleton behaviour.
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    DanishAlphabetSequence
  * @see    NorwegianAlphabetSequence
  * @see    SimpsonsAndBouvierFamilySequence
  * @see    SmileySequence
  * @see    MutatedSimpsonsFamilySequence
  */
@Participant("Singleton")
@Singleton("getFamily")
public class SimpsonsFamilySequence extends ArraySequence<String> {

    /**
      * The primary members in <i>The Simpsons</i> family. <p>
      */
    final static String[] members = {
        "Bartholomew Jojo Simpson", "Lisa Marie Simpson", "Maggie Simpson",
        "Homer Jay Simpson", "Marge Simpson", 
        "Abbie Simpson", "Herbert Powell",
        "Abraham Jay Simpson (Grampa)", "Mona Jay Simpson",
        "Cyrus Simpson", "Orville Jay Simpson", "Bill Simpson"
    };

    /**
      * The number of primary members in <i>The Simpsons</i> family. <p>
      * 
      * <b>Implementation notes</b>:<br>
      * Accessing this static field will <b>before</b> {@link #getFamily()}
      * will <b>not</b> cause the singleton instance to be created!
      * 
      * @see DanishAlphabetSequence#SIZE
      * @see NorwegianAlphabetSequence#SIZE
      * @see SimpsonsAndBouvierFamilySequence#SIZE
      * @see SmileySequence#SIZE
      */
    public final static int SIZE = members.length; 
        
    /**
      * An <i>instance container</i> class that when loaded will create 
      * the {@link SimpsonsFamilySequence} instance to use and return
      * from the {@link SimpsonsFamilySequence#getFamily()} method. <p>
      *
      * The initialisation of the sequence is thread-safe as it
      * utilises standard static initialisation. <p>
      * 
      * <b>Implementation notes</b>:<br>
      * This corresponds to the <i>initialise-on-demand holder class</i> idiom
      * described in [<a href="http://www.rode.dk/thesis/Bibliography.htm#Bloch01" target="thesis-bibliography">Bloch01</a>, p.194].
      * 
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    private final static class Instance {
        
        /**
          * The {@link SimpsonsFamilySequence singleton} instance. <p>
          * 
          * Not created until the {@link SimpsonsFamilySequence#getFamily()}
          * method is invoked.
          */
        private final static SimpsonsFamilySequence instance = 
            new SimpsonsFamilySequence();

        /**
          * Private off-limit constructor. 
          */
        private Instance() {
            throw new AssertionError(); // guard against reflection...
        }
    }
    
    /**
      * Returns the {@link SimpsonsFamilySequence} singleton 
      * instance to use. <p>
      * 
      * The singleton instance is not created before the first time
      * this method is invoked! <p>
      * 
      * <b>Implementation notes</b>:<br>
      * Because this method is static it cannot be overridden. However,
      * it cannot be declared final because we support singleton inheritance: 
      * a class does not allow two static methods with the same signature.
      * So, if declared final, sub-classes could not define their own
      * {@code getFamily} method to <i>hide</i> this method to return the 
      * proper instance.  
      *  
      * @return <i>The Simpsons</i> family sequence; never null.
      * @see    Singleton
      */
    public static SimpsonsFamilySequence getFamily() {
        // Forces static initialisation of the *inner* class, if
        // not already, in a thread-safe manner!
        return Instance.instance;
    }
    
    /**
      * Private no-arg constructor for singleton creation.
      * 
      * @throws SingletonError If another class tries to invoke
      *         this constructor (via reflection).
      */
    private SimpsonsFamilySequence() {
        super(members);
        Log out = LogFactory.getLog(this);
        // Guard against reflection...
        Caller caller = new Caller().getNonClassCaller(); // never null
        // Can only be invoked during static initialisation of the 
        // nested Instance class...
        if ((caller.callerClass != Instance.class) || (!caller.isStatic())) {
            out.error("Illegal access to singleton constructor: ", caller);
            throw new SingletonError(caller);
        }
        out.println("Created singleton: ", this);
    }
    
    /**
      * Sub-class constructor. <p>
      * 
      * Only a sub-class of this sequence can invoke this constructor!
      * Any other context invoking it will cause a 
      * {@link SingletonError} to be thrown.
      * 
      * @param  members The family members of the family; cannot be null, 
      *                 or empty.
      * @throws NullPointerException If {@code members} is null.
      * @throws IllegalArgumentException If {@code members} is empty.
      * @throws SingletonError If a non sub-class context
      *         tries to invoke this constructor.
      */
    protected SimpsonsFamilySequence(String... members) {
        super(members);
        Log out = LogFactory.getLog(this);
        // If the caller of this constructor is a sub-class constructor, 
        // allow the call. If not, we throw an exception. However, we 
        // *cannot* guarantee that the sub-class constructor invokes 
        // this constructor as "super" or uses "new" to create this
        // class explicitly! Hence, some trust must be placed in the 
        // sub-class... 
        Caller caller = new Caller();
        Caller subClass = caller.getSubClassCaller();
        if ((subClass == null) || (!subClass.isConstructor())) {
            out.error("Illegal access to sub-class constructor: ", caller);
            throw new SingletonError(caller);
        }
        out.println("Sub-class allowed to call constructor: ", this.toString(SequencePolicy.ID));
    }
    
    /**
      * Sub-class copy constructor constructor. <p>
      * 
      * Only a sub-class of this sequence can invoke this constructor!
      * Any other context invoking it will cause a 
      * {@link SingletonError} to be thrown.
      *
      * @param  sequence The sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      * @throws SingletonError If a non sub-class context
      *         tries to invoke this constructor.
      */
    protected SimpsonsFamilySequence(SimpsonsFamilySequence sequence) {
        super(sequence);
        Log out = LogFactory.getLog(this);
        // If the caller of this constructor is a sub-class constructor, 
        // allow the call. If not, we throw an exception. However, we 
        // *cannot* guarantee that the sub-class constructor invokes 
        // this constructor as "super" or uses "new" to create this
        // class explicitly! Hence, some trust must be placed in the 
        // sub-class... 
        Caller caller = new Caller();
        Caller subClass = caller.getSubClassCaller();
        if ((subClass == null) || (!subClass.isConstructor())) {
            out.error("Illegal access to sub-class copy constructor: ", caller);
            throw new SingletonError(caller);
        }
        out.println("Sub-class allowed to call copy constructor: ", this.toString(SequencePolicy.ID));
    }
    
    /**
      * Returns this instance.
      * 
      * @return This instance; never null.
      */
    public SimpsonsFamilySequence copy() {
        // Disallow copy because more than one Simpsons family is more
        // than the world can manage... :)
        return this;
    }
    
    /**
      * 
      */
    public int hashCode() {
        // Faster hash code since we know we are a singleton, though
        // the standard AlphabetSequence hash code works!
        return this.getClass().hashCode();
    }
    
    /**
      *  
      */
    public boolean equals(Object object) {
        // Faster equals since we know we are a singleton, though
        // the standard AlphabetSequence equals works!
        return (this == object);
    }
    
}
