/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.singleton;

import java.lang.annotation.Documented;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

import dk.rode.thesis.meta.reflect.Executor;
import dk.rode.thesis.observer.ObserverManager;

/**
  * A <i>singleton</i> annotation identifies a given type 
  * already at compile-time as a singleton by identifying the 
  * <b>static method returning an instance of the 
  * annotated type</b>. The method can either accept no arguments, 
  * or a fixed number of {@code java.lang.String} arguments. <p>
  * 
  * The method is known as the <i>singleton method</i>. By default, 
  * a method name of {@code getInstance} is assumed, corresponding
  * to the static method {@code Foo getInstance()} in the {@code Foo}
  * class. <p>
  * 
  * Any type declared as a {@code Singleton} can be utilised
  * in a stateless singleton {@link StatelessSingletonRegistry registry}. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * An annotation is preferred over an interface, because different 
  * singleton objects offer no common functionality except for conceptual 
  * behaviour in form of only a single instance available. The singleton 
  * method must in any case be static, and interfaces cannot declare 
  * static methods. An annotation also allows for different method
  * names for different types. <p>
  * 
  * Furthermore, neither an annotation or an interface can guarantee 
  * that the implementing class does not disclose an accessible constructor.   
  * However, by using annotations, Sun's <i>Annotation Processing Tool</i> 
  * (APT) could be primed with a handler to enforce that all found 
  * annotated types do not expose accessible constructors, and check
  * the existence of the specified singleton method, already at compile
  * time. <p> 
  * 
  * A consequence of not using an interface is that singleton registries
  * cannot impose direct bounds on the type parameters used, for example
  * {@code <T extends Singleton>} for a given {@code Singleton} 
  * interface. However, the {@link SingletonRegistry} still allows for
  * such parameterisation by specifying an appropriate type parameter. <p> 
  * 
  * A general parameterised singleton method - corresponding to a parameterised 
  * {@linkplain dk.rode.thesis.factorymethod Factory Method} - is not 
  * practically feasible using this annotation without involving a lot
  * of {@code String} parsing; that is, using an argument to determine the 
  * actual singleton instance to fetch of any type. Hence, we only allow 
  * no arguments or a fixed number of {@code String} arguments here. However, the 
  * idea to use reflection to access the singleton method could easily be extended 
  * to allow for more complex protocols as Java's reflection mechanism can mimic all 
  * normal method (and constructor) invocation scenarios. It could also be extended 
  * so <i>any</i> type could be a singleton type, but that would require a registry 
  * keeping taps  on created instances and would <b>only</b> work if <b>all</b> 
  * access to acquire/create instances were via the registry. A scenario were this 
  * could work could be an {@linkplain dk.rode.thesis.abstractfactory Abstract Factory} 
  * that internally uses such a registry. <p>
  * 
  * The annotation is <b>not</b> inherited, because there is no way a given
  * class can tell if a sub-class will be implemented as a singleton as well.
  * Even so, the sub-class could decide to use a different singleton method. <p>
  * 
  * An alternative to letting the annotation specify the method name could
  * be to use an annotation as a <i>marker</i> on the method to use. A 
  * visitor of some sorts would then be required to traverse the methods
  * declared in a given type to find the (first) one with the marker 
  * annotation that fits the profile of a static method returning a proper
  * singleton type. This roughly corresponds to the idea behind the
  * {@link Executor} annotation, which is used in the 
  * {@linkplain dk.rode.thesis.observer Observer} pattern (see the 
  * {@link ObserverManager} class). Here, however, annotation of the
  * singleton type itself is the logical design choice.
  *     
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Documented
@Retention(RetentionPolicy.RUNTIME)
@Target(ElementType.TYPE)
public @interface Singleton {

    /**
      * The name of the <i>singleton method</i> to access the singleton
      * instance. <p>
      * 
      * The method name must represent a <b>static method returning an instance 
      * of the annotated type</b>. The name must <b>not</b> include parentheses. 
      * The method can either be a no-argument method, or a method accepting
      * a fixed number of {@link String java.lang.String} arguments. <p> 
      * 
      * Argument values are fixed and specified by {@link #arguments()}.  
      * 
      * @return The method name, excluding parenthesis; default {@code getInstance}
      *         (accepting no arguments).
      */
    String value() default "getInstance";

    /**
      * Returns the fixed {@link String java.lang.String} arguments to supply each time the
      * singleton method identified by {@link #value()} is invoked, if any. <p>
      * 
      * The singleton method used must formally declare the same number of
      * parameters of type {@code java.lang.String}.
      * 
      * @return The fixed string arguments; default empty array (no-argument method).
      */
    String[] arguments() default {};
    
}
