/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.state;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.meta.model.PrimeSequence;
import dk.rode.thesis.meta.model.ReversibleSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.strategy.SequencePolicy;

/**
  * A <i>reversible prime sequence</i> returns with each call to 
  * {@link #next() next()} the next prime number in the sequence of 
  * prime numbers smaller than {@code maximum}, where {@code maximum} 
  * is supplied at construction time, <b>or</b> a previously calculated
  * prime number with each call to {@code next()} <b>after</b> 
  * {@link #reverse()} has been invoked. Hence, returned values are in 
  * range of {@code [2, maximum]}. <p>
  * 
  * A reversible prime sequence is {@link Sequence#bounded() bounded},
  * {@link Sequence#consistent() consistent}, and {@link #unique() unique}. <p>
  * 
  * <b>Implementation notes</b>:<br> 
  * A very simple version of the <b>Sieve of Eratosthenes</b> algorithm is used 
  * in this implementation. The amount of memory used by an a reversible prime 
  * sequence is linear with {@code maximum}. The general idea behind the 
  * implementation of the algorithm is identical with the standard 
  * {@link PrimeSequence} implementation, but the internal sequence design 
  * differs considerable between the two except for required attributes. Hence, 
  * it really makes no sense to inherit {@code PrimeSequence} here. Java has 
  * built-in support for generating primes in form of 
  * {@link java.math.BigInteger java.math.BigInteger}, but that is
  * beside the point. <p>
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("Context")
public class ReversiblePrimeSequence extends AbstractStateableSequence<Integer> implements ReversibleSequence<Integer> {

    /**
      * A <i>prime state</i> represents a <b>stateless</b> 
      * <i>functional</i> {@link FunctionalState state} for all 
      * reversible prime sequences, while the actual internal
      * state and attributes are stored in a given reversible 
      * prime sequence instance. <p>
      * 
      * State transitions are handled both by prime states, but also
      * by reversible prime sequences requesting a given action,
      * i.e. in {@link ReversiblePrimeSequence#reverse()}. <p>
      * 
      * <b>Implementation notes</b>:<br>
      * Each state may be considered implemented according to the 
      * {@linkplain dk.rode.thesis.flyweight Flyweight} pattern in that they
      * are singletons that only operate on extrinsic state. 
      *
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    private static enum PrimeState implements FunctionalState<Integer> { 

        /**
          * Functional state <i>initialising</i> a reversible prime sequence. 
          * It is the first state set for any reversible prime sequence. <p>
          * 
          * <b>State transitions</b>: <ul><p>
          * 
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#START}: 
          *      none, but initialises the sieve and validates the maximum number 
          *      of primes to investigate. Once this state is discarded, it will 
          *      never be used again for the same sequence. <p>
          *      
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#NORMAL}: 
          *      {@link #Calculate} if all primes have not been calculated, 
          *      otherwise {@link #Next}. <p>
          *  
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESET}: 
          *      {@link #ResetLow}. <p> 
          * 
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESTART}: 
          *      illegal state exception. <p>
          *  
          * </ul>
          * 
          * @see #action(Sequence.State, ReversiblePrimeSequence)
          */
        @Participant("ConcreteState")
        Start {
            protected void action(State state, ReversiblePrimeSequence sequence) {
                switch (state) {
                    case START: {
                        if (sequence.maximum < 2) {
                            throw new IllegalStateException("Maximum number must be > 2");
                        }
                        sequence.calculate = (sequence.maximum > 3);
                        // Identify multiples of 2 as not prime, because 2 is the
                        // initial value...
                        for (int j = 0; j * 2 <= sequence.maximum; j++) {
                            sequence.sieve[j * 2] = true; // not prime
                        }
                        sequence.sieve[1] = true; // 0 and 1 unused, but are not prime!
                        
                        // To keep this state initially...
                        ResetLow.action(state, sequence);
                        break;
                    }
                    case RESET: {    
                        sequence.setFunctionalState(ResetLow, state);
                        break;
                    }
                    case NORMAL: {
                        sequence.setFunctionalState(sequence.calculate ? Calculate : Next, state);
                        break;
                    }
                    default: {
                        throw new IllegalStateException(state.name());
                    }
                }  
            }
        },
        
        /**
          * Functional state <i>resetting</i> a reversible prime sequence to start
          * over either because of initialisation, or in case of restart or 
          * {@link ReversiblePrimeSequence#reset() reset} explicitly if
          * prime numbers are delivered in ascending order ("forward"). <p>
          *
          * <b>State transitions</b>: <ul><p>
          * 
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#START}: 
          *      initialises the reversible prime sequence the very first
          *      time (called from {@link #Start}). Henceforth an illegal
          *      state exception is thrown if used again. <p>
          *      
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#NORMAL}: 
          *      {@link #Calculate} if all primes have not been calculated, 
          *      otherwise {@link #Next}. <p>
          * 
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESET}: 
          *      None, but resets the sequence (low). <p>
          *  
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESTART}: 
          *      None, but resets the sequence (low). <p>
          *  
          * </ul>
          * 
          * Can also be set by {@link ReversiblePrimeSequence#reverse()}. <p> 
          * 
          * @see #ResetHigh
          * @see #action(Sequence.State, ReversiblePrimeSequence)
          */
        @Participant("ConcreteState")
        ResetLow {
            protected void action(State state, ReversiblePrimeSequence sequence) {
                switch (state) {
                    case NORMAL: {
                        sequence.setFunctionalState(sequence.calculate ? Calculate : Next, state);
                        break;
                    }
                    case START: {
                        // Initially index is zero...
                        if (sequence.index != 0) {
                            throw new IllegalStateException(state.name());
                        }
                        // Fall-through!
                    }
                    case RESTART: {
                        // Adapt maximum to the last known prime...
                        if (sequence.index != 0) {
                            sequence.maximum = sequence.prime;
                        }
                        // Fall-through!
                    }
                    case RESET: { 
                        sequence.state = state;
                        sequence.index = 3;
                        sequence.prime = 2;
                    }
                }
            }
        },
        
        /**
          * Functional state <i>resetting</i> a reversible prime sequence if 
          * prime numbers are delivered in descending order ("reverse"). If so, 
          * all prime numbers have been calculated. <p>
          * 
          * <b>State transitions</b>: <ul><p>
          * 
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#START}: 
          *      illegal state exception. <p>
          *      
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#NORMAL}: 
          *      {@link #Previous}. <p>
          * 
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESET}: 
          *      None, but resets the sequence (high). <p>
          *  
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESTART}: 
          *      None, but resets the sequence (high). <p>
          *  
          * </ul>
          * 
          * Can also be set by {@link ReversiblePrimeSequence#reverse()}. <p> 
          * 
          * @see #ResetLow
          * @see #action(Sequence.State, ReversiblePrimeSequence)
          */
        @Participant("ConcreteState")
        ResetHigh {
            protected void action(State state, ReversiblePrimeSequence sequence) {
                switch (state) {
                    case NORMAL: {
                        sequence.setFunctionalState(Previous, state);
                        break;
                    }
                    case RESET: 
                    case RESTART: {
                        sequence.state = state;
                        // Since this functional state is stateless, we have
                        // to traverse the sieve to find the highest prime
                        // number. We know that all primes have been calculated,
                        // so we traverse simply until we find a false value.
                        // This can be quite expensive! But once it is done, we
                        // adjust the maximum accordingly for the next reset...
                        sequence.index = sequence.maximum;
                        while (sequence.sieve[sequence.index]) {
                            sequence.index--;
                        }
                        sequence.prime = sequence.index--; 
                        sequence.maximum = sequence.prime;
                        break;
                    }
                    default: {
                        throw new IllegalStateException(state.name());
                    }
                }
            }
        },
        
        /**
          * Functional state <i>calculating</i> prime numbers for a reversible 
          * prime sequence, and delivering the next prime number (just calculated). <p>
          * 
          * <b>State transitions</b>: <ul><p>
          * 
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#START}: 
          *      illegal state exception. <p>
          *      
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#NORMAL}: 
          *      {@link #Next} if all primes have been calculated. <p> 
          * 
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESET}: 
          *      {@link #ResetLow}. <p>
          *  
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESTART}: 
          *      illegal state exception. <p>
          *  
          * </ul>
          * 
          * @see #action(Sequence.State, ReversiblePrimeSequence)
          */
        @Participant("ConcreteState")
        Calculate {
            protected void action(State state, ReversiblePrimeSequence sequence) {
                switch (state) {
                    // Calculate...
                    case NORMAL: {
                        for (; sequence.calculate = sequence.index * sequence.index <= sequence.maximum; sequence.index++) {
                            if (!sequence.sieve[sequence.index]) {
                                // Identify multiples of "index" as not prime...
                                for (int j = sequence.index; j * sequence.index <= sequence.maximum; j++) {
                                    sequence.sieve[j * sequence.index] = true; // not prime
                                }
                                sequence.prime = sequence.index++;
                                return;
                            } 
                        }
                        // Done calculating...
                        sequence.setFunctionalState(Next, state);
                        break;
                    }
                    // Reset...
                    case RESET: {
                        sequence.setFunctionalState(ResetLow, state);
                        break;
                    }
                    default: {
                        throw new IllegalStateException(state.name());
                    }
                }
            }
        },
        
        /**
          * Functional state <i>delivering the next</i> prime 
          * number for a reversible prime sequence ("forward"), where
          * all prime numbers have been calculated. <p>
          * 
          * <b>State transitions</b>: <ul><p>
          * 
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#START}: 
          *      illegal state exception. <p>
          *      
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#NORMAL}: 
          *      {@link #ResetLow} if restarting. <p> 
          * 
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESET}: 
          *      {@link #ResetLow}. <p>
          *  
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESTART}: 
          *      {@link #ResetLow}. <p>
          *  
          * </ul>
          * 
          * Can also be set by {@link ReversiblePrimeSequence#reverse()}. <p>
          * 
          * @see #Previous
          * @see #action(Sequence.State, ReversiblePrimeSequence)
          */
        @Participant("ConcreteState")
        Next {
            protected void action(State state, ReversiblePrimeSequence sequence) {
                switch (state) {
                    case NORMAL: {
                        // Advance until next prime is found, i.e. a false value...
                        while (sequence.index <= sequence.maximum) {
                            // Not prime...
                            if (sequence.sieve[sequence.index]) {
                                sequence.index++;
                            // Prime...
                            } else {
                                sequence.prime = sequence.index++;
                                return;
                            }
                        }
                        // Fall-through!
                    }
                    // Restart...
                    case RESTART: {
                        state = State.RESTART;
                        // Fall-through!
                    }
                    // Reset...
                    case RESET: {
                        sequence.setFunctionalState(ResetLow, state);
                        break;
                    }
                    default: {
                        throw new IllegalStateException(state.name());
                    }
                }
            }
        },
        
        /**
          * Functional state <i>delivering the previous</i> prime 
          * number for a reversible prime sequence ("reverse"), where
          * all prime numbers have been calculated. <p>
          * 
          * <b>State transitions</b>: <ul><p>
          * 
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#START}: 
          *      illegal state exception. <p>
          *      
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#NORMAL}: 
          *      {@link #ResetHigh} if restarting. <p> 
          * 
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESET}: 
          *      {@link #ResetHigh}. <p>
          *  
          * <li> {@link dk.rode.thesis.meta.model.Sequence.State#RESTART}: 
          *      {@link #ResetHigh}. <p>
          *  
          * </ul>
          * 
          * Can also be set by {@link ReversiblePrimeSequence#reverse()}. <p> 
          * 
          * @see #Next
          * @see #action(Sequence.State, ReversiblePrimeSequence)
          */
        @Participant("ConcreteState")
        Previous {
            protected void action(State state, ReversiblePrimeSequence sequence) {
                switch (state) {
                    case NORMAL: {
                        // Advance until next prime is found, i.e. a false value...
                        while (sequence.index >= 2) { 
                            // Not prime...
                            if (sequence.sieve[sequence.index]) {
                                sequence.index--;
                            // Prime...
                            } else {
                                sequence.prime = sequence.index--;
                                return;
                            }
                        }
                        // Fall-through!
                    }
                    // Restart...
                    case RESTART: {
                        state = State.RESTART;
                        // Fall-through!
                    }
                    // Reset...
                    case RESET: {
                        sequence.setFunctionalState(ResetHigh, state);
                        break;
                    }
                    default: {
                        throw new IllegalStateException(state.name());
                    }
                }
            }
        };

        /**
          * 
          */
        public final Integer action(State internalState, StateableSequence<Integer> sequence) {
            ReversiblePrimeSequence rps = (ReversiblePrimeSequence)sequence;
            if (internalState != null) {
                this.action(internalState, rps);
            }
            return rps.prime;
        }
        
        /**
          * Handles the <i>internal</i> state request supplied as {@code internalState} 
          * for the reversible prime sequence supplied as {@code sequence}. <p>
          * 
          * @param  internalState The internal state; never null, i.e. current 
          *                       state already handled at this point.
          * @param  sequence      The reversible prime sequence in question; 
          *                       cannot be null.
          * @throws NullPointerException If {@code sequence} is null.              
          */
        protected abstract void action(State internalState, ReversiblePrimeSequence sequence);
        
    }
    
    /**
      * Maximum number to consider. <p>
      * 
      * Will be adjusted to the last prime number on first restart.
      */
    private int maximum;
    /**
      * The next index to start processing from on calls
      * to {@link #next() next()}. After each call to next has
      * completed, the value is {@link #prime} + 1 (forward) or
      * {@code prime} - 1 (reverse). <p>
      * 
      * Range: [2, {@link #maximum}].
      */
    private int index;
    /**
      * The current prime number. <p>
      *   
      * Range: [2, {@link #maximum}].
      */
    private int prime;
    /**
      * The actual sieve. <p>
      * 
      * If {@link #calculate} is true, an index smaller than 
      * {@link #index} having a false value represents a prime 
      * number (normally a true value represent
      * a prime number, but by flipping it, we do not have to 
      * initialise the array with true). <p>
      * 
      * If {@code calculate} is false, any index having a false
      * value represents a prime number. <p>
      * 
      * The length is {@link #maximum} + 1. Index zero and one 
      * are unused.
      */
    private boolean[] sieve;
    /**
      * True if all primes have been calculated, false if not. <p>
      * 
      * @see PrimeState#Calculate
      */
    private boolean calculate;
    
    /**
      * Constructor.
      * 
      * @param  maximum The maximum number to consider. Must be larger than 1.
      * @throws IllegalStateException If {@code maximum} is smaller than 2.
      */
    public ReversiblePrimeSequence(int maximum) {
        super();
        this.sieve = new boolean[(this.maximum = maximum) + 1];
        // Initialise...
        this.setFunctionalState(PrimeState.Start, State.START);
    }

    /**
      * Copy constructor.
      *
      * @param  sequence The reversible prime sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public ReversiblePrimeSequence(ReversiblePrimeSequence sequence) {
        super(sequence);
        this.calculate = sequence.calculate;
        this.maximum = sequence.maximum;
        this.prime = sequence.prime;
        this.index = sequence.index;
        this.sieve = new boolean[sequence.sieve.length];
        System.arraycopy(sequence.sieve, 2, this.sieve, 2, sequence.sieve.length - 2);
    }
    
    /**
      * 
      */
    public ReversiblePrimeSequence copy() {
        return new ReversiblePrimeSequence(this);
    }
    
    /**
      * The maximum prime considered by this sequence. <p>
      * 
      * The returned value is not guaranteed to be equal to the maximum
      * value supplied at construction time.
      *  
      * @return The maximum prime considered.
      */
    public int maximum() {
        return this.maximum;
    }
    
    /**
      * Returns true. 
      *
      * @return True.
      */
    public boolean bounded() {
        return true;
    }

    /**
      * Returns true. 
      *
      * @return True.
      */
    public boolean consistent() {
        return true;
    }

    /**
      * Returns true.
      * 
      * @return True.
      */
    public boolean unique() {
        return true;
    }
    
    /**
      * @throws IllegalStateException {@inheritDoc}
      */
    public Integer reverse() {
        FunctionalState<Integer> fs = this.getFunctionalState();
        if (fs == PrimeState.Next) {
            // Have to advance the sequence "in reverse", because the 
            // index to examine next will be higher than the prime. 
            // This will find the current prime again, but at the same
            // time update the index correctly to be lower than the
            // current prime. Hence, the current value is not changed...
            this.setFunctionalState(PrimeState.Previous, State.NORMAL);
            
        } else if (fs == PrimeState.Previous) {
            // Have to advance the sequence "forward", because the 
            // index to examine next will be lower than the prime.
            // This will find the current prime again, but at the same
            // time update the index correctly to be higher than the
            // current prime. Hence, the current value is not changed...
            this.setFunctionalState(PrimeState.Next, State.NORMAL);

        } else if (fs == PrimeState.ResetLow) {
            if (this.calculate) {
                throw new IllegalStateException(fs.toString());
            }
            this.setFunctionalState(PrimeState.ResetHigh, this.state());
            
        } else if (fs == PrimeState.ResetHigh) {
            this.setFunctionalState(PrimeState.ResetLow, this.state());
            
        }  else {
            throw new IllegalStateException(fs.toString());
        }
        LogFactory.getLog(this).println("Reversed: ", this.toString(SequencePolicy.IDENTITY));
        return this.prime;
    }

    /**
      * 
      */
    public boolean reversible() {
        FunctionalState<Integer> fs = this.getFunctionalState();
        if ((fs == PrimeState.Next)     || 
            (fs == PrimeState.Previous) || 
            (fs == PrimeState.ResetHigh)) {
            return true;
        } else if (fs == PrimeState.ResetLow) {
            return !this.calculate;
        }
        return false;
    }
    
}


