/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.state;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;

/**
  * A <i>stateable sequence</i> is implemented internally using
  * <i>functional</i> {@link FunctionalState state} objects to 
  * determine the current functional state.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by this stateable sequence. 
  */
@Participant("Context")
public interface StateableSequence<E> extends Sequence<E> {
    
    /**
      * Returns the <i>functional</i> {@link FunctionalState state} currently
      * set for this stateable sequence if this sequence allows disclosure
      * of such states. <p>
      * 
      * A functional state is not the same as an <i>internal</i> sequence
      * {@link Sequence#state() state}, but functional states will update the
      * internal state. <p>
      * 
      * @return The current functional state; can be null if the this sequence
      *         does not allow functional states to be disclosed.
      * @see    #setFunctionalState(FunctionalState, Sequence.State)
      */
    public FunctionalState<E> getFunctionalState();
    
    /**
      * Updates the <i>functional</i> {@link FunctionalState state} for this 
      * stateable sequence to {@code functionalState}, as well as the 
      * <i>internal</i> {@link dk.rode.thesis.meta.model.Sequence.State 
      * state} to {@code state}, if supplied. <p>
      * 
      * A functional state is not the same as an <i>internal</i> sequence
      * state, but functional states will update the internal state. After 
      * the functional sequence state has been set, its 
      * {@link FunctionalState#action(Sequence.State, StateableSequence)}
      * method is invoked using the internal state supplied as argument.
      *  
      * @param  functionalState The functional state; cannot be null.
      * @param  internalState   The internal state; can be null.
      * @return The previous functional state; never null.
      * @throws NullPointerException If {@code functionalState} is null.
      * @throws IllegalStateException If {@code functionalState} cannot be set, or
      *         if it does not accept a {@code internalState} request. 
      * @see    #getFunctionalState()
      */
    public FunctionalState<E> setFunctionalState(FunctionalState<E> functionalState, State internalState);

}
