/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.strategy;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.Sequence;

/**
  * Default policies for formatting {@link Sequence} 
  * objects into textual representations in form of 
  * char sequences. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * Though defined as an enumeration, different applicable policies 
  * may still be applied to format sequences, i.e. any policy having 
  * the type {@code Stringable&ltSequence<?>>}. <p>
  * 
  * See implementation notes for the {@link ObjectPolicy} enumeration.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    ObjectPolicy
  */
public enum SequencePolicy implements StringablePolicy<Sequence<?>> {
    
    /**
      * Generates a char sequence representation of a sequence by 
      * invoking {@code toString()} on it.
      */
    @Participant("ConcreteStrategy")
    DEFAULT {
        @SuppressWarnings("unchecked") // See comment below...
        public CharSequence toString(Sequence<?> sequence) {
            // Have to use the raw type because we cannot capture
            // the same wild-card in both method invocations below...!
            Sequence s = sequence;
            return s.getStringablePolicy(null).toString(s);
        }
    },
    
    /**
      * Generates a char sequence representation of a sequence having
      * the following format:
      * <pre>
      *   Sequence
      * </pre>
      * 
      * Where {@code Sequence} is the simple class name of the class
      * of the sequence. <p>
      * 
      * This policy is considered {@link StringablePolicy.Type#CLOSED}.
      */
    @Participant("ConcreteStrategy")
    NAME {
        public CharSequence toString(Sequence<?> sequence) {
            return super.toString(sequence);
        }
        public Type type() {
            return Type.CLOSED;
        }
    },

    /**
      * Generates a char sequence representation of a sequence having
      * the following format:
      * <pre>
      *   ..
      * </pre>
      * 
      * Where {@code ..} is the current value of the sequence. <p>
      * 
      * This policy is considered {@link StringablePolicy.Type#CLOSED}.
      */
    @Participant("ConcreteStrategy")
    VALUE {
        public CharSequence toString(Sequence<?> sequence) {
            return sequence.current().toString();
        }
        public Type type() {
            return Type.CLOSED;
        }
    },

    /**
      * Generates a char sequence representation of a sequence having
      * the following format:
      * <pre>
      *   ..
      * </pre>
      * 
      * Where {@code ..} is the current value of the sequence as
      * an upper-cased string. <p>
      * 
      * This policy is considered {@link StringablePolicy.Type#CLOSED}.
      */
    @Participant("ConcreteStrategy")
    VALUE_UPPERCASE {
        public CharSequence toString(Sequence<?> sequence) {
            return sequence.current().toString().toUpperCase();
        }
        public Type type() {
            return Type.CLOSED;
        }
    },
    
    /**
      * Generates a char sequence representation of a sequence having
      * the following format:
      * <pre>
      *   Sequence[..]
      * </pre>
      * 
      * Where {@code Sequence} is the simple class name of the class
      * of the sequence, and {@code ..} is the current sequence
      * value.
      */
    @Participant("ConcreteStrategy")
    SHORT {
        public CharSequence toString(Sequence<?> sequence) {
            StringBuilder sb = (StringBuilder)super.toString(sequence);
            return sb.append('[')
                     .append(sequence.current())
                     .append(']');
        }
    },

    /**
      * Generates a char sequence representation of a sequence having
      * the following format:
      * <pre>
      *   Sequence[..]
      * </pre>
      * 
      * Where {@code Sequence} is the simple class name of the class
      * of the sequence, and {@code ..} is the <b>identity</b> 
      * hash-code of the sequence.
      */
    @Participant("ConcreteStrategy")
    ID {
        public CharSequence toString(Sequence<?> sequence) {
            StringBuilder sb = (StringBuilder)super.toString(sequence);
            return sb.append('[')
                     .append(System.identityHashCode(sequence))
                     .append(']');
        }
    },
    
    /**
      * Generates a char sequence representation of a sequence having
      * the following format:
      * <pre>
      *   Sequence[hash-code: ..[; identity-hash-code: ..]; current: ..]
      * </pre>
      * 
      * Where {@code Sequence} is the simple class name of the class
      * of the sequence, and {@code ..} are the hash-code and
      * current value of the sequence, respectively.
      * 
      * @see ObjectPolicy#HASHCODE
      */
    @Participant("ConcreteStrategy")
    IDENTITY {
        public CharSequence toString(Sequence<?> sequence) {
            StringBuilder sb = (StringBuilder)super.toString(sequence);
            return sb.append('[')
                     .append(ObjectPolicy.HASHCODE.toString(sequence))
                     .append("; current: ")
                     .append(sequence.current())
                     .append(']');
        }
    },
    
    /**
      * Generates a char sequence representation of a sequence having
      * the following format:
      * <pre>
      *   Sequence[hash-code: ..[; identity-hash-code: ..]; current: ..; bounded: ..; consistent: ..; unique: ..; state: ..]
      * </pre>
      * 
      * Where {@code Sequence} is the simple class name of the class
      * of the sequence.
      * 
      * @see #VERBOSE_OPEN
      * @see ObjectPolicy#HASHCODE
      */
    @Participant("ConcreteStrategy")
    VERBOSE {
        public CharSequence toString(Sequence<?> sequence) {
            return ((StringBuilder)VERBOSE_OPEN.toString(sequence)).append(']');
        }
    },
    
    /**
      * Generates a char sequence (StringBuilder) representation of a sequence 
      * having the following format:
      * <pre>
      *   Sequence[hash-code: ..[; identity-hash-code: ..]; current: ..; bounded: ..; consistent: ..; unique: ..; state: .. 
      * </pre>
      * 
      * Where {@code Sequence} is the simple class name of the class
      * of the sequence. <p>
      * 
      * The difference between this policy and {@link #VERBOSE}
      * is that the representation formatted by this policy does 
      * not end with a closing bracket (]), i.e. is 
      * {@link StringablePolicy.Type#OPEN}. 
      *
      * @see ObjectPolicy#HASHCODE
      */
    @Participant("ConcreteStrategy")
    VERBOSE_OPEN {
        public StringBuilder toString(Sequence<?> sequence) {
            StringBuilder sb = (StringBuilder)super.toString(sequence);
            return sb.append('[')
                     .append(ObjectPolicy.HASHCODE.toString(sequence))
                     .append("; current: ")
                     .append(sequence.current())
                     .append("; bounded: ")
                     .append(sequence.bounded())
                     .append("; consistent: ")
                     .append(sequence.consistent())
                     .append("; unique: ")
                     .append(sequence.unique())
                     .append("; state: ")
                     .append(sequence.state());
        }
        public Type type() {
            return Type.OPEN;
        }
    };
    
    /**
      * @see ObjectPolicy#NAME
      */
    public CharSequence toString(Sequence<?> sequence) {
        return ObjectPolicy.NAME.toString(sequence);
    }

    /**
      * @return {@link StringablePolicy.Type#APPENDABLE} is returned unless 
      *         explicitly documented otherwise for a given constant. 
      */
    public Type type() {
        return Type.APPENDABLE;
    }
    
}
