/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.strategy;

import dk.rode.thesis.meta.Participant;

/**
  * A <i>stringable</i> object can be formatted into different
  * textual char sequence representations using a given 
  * {@link StringablePolicy policy} to determine the format. 
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @see    StringablePolicy
  * 
  * @param  <T> The type of this stringable object. 
  */
@Participant("Context")
public interface Stringable<T extends Stringable<? super T>> {

    /**
      * Returns a char sequence representation of this stringable object 
      * using the format determined by {@code policy} or the 
      * {@link #getStringablePolicy(StringablePolicy) default} policy
      * in case {@code policy} is null. <p>
      * 
      * In {@code Foo}, a typical implementation of this method 
      * <i>could</i> be:
      * <pre>
      *    public CharSequence toString(StringablePolicy&lt;? super Foo> policy) {
      *      return this.{@link #getStringablePolicy(StringablePolicy) getStringablePolicy}(policy).{@link StringablePolicy#toString(Object) toString}(this);
      *    }
      * </pre>
      * 
      * There are two approaches to formatting this stringable object
      * into a char sequence representation: <ol><p>
      *  
      * <li> Let {@code policy} decide the <i>entire</i> format, 
      *      as in the {@code Foo} example above; or <p>
      *      
      * <li> Use {@code policy} to format <i>part</i> of the overall 
      *      representation, for example letting this method append 
      *      certain text regardless of the policy used. <p>
      *      
      * </ol>
      * 
      * Bullet 1) is not always applicable because a given policy 
      * implementation may not have access to all required information 
      * in its {@link StringablePolicy#toString(Object)} method, for 
      * example in case multiple stringable objects should be formatted 
      * into an overall representation. <p>
      * 
      * In case an implementation uses the approach from bullet 2), care
      * must be take to respect the policy {@link StringablePolicy.Type
      * hints} so the overall format remains meaningful.
      *  
      * @param  policy The policy to dictate the formatting; can be null, in
      *                which case the result of {@code toString} method
      *                is returned. 
      * @return The char sequence representation; never null.
      * @see    StringablePolicy.Type
      */
    public CharSequence toString(StringablePolicy<? super T> policy);
    
    /**
      * Always return a non-null policy: <ol><p>
      * 
      * <li> {@code policy} is not null: {@code policy} is returned. <p>
      * 
      * <li> {@code policy} is null: a default, non-null policy is returned.
      * 
      * </ol>
      * 
      * @param  policy The supplied policy; can be null.
      * @return The policy to use; never null.
      * @see    #toString(StringablePolicy)
      */
    public StringablePolicy<? super T> getStringablePolicy(StringablePolicy<? super T> policy);
        
}
