/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.templatemethod;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.log.LogFactory;
import dk.rode.thesis.strategy.ObjectPolicy;

/**
  * A <i>zip sequence</i> delivers input streams as sequence
  * values, where each stream correspond to the current <i>zip file
  * entry</i>. <p>
  *
  * Note, that since sequences always have a current value, a zip stream
  * sequence cannot describe zip files with no zip entries. <p>
  * 
  * <b>Implementation notes</b>:<br>
  * A zip sequence utilises a single internal zip input stream to
  * deliver all sequence values until reset or restart, where it
  * is discarded, and a new one created and so forth. The returned
  * input streams are protection {@linkplain dk.rode.thesis.proxy proxies} 
  * to ensure that the underlying stream cannot be closed or 
  * reset. <p>
  *  
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("ConcreteClass")
public class ZipSequence extends SequenceTemplate<ZipEntry,InputStream> {
    
    /**
      * The file representing the physical zip file. <p>
      * 
      * Never null.
      */
    private File file;
    /**
      * The zip input stream used to read the zip entries from
      * the zip archive represented by {@link #file}. <p>
      * 
      * Can be null.
      */
    private ZipInputStream stream;
    
    /**
      * Constructor. <p>
      * 
      * @param  file The file representing the physical zip file; cannot be null.
      * @throws NullPointerException If {@code file} is null.
      * @throws SequenceValueException If {@code file} does not exist, or if it 
      *         represents a directory, or if the initialisation fails otherwise. 
      */
    public ZipSequence(File file) throws SequenceValueException {
        super(file);
    }

    /**
      * Constructor. <p>
      * 
      * @param  directory The directory where the zip file is to be found; can be
      *                   null, i.e. "current" directory.
      * @param  filename  The zip file name; cannot be null.
      * @throws NullPointerException If {@code name} is null.
      * @throws SequenceValueException If a zip file with {@code name} does not exist
      *         in {@code directory}, or if it represents a directory, or if the 
      *         initialisation fails otherwise. 
      */
    public ZipSequence(File directory, String filename) throws SequenceValueException {
        super(new File(directory, filename));
    }
    
    /**
      * Copy constructor. <p>
      * 
      * If {@code sequence} is {@link #isClosed() closed}, this 
      * sequence will be closed as well.
      * 
      * @param  sequence The sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      * @throws SequenceValueException If the initialisation fails, or if the 
      *         current value of this sequence cannot be fetched.
      */
    public ZipSequence(ZipSequence sequence) throws SequenceValueException {
        // File is passed to the initialisation method, and buffer size will
        // automatically be inserted by the super constructor...
        super(sequence, sequence.file.getAbsoluteFile());
    }

    /**
      * 
      */
    public ZipSequence copy() {
        return new ZipSequence(this);
    }

    /**
      * 
      */
    public boolean consistent() {
        return false;
    }
    
    /**
      * Opens a new internal zip stream based on the 
      * physical zip file used by this zip sequence.
      * 
      * @throws FileNotFoundException If the file supplied at construction
      *         time does not exist, or if it represents a directory. 
      * @throws IOException If the stream cannot be opened otherwise.
      */
    private final void openStream() throws IOException {
        // Zip streams cannot be reset, so we have to (re-)create...
        this.stream = new ZipInputStream(new FileInputStream(this.file));
        LogFactory.getLog(this).println("Opened stream: ", this.stream);
    }

    /**
      * Closes the current internal zip stream used by this sequence. 
      *  
      * @throws IOException If the zip entry corresponding to the current
      *         value cannot be closed, or if the zip stream cannot be 
      *         closed.
      */
    private final void closeStream() throws IOException {
        if (this.stream != null) {
            try {
                this.stream.closeEntry();
            } finally {
                this.stream.close();
                LogFactory.getLog(this).println("Closed stream: ", this.stream);
                this.stream = null;
            }
        }
    }

    /**
      * @throws NullPointerException If no file is supplied at construction time.
      * @throws Exception If the file supplied at construction time does not exist; if it 
      *                   represents a directory; or if the stream cannot be opened otherwise.
      */
    @Override
    protected boolean sequenceOpen(Object... arguments) throws Exception {
        // Varargs will never be null here, because the applicable 
        // constructors declares a File parameter. The cast below is known 
        // to succeed for the same reason. We could have used an additional 
        // generic type parameter on SequenceTemplate to narrow the argument 
        // type for the constructors and this method even further, but if 
        // more than one argument is needed with different type hierarchies,
        // the actual type parameter would still need to be Object. Here,
        // we could have used File to avoid the explicit cast...
        File file = (File)arguments[0];
        // The actual argument can still be null in case a single null 
        // argument was supplied to the constructor...
        if ((this.file = file) == null) {
            throw new NullPointerException("Zip file cannot be null");
        } 
        // Open stream...
        this.openStream();
        // Ok...
        return true;
    }

    
    /**
      * 
      */
    @Override
    protected void sequenceReset() throws Exception {
        // Close existing stream because a zip stream cannot be reset...
        this.closeStream();
        // ...and create a new one...
        this.openStream(); 
    }
    
    /**
      * @throws Exception If the super sequence close method throws an
      *         exception.
      */
    @Override
    protected void sequenceClose() throws Exception {
        // Close stream...
        this.closeStream(); 
        // Simply for logging purposes, not required!
        super.sequenceClose(); // (abstract) operation
    }
    
    /**
      * @return The next {@code ZipEntry} in the zip stream; can be null in
      *         case the stream has no more entries (restart). 
      * @throws Exception If the zip stream entry represented by 
      *         {@code current} cannot be closed, or if the next
      *         zip entry cannot be fetched. 
      */
    @Override
    protected ZipEntry nextValueOpen(InputStream current) throws Exception {
        // If "old" current is not null, we close the zip entry
        // corresponding to it...
        if (current != null) {
            // Close current stream entry...
            this.stream.closeEntry();
        }
        // Will be null in case there are not more entries...
        return this.stream.getNextEntry();
    }
    
    /**
      * Creates and returns a new input stream to be delivered as 
      * the sequence value.
      */
    @Override
    protected InputStream nextValue(ZipEntry key, InputStream current) {
        // If entry is null, we are about to restart...
        if (key == null) {
            return null;
        }
        // The stream is already positioned, but we must protect it from
        // being closed or reset...
        return new ReadOnlyInputStream();
    }
    
    /**
      * A <i>read-only input stream</i> is a protection proxy (decorator)
      * that ensures that a given input stream cannot be closed or reset 
      * by an external context. <p> 
      *
      * <b>Implementation notes</b>:<br>
      * This is an application of the {@linkplain dk.rode.thesis.proxy Proxy} pattern.
      * 
      * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
      */
    class ReadOnlyInputStream extends InputStream {
    
        /**
          * No-arg constructor.
          */
        ReadOnlyInputStream() {
            super();
        }
    
        /**
          * Return the zip sequence this stream is associated with.
          *  
          * @return The sequence; never null.
          */
        ZipSequence getSequence() {
            return ZipSequence.this;
        }
        
        /**
          * 
          */
        public int read() throws IOException {
            return ZipSequence.this.stream.read();
        }
    
        /**
          * 
          */
        @Override
        public int available() throws IOException {
            return ZipSequence.this.stream.available();
        }
    
        /**
          * All calls are ignored.
          */
        @Override
        public void close() {
            LogFactory.getLog(this).setActive(true).warn("Denied context request to close input stream");
        }
    
        /**
          * 
          */
        @Override
        public void mark(int readlimit) {
            ZipSequence.this.stream.mark(readlimit);
        }
    
        /**
          * 
          */
        @Override
        public boolean markSupported() {
            return ZipSequence.this.stream.markSupported();
        }
    
        /**
          * 
          */
        @Override
        public int read(byte[] b, int off, int len) throws IOException {
            return ZipSequence.this.stream.read(b, off, len);
        }
    
        /**
          * 
          */
        @Override
        public int read(byte[] b) throws IOException {
            return ZipSequence.this.stream.read(b);
        }
    
        /**
          * All calls are ignored.
          */
        @Override
        public void reset() {
            LogFactory.getLog(this).setActive(true).warn("Denied context request to reset input stream");
        }
    
        /**
          * 
          */
        @Override
        public long skip(long n) throws IOException {
            return ZipSequence.this.stream.skip(n);
        }

        /**
          * 
          */
        public String toString() {
            return ObjectPolicy.IDENTITY.asString(this);
        }
        
    }

}