/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.visitor;

import dk.rode.thesis.composite.CompositeSequence;
import dk.rode.thesis.decorator.SequenceDecorator;
import dk.rode.thesis.memento.MemorizableSequence;
import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.ReversibleSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.observer.ObservableSequence;
import dk.rode.thesis.state.StateableSequence;

/**
  * Any sequence <i>decorated</i> by an <i>abstract visitable sequence</i> 
  * implementation is {@link ValueVisitableSequence value} <b>and</b> 
  * {@link TypeVisitableSequence type} visitable. <p>
  * 
  * <b>Implementation notes</b>:<br/>
  * The {@linkplain dk.rode.thesis.decorator Decorator} pattern is used
  * to implement this visitable sequence. <p>
  * 
  * The type visitable functionality is implemented by this class,
  * but the value visitable functionality has to be implemented
  * by specific sub-classes.
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  * 
  * @param  <E> The type of values delivered by this visitable sequence.
  */
@Participant("Element")
public abstract class AbstractVisitableSequence<E> 
              extends SequenceDecorator<E> 
           implements ValueVisitableSequence<E>, TypeVisitableSequence<E> {

    /**
      * Constructor.
      * 
      * @param sequence The decorated sequence that will become visitable;
      *                 cannot be null.
      * @throws NullPointerException If {@code sequence} is null.                  
      */
    protected AbstractVisitableSequence(Sequence<E> sequence) {
        super(sequence);
    }

    /**
      * 
      */
    public <P> void accept(SequenceTypeVisitor<P> visitor, P argument) {
        // Fetch actual sequence, but note that it may be adapted,
        // decorated, or proxied, and hence the actual type may not be
        // inferred...
        Sequence<?> sequence = this.getSequence(false);
        boolean match = false;
        if (sequence instanceof ReversibleSequence) {
            match = true;
            visitor.visitReversible((ReversibleSequence<?>)sequence, argument);
        }
        if (sequence instanceof MemorizableSequence) {
            match = true;
            visitor.visitMemorizable((MemorizableSequence<?>)sequence, argument);
        }
        if (sequence instanceof ObservableSequence) {
            match = true;
            visitor.visitObservable((ObservableSequence<?,?,?>)sequence, argument);
        }
        if (sequence instanceof StateableSequence) {
            match = true;
            visitor.visitStateable((StateableSequence<?>)sequence, argument);
        }
        if (sequence instanceof CompositeSequence) {
            match = true;
            visitor.visitComposite((CompositeSequence<?>)sequence, argument);
        }
        if (!match) {
            visitor.visitUnknown(this, argument);
        }
    }

}
