/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 *
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at
 * http://www.rode.dk/thesis
 *
 */
package dk.rode.thesis.visitor;

import dk.rode.thesis.meta.Participant;
import dk.rode.thesis.meta.model.ReversibleSequence;
import dk.rode.thesis.meta.model.Sequence;
import dk.rode.thesis.state.ReversiblePrimeSequence;
import dk.rode.thesis.state.StateableSequence;

/**
  * A <i>visitable reversible prime sequence</i> is a
  * {@link ReversiblePrimeSequence} that is {@link ValueVisitableSequence
  * value} visitable using the {@link SequenceValueVisitor#visitIntegerValued(Sequence, Object)}
  * visitation method <b>and</b> is {@link TypeVisitableSequence type}
  * visitable as well. <p>
  *
  * This sequence adds no additional state to a reversible prime
  * sequence, only visitation behaviour.
  *
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a>
  */
@Participant("ConcreteElement")
public class VisitableReversiblePrimeSequence
     extends ReversiblePrimeSequence
  implements ValueVisitableSequence<Integer>, TypeVisitableSequence<Integer> {

    /**
      * Constructor.
      *
      * @param  maximum The maximum number to consider. Must be larger than 1.
      * @throws RuntimeException If {@code maximum} is smaller than 2.
      */
    public VisitableReversiblePrimeSequence(int maximum) {
        super(maximum);
    }

    /**
      * Copy constructor.
      *
      * @param  sequence The reversible and/or visitable prime
      *                  sequence to copy; cannot be null.
      * @throws NullPointerException If {@code sequence} is null.
      */
    public VisitableReversiblePrimeSequence(ReversiblePrimeSequence sequence) {
        // ok to copy super class since we add no additional state, behaviour only...
        super(sequence);
    }

    /**
      *
      */
    public VisitableReversiblePrimeSequence copy() {
        return new VisitableReversiblePrimeSequence(this);
    }
    
    /**
      * @see SequenceValueVisitor#visitIntegerValued(Sequence, Object)
      */
    public <P> void accept(SequenceValueVisitor<P> visitor, P argument) {
        visitor.visitIntegerValued(this, argument);
    }

    /**
      * @see SequenceTypeVisitor#visitReversible(ReversibleSequence, Object)
      * @see SequenceTypeVisitor#visitStateable(StateableSequence, Object)
      */
    public <P> void accept(SequenceTypeVisitor<P> visitor, P argument) {
        visitor.visitReversible(this, argument);
        visitor.visitStateable(this, argument);
    }

}
