/*
 * Created: May-August, 2007
 *
 * Evaluating Software Design Patterns
 * - the "Gang of Four" patterns implemented in Java 6
 *
 * Gunni Rode - http://www.rode.dk 
 *
 * Master's Thesis, Computer Science
 * Department of Computer Science,
 * The Faculty of Science,
 * University of Copenhagen
 * 
 * Feel free to use and/or modify the Java 6 source code 
 * developed for this thesis AT YOUR OWN RISK, but note 
 * that the source code comes WITHOUT ANY - and I do mean 
 * WITHOUT ANY - form of warranty WHAT SO EVER!
 *
 * The original thesis and source code are available at 
 * http://www.rode.dk/thesis
 * 
 */

/**
  * Implementations and examples of the <b>Visitor</b> 
  * design pattern [<a href="http://www.rode.dk/thesis/Bibliography.htm#Gamma95" target="thesis-bibliography">Gamma95</a>, p.331]. <p>
  * 
  * <b>Intent</b>:<ul><p><i>
  * 
  *    Represent an operation to be performed on the elements of an object 
  *    structure. Visitor lets you define a new operation without changing 
  *    the classes of the elements on which it operates.
  * 
  * </i></ul>
  * 
  * Here, the <b>Visitor</b> participant is represented by not one, 
  * but two separate visitor implementations, namely the 
  * {@link dk.rode.thesis.visitor.SequenceTypeVisitor} and
  * {@link dk.rode.thesis.visitor.SequenceValueVisitor} classes.
  * The first is a <i>type visitor</i> to visit a given  
  * {@link dk.rode.thesis.meta.model.Sequence} based on its type, 
  * e.g. {@link dk.rode.thesis.meta.model.ReversibleSequence}, 
  * while the latter is a <i>value visitor</i> that visits based on
  * the type of sequence values, e.g. {@code Integer}, {@code Date},
  * {@code CharSequence}, etc. The use of two different visitors 
  * help illustrate visitor usage more clearly than just using a 
  * single type. <p>
  * 
  * The <b>Element</b> participant is represented by the 
  * {@link dk.rode.thesis.visitor.TypeVisitableSequence} and
  * {@link dk.rode.thesis.visitor.ValueVisitableSequence}
  * interfaces, respectively for visitation by {@code SequenceTypeVisitor}
  * and {@code SequenceValueVisitor} implementations. A given 
  * {@code Sequence} may naturally implement both interfaces. The 
  * <b>ConcreteElement</b> participant is represented by any sequence 
  * that implement either or both of these interfaces. <p> 
  * 
  * The <b>ConcreteVisitor</b> participant is represented by any
  * {@code SequenceTypeVisitor} and {@code SequenceValueVisitor} 
  * implementation. In this package, three visitors are implemented, 
  * namely {@link dk.rode.thesis.visitor.TypeVisitor} (<i>type 
  * visitor</i>), {@link dk.rode.thesis.visitor.CountingVisitor} 
  * (<i>value visitor</i>), and 
  * {@link dk.rode.thesis.visitor.LoggingVisitor} (<i>type</i> 
  * <b>and</b> <i>value visitor</i>). <p> 
  * 
  * A tree-like element (sequence) structure is achieved by using the 
  * {@link dk.rode.thesis.composite.CompositeSequence} or 
  * {@link dk.rode.thesis.visitor.VisitableCompositeSequence} types,
  * combined with non-composite (visitable) sequences. To traverse
  * such a {@code Sequence} structure for a given <i>type</i> or <i>value 
  * visitor</i>, {@link dk.rode.thesis.visitor.SequenceTypeScanner} 
  * and {@link dk.rode.thesis.visitor.SequenceValueScanner} 
  * implementations should be used, respectively. They represent the 
  * <b>ObjectStructure</b> participant. A concrete implementation has been
  * made than can traverse a sequence structure for both <i>type</i> and 
  * <i>value visitors</i> in form of
  * {@link dk.rode.thesis.visitor.SimpleScanner}. <p>
  * 
  * This package defines three concrete visitable sequences (excluding
  * the composite visitable sequence): 
  * {@link dk.rode.thesis.visitor.VisitableRandomSequence}, 
  * {@link dk.rode.thesis.visitor.VisitableLongSequence}, and
  * {@link dk.rode.thesis.visitor.VisitableReversiblePrimeSequence}.
  * The first is only visitable by <i>value visitors</i>, while last two
  * can be visited by both <i>type</i> and <i>value visitors</i>. <p>
  * 
  * A visitable sequence rarely adds new state to the sequence it
  * inherits. In fact, the primary reason for inheritance is to
  * implement the {@code TypeVisitableSequence} and/or 
  * {@code ValueVisitableSequence} interfaces because of Java's 
  * fixed class structure. To avoid overly use of inheritance, several
  * <i>wrappers</i> (decorators) have been developed in this package 
  * that will wrap a given sequence and make it visitable. For example,
  * the {@link dk.rode.thesis.visitor.StringValuedVisitableSequence}
  * class allows any {@code Sequence<? extends CharSequence>} type
  * to become both <i>type</i> and <i>value</i> visitable. Such wrappers
  * all extend {@link dk.rode.thesis.visitor.AbstractVisitableSequence}. <p>   
  * 
  * Finally, a <i>reflective wrapper</i> in form of the 
  * {@link dk.rode.thesis.visitor.ReflectiveVisitableSequence} class has been 
  * created that allows <b>any</b> sequence to become both <i>value</i> and <i>type</i> 
  * visitable based on reflective {@link dk.rode.thesis.meta.reflect.TypeLiteral} 
  * functionality. <p> 
  * 
  * <a name="uml"></a><b>UML Class Diagram</b>:<p>
  * <img src="doc-files/Visitor.png" border="0"><p>
  * 
  * <b>Implementation notes</b>:<br/>
  * Though two different types of visitors are used, they operate on the
  * same class hierarchy, i.e. {@code Sequence}, but with different
  * focus, i.e. sequence type vs. sequence values. <p>
  * 
  * {@code SequenceTypeVisitor} implementations are not completely 
  * accurate when decorated, adapted, and/or proxied sequences are
  * used, because the {@code instanceof} test will fail in such cases.
  * However, this is a general problem related to the use of decorators,
  * adapters, and proxies, not specifically to visitors. <p>
  * 
  * While the {@link dk.rode.thesis.visitor.ReflectiveVisitableSequence}
  * class can make any sequence visitable, it may not be able to infer
  * the correct sequence value type in case the sequence is generic. The
  * sequence in question may also deliver a value not supported by the
  * {@code SequenceValueVisitor} interface. In such cases, the 
  * default visitation method is used. <p>  
  *  
  * In the implementation here, visiting a composite sequence is just
  * like visiting a non-composite sequence, because the 
  * {@code SequenceTypeScanner} and {@code SequenceValueScanner}
  * implementations, respectively, will handle the sequence traversal 
  * explicitly.  
  * 
  * @author Gunni Rode / <a href="http://www.rode.dk">rode.dk</a> 
  */
@Pattern(name         = "Visitor", 
         scope        = Scope.Object,
         purpose      = Purpose.Behavioural,
         participants = {"Visitor", "ConcreteVisitor", "Element", "ConcreteElement", "ObjectStructure"})
package dk.rode.thesis.visitor;

import dk.rode.thesis.meta.Pattern;
import dk.rode.thesis.meta.Purpose;
import dk.rode.thesis.meta.Scope;

